!+
!
! Module: lccorr_mod
!
! Authors:  Ian Stewart - SSC
!          
! Description: This task calculates and optionally applies an exposure
! correction for timeseries.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
!

! Version 1.0*****************
!
!
! ToDo:
!********** put in doco/release notes that only handles region filtering in DETX/Y!!! Mention also about square pixels better because => no ellipses.
!********** recommend select PI > about 100 eV to avoid getting the 'Hermite' PossiblyInaccurateResult warning.
!********** regions extending over more than 1 ccd are not yet supported.
!********** binning noise corrn not yet supported. Should run e*events with randomizetimes=yes.
!********** oote corrn not yet supported.
!	- implement RAND_TIM stuff in binning noise correction.
!	- nint() calls deprecated??
!	- Sort out EUNIT/E_UNITS stuff.
!	- look for submode keyword.
!	- doco: treatPatterns has no effect if treatQE=no.
!	- logarithmic energy bins better.
!	- write also comments for *APP keywords.
!	- break into at least 2 modules. It's too big.
!	- write bkg/src area ratio as a keyword?
!	- make the psf normalisation more elegant.
!	- if submode kwd exists, compare with mode worked out within lccorr.
!	- call CAL_releaseMemory when appropriate.
!	- Make test files on the fly! That will shrink the size of the upload by a large factor.
!	Change filter names and warn if filter='NONE' found - see UL emails.
!	If exposure keyword found in ts, overwrite with the time-average of fracexp. evselect writes a rough-and-ready exposure kwd but it deserves to be improved in lcurves that have been through lccorr.
!	Fix up column dimensions stuff - see runrgs on bryan.
!	Look for RAND_TIM kwd (0 or 1) in EXPOSU exts - if 1, don't do binning noise corr for that ccd/node.
!	Check more carefully to see that what happens when the bkg comes from BACKC etc columsn within the src TS is the same as when it comes from a separate file - eg, is it corrected for area? Subtracted if this is desired?
!	Check that, if bkg sub is desired but exp corr not, the bkg is divided by its own fracexp then multiplied by the source one. (Or maybe this is done when the areas are adjusted??)
!	Test for existence of DSS before attempting to open it. Not every TS gets region-filtered. Maybe the whole camera is desired!
!	Pixel areas need to be passed to sub lccorr by calcExposureCorrections. The purpose is not the same as the regionAreaMm2, which is needed in the calcSpatial routines.
!	However..... check this statement, 'the purpose is not the same' out. I have a feeling that it may be the same after all...
!	PN missing frames - see how this is handled.
!	If the mode (eg small window) does not allow a tally of all OOTEs, change the calc of dead time to take into account the region y-dimension for each column. This makes it more accurate.
!	Speed up function psf() by preloading coordinates referred to the psf map, instead of CC2 coords.
!	Change DIM of patternIsCanonical() and patternWasSelected() sometime from 1:32 to 0:31 (saves on lots of -1s in the CAL_getQM calls). NOTE! Tried this and ran into troubles.
!	Copy TIME column to output if it is present in both src and bkg TS. May need to check that all vals of the column are the same for src and bkg. (This may seem like an exacting criterion, but src and bkg should be made from the same evlist.)
!	Fix up messages to format numbers properly (without line breaks!)
!	Modify to cope with unequal-width time bins. Check out whether can enable ((TIME column) .and. (no TIMEDEL kwd)). Does the OGIP standard specify a TIMEDEL column in this case to take care of unequal-width bins?
!	Modify the time routines to use the TIME column where this is available, only calculating bin times from the bin number + srcTsTime1stBin where it is not available.
!	Carefully compare the GTI routine to the subsequent one - changes made to the latter which need to be aped by the former.
!	.Take out workarounds when JB fixes timeunits in evlists (+ exposu exts) from SECONDS to s.
!	Cut up into finer subroutines.
!	Reinstate all the (i) eff area, (ii) psf, (iii) vignet stuff when the relevant CAL routines start working.
!	Clean up some of the -1s inserted when I found that the DAL and CAL return arrays starting from 0 not 1!! Bastards.
!	Change the calculation of the bounding box, eliminating boxM??CC2? and keeping the centre CC2X and Y and minRadius, converting these to pixcoord for each ccd, then working out the BB coords.
!	Some sort of check that the BB actually does contain all the region? maybe another evlist extension containing pixels 1 step outside the BB? Then check that none are within region?
!	Is there a function that returns all upper (or lower) case of a string? Want it when checking region filter shapes.
!	Check that src coords are within region - warn if not.
!	Empty ep frames - how to handle?
!	See DLumb's email of 17 Mar re integ time definition in the ccf.
!	Note in the doco that dss not copied over (because of confusion with bkg)
!	Use foreachsubtable when implemented.
!	Check that all opened objects are closed
!	Minimise memory consumption by not leaving objects open unless nec.
!	Think about timing modes, see if it makes sense to do exp corr for these, and if so which and how.
!	Loops that do more than 1 column at once: see if can break into >1 loop, each doing 1 column.
!	Mention in docs that burst, timing modes not yet supported.
!	Could be a problem if source GTI correction proceeds smoothly, but bkg doesn't. Better if the 2 TS are consistent in the corrections applied.

module lccorr_mod

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.
  use dal
  use cal
  use param  
  use errorHandling
  use dataFilter
  use dss
  use caloaldefs
  use constants
!use plotlib_mod
  implicit none 
  
  private 

! Command-line parameters:
  real(kind=single) ::&
    specBinDiv,&
    pointSrcRA,&
    pointSrcDec
  integer(kind=int16) ::&
    numRBspecRows
  logical ::&
    withBkgTsSet,&
    withSrcSpecSet,&
    withBkgSpecSet,&
    rebinSpec,&
    withSrcImgSet,&
    getPosFromRegn,&
    withPointSrcPosn,&
    withBkgImgSet,&
    treatVignet,&
    treatBorderEvents,&
    treatSelPatterns,&
    treatFilterTrans,&
    treatQuantumEff,&
    treatCosmicRays,&
    treatDeadTime,&
    treatGti,&
    applyCorrections,&
    treatOotes,&
    subtractBkg
  character(len=255) ::&
    srcTsSetName,&
    eventSetName,&
    bkgTsSetName,&
    srcSpecSetName,&
    bkgSpecSetName,&
    outSetName
  character(len=20) ::&
    rebinSpecMode,&
    srcImgModel,&
    bkgImgModel

!.................
! Global parameters:

  real(kind=single), parameter ::&
    detUnitArcsec = 0.05,& ! The unit of the DETX,Y system.
    myRealSingleNull = -999.9
  real(kind=double), parameter ::&
    psfMapPixelWidthArcsec = 5d-1
  integer(kind=int32), parameter ::&
    psfMapHalfsizePixels = 30
  integer(kind=int16), parameter ::&
    numCoarseEnergySamples = 10,&
!    leastNTimeBins = 1,&
    leastNSpecBins = 1
!***** a bit unsatisfactory. Be nice to be able to read them from the (command-line) parameter file.

!.................
! Global variables:

!  character(len=80) ::&
!    eventSetName,&
!    outSetName		! Set from command-line parameters.

  real(kind=double) ::&
    srcTsTimeZero,&
    srcTsTime1stBin,&
    srcTsBinTimeDel
  integer(kind=int32) ::&
    numTsBins
  integer ::&
    numPatterns
  character(len=20) ::&
    srcInstrument,&
    srcTimeUnit
  real(kind=double) ::&
    srcMJDref		! All set from attributes or properties of
			! the source TS file.

  integer(kind=int8) ::&
    numCcds,&
    numNodes
  integer(kind=int8), dimension(12,2) ::&
    ccd2ccdnr  = -1
  integer(kind=int8), dimension(17) ::&
    ccdnr2ccd  = -1,&
    ccdnr2node = -1	! Set within ccdNumbering().

  logical, dimension(12,2) ::&
    exposuExists,&
    neededColumnsPresent,&
    eventTimesRandomized
  logical, dimension(:), allocatable ::&
!    patternIsCanonical,&
    patternWasSelected
  integer(kind=int32), dimension(12,2) ::&
    ccdMode = -1
  integer(kind=int16), dimension(12,2) ::&
    windowMinRawX = -1,&
    windowMinRawY = -1,&
    windowMaxRawX = -1,&
    windowMaxRawY = -1,&
    windowRawDX = -1,&
    windowRawDY = -1
  real(kind=double), dimension(12,2) ::&
    Tshift = -1.0d0,&
    Tdead = -1.0d0
  real(kind=single) ::&
    pixelAreaMm2,&
    mmToDetUnits,&
    psfNormalisation
  real(kind=double) ::&
    pixelWidthXMm,&
    pixelWidthYMm,&
    mmToArcsec,&
    focalLengthMm,&
    radiusFovDET
  real (kind=single) ::&
    srcTsEmin,&
    srcTsEmax,&
    andedEmin,&
    andedEmax
  integer(kind=int16) ::&
    chipMinRawX,&
    chipMaxRawX			! All set within evlistCheckAndSetup()

  logical, dimension(12,2) ::&
    ccdInRegion = .false.	! Set within createPixelLists().

  type arrayType
    real(single) :: lo(3), hi(3), binWidth(3)
    integer(int16) :: iLo(3), iHi(3), numSamples(3)
    character(20) :: units(3)
    real(double), dimension(:,:,:), allocatable :: array
  end type
  type (arrayType) :: cvf

  integer ::&
    numSpecSamples

  real(kind=single) ::&
    boxMinDETX,&
    boxMinDETY,&
    boxMaxDETX,&
    boxMaxDETY	! Set by calculateBoundingBox();
		! used in calculateRegionArea().

  real(kind=single) ::&
    boxMinCC2X,&
    boxMinCC2Y,&
    boxMaxCC2X,&
    boxMaxCC2Y	! Set by calculateBoundingBox();
		! used in createPixelLists().

  public lccorr

contains 
  
  !---------------------------------------------------------------------
  !#####################################################################
  !---------------------------------------------------------------------
  subroutine lccorr
  !---------------------------------------------------------------------
!
! Lccorr performs the following operations:
!	(i) If the input source TS has integer COUNT values, these are
! converted to real-valued count rates by dividing by the bin duration
! and stored in a column RATE in the output TS. An ERROR column, containing
! uncertainties in the count rates, is also created appropriately.
!	(ii) Exposure (+ oote & binning noise) corrections are calculated
! for the source TS. Most individual corrections are optional. The exposure
! factor per bin is written to a column FRACEXP in the output timeseries,
! whereas the oote and binning noise corrections are performed directly on
! the RATE values.
!	(iii) Optionally, the exposure correction is performed by dividing
! the RATE column by the FRACEXP column.
!	(iv) If a bkg TS is supplied, identical calculations are carried
! out on this TS. The options are shared with the source TS - eg it is not
! possible to correct the source TS for vignetting but not the background TS.
!	(v) The bkg count rate values are normalised to the same collection
! area as the source. This is non-optional.
!	(vi) Optionally, the background is subtracted from the source.
! Before the above operations are performed, the relevant columns of the
! source and background TS are copied to the output TS and all the
! manipulations are performed on this output TS. When all operations are
! complete, the columns of the output TS are renamed and pruned as
! appropriate until it complies with the specification for the output of
! lccorr. At present this dictates that the output file has the same
! headers as the source TS, the same number and size of bins, and five
! real32-valued columns:
!		RATE
!		ERROR
!		BACKC
!		BACKE
!		FRACEXP
!........................................................................

    implicit none

    type(DataSetT) ::&
      outSet
    type(TableT)   ::&
      outTab
    real(kind=single) ::&
      srcRegionAreaMm2,&
      srcRegionAreaSr,&
      srcWeightedEffAreaCm2                
    character(len=20) ::&
      bkgInstrument,&
      bkgTimeUnit
    integer(kind=int32) ::&
      numBkgTsBins
    real(kind=single) ::&
      bkgTsEmin,&
      bkgTsEmax,&
      bkgRegionAreaMm2,&
      bkgRegionAreaSr,&
      bkgWeightedEffAreaCm2
    real(kind=double) ::&
      bkgTsBinTimeDel,&
      bkgTsTime1stBin,&
      bkgMJDref,&
      bkgTsTimeZero
    logical ::&
      rebinNecessary

    !--------------------------------------------------------------------
    ! Start:

    call readParameters

!call test
!return

!........................................................................
! Open the source and bkg timeseries and perform some checks:

    call checkTSheader(srcTsSetName, srcInstrument, numTsBins, srcTsEmin&
    , srcTsEmax, srcTimeUnit, srcMJDref, srcTsTimeZero, srcTsBinTimeDel&
    , srcTsTime1stBin)

    if (withBkgTsSet) then
      call checkTSheader(bkgTsSetName, bkgInstrument, numBkgTsBins, bkgTsEmin&
      , bkgTsEmax, bkgTimeUnit, bkgMJDref, bkgTsTimeZero, bkgTsBinTimeDel&
      , bkgTsTime1stBin)

      call compareSrcBkgKwds(&
        srcInstrument,   bkgInstrument,&
        numBkgTsBins,    numTsBins,&
        srcTimeUnit,     bkgTimeUnit,&
        srcMJDref,       bkgMJDref,&
        srcTsTimeZero,   bkgTsTimeZero,&
        srcTsBinTimeDel, bkgTsBinTimeDel,&
        rebinNecessary&
      )
    end if

    call initOutSet(numTsBins)

    andedEmin = srcTsEmin
    andedEmax = srcTsEmax

!........................................................................
! Open the event set and load some arrays from its header keywords and
! those of its EXPOSUnn extensions; load other miscellaneous parameters.

    call ccdNumbering(&
      srcInstrument,&
      ccdnr2ccd,&
      ccdnr2node,&
      ccd2ccdnr,&
      numCcds,&
      numNodes&
    )

    call evlistCheckAndSetup(&
!      exposuExists,&
!      patternIsCanonical,&
!      patternWasSelected,&
      ccdMode,&
      windowMinRawX,&
      windowMinRawY,&
      windowMaxRawX,&
      windowMaxRawY,&
      windowRawDX,&
      windowRawDY,&
      Tshift,&
      Tdead,&
      pixelWidthXMm,&
      pixelWidthYMm,&
      pixelAreaMm2,&
      mmToDetUnits,&
      mmToArcsec,&
      focalLengthMm,&
      radiusFovDET,&
      chipMinRawX,&
      chipMaxRawX&
    )

!........................................................................
! Perform the required exposure corrections on the source dataset:

    call calcExposureCorrections(&
      .true.,&
      withSrcImgSet,&
      srcTsSetName,&
      srcSpecSetName,&
      pointSrcRA,&
      pointSrcDec,&
      srcImgModel,&
      withSrcSpecSet,&
      srcRegionAreaMm2,&
      srcRegionAreaSr,&
      srcWeightedEffAreaCm2&
    )

    outSet = dataSet(outSetName, MODIFY)
    outTab = table(outSet, 'RATE')
    call setAttribute(outTab, 'EFFAREA', srcWeightedEffAreaCm2&
    , comment='Mirror effective area in cm^2, weighted by spectrum.')
    call setAttribute(outTab, 'SKYAREA', srcRegionAreaSr&
    , comment='Area in steradians of collection region')
    call release(outSet)

!........................................................................
!........................................................................

    if (withBkgTsSet) then
      call addBkgColToOutSet(rebinNecessary, numBkgTsBins)

!........................................................................
! Perform the required corrections on the background timeseries:

      call calcExposureCorrections(&
        .false.,&
        withBkgImgSet,&
        bkgTsSetName,&
        bkgSpecSetName,&
        pointSrcRA,&
        pointSrcDec,&
        bkgImgModel,&
        withBkgSpecSet,&
        bkgRegionAreaMm2,&
        bkgRegionAreaSr,&
        bkgWeightedEffAreaCm2&
      )

      call subtractCompensatedBkg(srcRegionAreaMm2, bkgRegionAreaMm2)

    else ! no separate background TS supplied.
      call copyOverBackvColToOutSet(numTsBins)

    end if ! withBkgTsSet.

!........................................................................
! Up until now, all possible information has been preserved - counts, rate,
! error in counts and rate, and fracexp, for both of source
! and background if applicable (although the correction for collection area
! and the background subtraction is at present only applied to RATE, RERROR
! (-> ERROR in the output file), BRATE (-> BACKV) and BRERROR (-> BACKE).
! Now these columns are pruned to just that set desired in the output. If
! the specification for the product changes at a later date, such a change
! can be implemented reasonably simply by changing the routine below.

    call finaliseOutSet()

!    call message(AppMsg, Verbose, 'FINISHED!')

  end subroutine lccorr

  !---------------------------------------------------------------------
  subroutine test
  !---------------------------------------------------------------------

    implicit none 

    integer :: instrument, filter, ccd, node
    integer(int16) :: xsize, ysize, xi, yi
!, n
    integer(int16), dimension(:), allocatable :: rawX, rawY
    integer(int32) :: index, maxIndex, pattern
    real(double), dimension(:,:), allocatable :: filterTrans, totalQe, vigFactor
    real(single), dimension(:), pointer :: cc2X, cc2Y, cc2Z, thetaArcsec, phi
    real(double) :: eV, eVstart, eVfinis, summ
!, frac
    integer(int16), dimension(:,:), pointer :: map
    integer(int16), parameter :: numSamples=100
    real(double), dimension(numSamples) :: totalQeV
!, filterTransV, vigFactorV, eVvar

    !-------------------------------------------------------------------
    ! Start:

!    xsize = 100
!    ysize = 100
!    cc2Xmin = ??????
!    cc2Xmax = ??????
!    cc2Ymin = ??????
!    cc2Ymax = ??????

    instrument = EMOS1
    filter = MEDIUM
    ccd = EMOS_CCD_1
    node = EMOS_PRIMARY_NODE

    call CAL_setState(instrument = instrument)
    call CAL_setState(filterId   = filter)
    call CAL_setState(ccdChipId  = ccd)
    call CAL_setState(ccdNodeId  = node)

    call CAL_getFOVmap(map)!, rawX0=int(1, int16), rawY0=int(1, int16)&
!    , rawXsize=int(20, int16), rawYsize=int(20, int16))
    xsize = size(map, 1)
    ysize = size(map, 2)

    allocate(filterTrans(xsize, ysize), totalQe(xsize, ysize)&
    , vigFactor(xsize, ysize))

    filterTrans(1:xsize, 1:ysize) = map(0:xsize-1, 0:ysize-1)
    totalQe(1:xsize, 1:ysize)     = map(0:xsize-1, 0:ysize-1)
    vigFactor(1:xsize, 1:ysize)   = map(0:xsize-1, 0:ysize-1)

    maxIndex = xsize * ysize - 1
    allocate(rawX(0:maxIndex), rawY(0:maxIndex))
    do xi = 1, xsize
      do yi = 1, ysize
        index = ysize * (xi - 1) + yi - 1
        rawX(index) = xi
        rawY(index) = yi
      end do
    end do

    call CAL_rawXY2mm(rawX, rawY, cc2X, cc2Y, cc2Z, coordSysType=CAMCOORD2)
    call CAL_camCoord2ToTelCoord(cc2X, cc2Y, cc2Z, thetaArcsec, phi)

xi = xsize / 2
yi = ysize / 2
index = ysize * (xi - 1) + yi - 1
eVstart = 100.0
eVfinis = 10000.0

!do n = 1, numSamples
!print *, n, '/', numSamples
!  frac = (n - 1) / real(numSamples - 1)
!  eV = eVstart * (1.0 - frac) + eVfinis * frac
!eVvar(n) = eV
!
!!  filterTransV(n) = CAL_getFilterTransmission(eV&
!!  , real(thetaArcsec(index), double), real(phi(index), double))
!
!  totalQeV(n) = CAL_getQuantumEfficiency(eV, rawX(index)&
!  , rawY(index), patternId=int(0, int32))
!
!!  vigFactorV(n) = CAL_getVignettingFactor(eV&
!!  , real(thetaArcsec(index), double), real(phi(index), double))
!end do
!!call plotLine(int(numSamples), totalQeV, x=eVvar, xLabel='eV', label='QE, M1, chip 1 ctr, pattern 0', device='qe.ps/ps')
!call plotLine(int(numSamples), totalQeV, x=eVvar, xLabel='eV'&
!, label='QE, M1, chip 1 ctr, pattern 0', device='qe.ps/ps')
!return

    eV = 1000.0d0

summ = 0.0
do pattern = 0, numPatterns - 1
  totalQeV(1) = CAL_getQuantumEfficiency(eV, rawX(index)&
  , rawY(index), patternId=pattern)
  summ = summ + totalQeV(1)
print *, pattern, totalQeV(1)
end do
print *, ' '
print *, 'total qe:', summ
return

    do xi = 1, xsize
print *, xi, '/', xsize
      do yi = 1, ysize
        index = ysize * (xi - 1) + yi - 1
        filterTrans(xi, yi) = CAL_getFilterTransmission(eV&
        , real(thetaArcsec(index), double), real(phi(index), double))

!        call calcQe(treatQuantumEff, treatSelPatterns, eV,&
!        rawX(index), rawY(index), totalQe(xi, yi), fracEventsSel)
        totalQe(xi, yi) = CAL_getQuantumEfficiency(eV, rawX(index), rawY(index), patternId=int(0, int32))

        vigFactor(xi, yi) = CAL_getVignettingFactor(eV&
        , real(thetaArcsec(index), double), real(phi(index), double))
      end do
    end do

    call saveImage(filterTrans, xsize, ysize, 'ftrans_image.ds')
    call saveImage(totalQe,     xsize, ysize, 'qe_image.ds')
    call saveImage(vigFactor,   xsize, ysize, 'vignet_image.ds')

  end subroutine test

  !---------------------------------------------------------------------
  subroutine saveImage(image, xsize, ysize, filename)
  !---------------------------------------------------------------------

    implicit none

    integer(kind=int16), intent(in) :: xsize, ysize
    real(kind=double), dimension(xsize, ysize) :: image
    character(len=*), intent(in) :: filename
    real(kind=single) ::&
      yStretch
    type(DataSetT) :: outSet
    type(ArrayT) :: outArray
    integer, dimension(:), pointer ::&
      dimVector
!    real(kind=double) :: xRefPixel, yRefPixel, crpix1, crpix2, crval1, crval2, cdelt1, cdelt2
!    character(len=20) :: ctype1, ctype2, cunit1, cunit2
    real(kind=double), dimension(:,:), pointer ::&
      arrayDataReal64
!    type(TableT), intent(in) ::&
!      eventTab
!    type(ColumnT) ::&
!      xCol, yCol

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Writing output image '&
    //trim(filename)//'.')

    allocate(dimVector(0:1))

    outSet = dataSet(filename, CREATE)
    dimVector(0) = int(xsize) ! the 'ints' just do a conversion from the int16
    dimVector(1) = int(ysize) ! of xsize, ysize to the generic int of dimVector.
yStretch = 1.0

    outArray = addArray(outSet, 'PRIMARY', Real64, dimVector, '', '', 0)

    call setAttribute(outArray, 'HDUCLASS', 'OGIP',&
    "Format conforms to OGIP/GSFC conventions.")

    call setAttribute(outArray, 'HDUCLAS1', 'IMAGE',&
    "File contains an Image.")

!    call setAttribute(outArray, 'HDUCLAS2', 'TOTAL',&
!    "Gross Image.")
!?????

    call setAttribute(outArray, 'HDUVERS1', '1.1.0',&
    "Version of format.")

    arrayDataReal64 => real64Array2Data(outArray)
    arrayDataReal64 = image(1:xsize, 1:ysize)

!*** write dss???

    call release(outSet)

  end subroutine saveImage

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Getting command-line parameters.')

    srcTsSetName = stringParameter('srctsset')
      eventSetName = stringParameter('eventset')

    withBkgTsSet = booleanParameter('withbkgtsset')
    if (withBkgTsSet) then
      bkgTsSetName = stringParameter('bkgtsset')
    end if

    withSrcSpecSet = booleanParameter('withsrcspecset')
    if (withSrcSpecSet) then
!call error('notSupported', 'Sorry, use of spectra is not supported yet.')
      srcSpecSetName = stringParameter('srcspecset')
    end if

    withBkgSpecSet = booleanParameter('withbkgspecset')
    if (withBkgSpecSet) then
!call error('notSupported', 'Sorry, use of spectra is not supported yet.')
      bkgSpecSetName = stringParameter('bkgspecset')
    end if

    numRBspecRows = 0
    specBinDiv = 1.0
    rebinSpec = booleanParameter('rebinspec')
    if (rebinSpec) then
      rebinSpecMode = stringParameter('rebinspecmode')
      if (trim(rebinSpecMode) .eq. 'tonumber') then
        numRBspecRows = intParameter('nspecbins')
      else
        specBinDiv = realParameter('specbindiv')
      end if
    end if

!    withSrcImgSet = booleanParameter('withsrcimgset')
!    if (withSrcImgSet) then
!      srcImgSetName = stringParameter('srcimgset')
!    end if

withSrcImgSet = .false.

    getPosFromRegn = .false.
    srcImgModel = ''
    if (.not. withSrcImgSet) then
      srcImgModel = stringParameter('srcimgmodel')
      if (trim(srcImgModel) .eq. 'point') then
!        withPointSrcPosn = booleanParameter('withpointsrcposn')
withPointSrcPosn = .false.
!        if (withPointSrcPosn) then
!          pointSrcRA = realParameter('pointsrra')
!          pointSrcDec = realParameter('pointsrdec')
!        else
          getPosFromRegn = .true.
!        end if
      end if
    end if

!    withBkgImgSet = booleanParameter('withbkgimgset')
!    if (withBkgImgSet) then
!      bkgImgSetName = stringParameter('bkgimgset')
!    end if

withBkgImgSet = .false.

!    bkgImgModel = ''
    if (.not. withBkgImgSet) then
      bkgImgModel = 'uniform'
    end if

    outSetName = stringParameter('outset')

    treatVignet = booleanParameter('treatvignet')
!    treatBorderEvents = booleanParameter('treatborderevents')
!    treatSelPatterns = booleanParameter('treatselpatterns')
treatSelPatterns = .true.
    treatFilterTrans = booleanParameter('treatfiltertrans')
    treatQuantumEff = booleanParameter('treatquantumeff')
    treatCosmicRays = booleanParameter('treatcosmicrays')
    treatDeadTime = booleanParameter('treatdeadtime')
    treatGti = booleanParameter('treatgti')

    applyCorrections = booleanParameter('applycorrections')

    treatOotes = booleanParameter('treatootes')  
!    if (treatOotes .and. .not. useEventSet) then
!      treatOotes = .false.
!      write(errstr, *) "Can't correct for OOTEs because &
!      &you haven't supplied an event list."
!      call warning('OoteNoEventSet')
!    end if

    subtractBkg = booleanParameter('subtractbkg')

! NOTE!!! Rebinning of time-dependent factors is currently not done. With
! the present algorithms for calculating fracexp, little or no time would
! be saved by rebinning.
!    nTimeBins = 0
!    timeBinDiv = 1.0
!    rebinTime = booleanParameter('rebintime')
!    if (rebinTime) then
!      rebinTimeMode = stringParameter('rebintimemode')
!      if (trim(rebinTimeMode) .eq. 'tonumber') then
!        nTimeBins = intParameter('ntimebins')
!      else
!        timeBinDiv = realParameter('timebindiv')
!      end if
!    end if

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine checkTSheader(tsSetName, tsInstrum, tsNumBins, tsEmin, tsEmax&
  , tsTimeUnit, tsMJDref, tsTimeZero, tsBinTimeDel, tsTime1stBin)
  !---------------------------------------------------------------------

    implicit none

    character(len=255), intent(in) ::&
      tsSetName
    character(len=20), intent(out) ::&
      tsInstrum, tsTimeUnit
    real(kind=double), intent(out) ::&
      tsMJDref, tsTimeZero, tsBinTimeDel, tsTime1stBin
    integer(kind=int32), intent(out) ::&
      tsNumBins
    real(kind=single), intent(out) :: &
      tsEmin, tsEmax

    type(DataSetT) ::&
      tsSet
    type(TableT) ::&
      tsTab
    type(ColumnT) ::&
      timeCol
    real(kind=double), dimension(:), pointer ::&
      time
    character(len=20) ::&
      tsEunit,&
      timeColUnits
    real(kind=double) ::&
      tsTimeZerF
    integer(kind=int32) ::&
      tsTimeZerI

character(len=255) :: eUnitLabel
integer(kind=int16), dimension(:), pointer ::&
  EBchannels
real(kind=double), dimension(:), pointer ::&
  Emin,&
  Emax

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, "Checking headers of "//trim(tsSetName))

    tsSet = dataSet(tsSetName, READ)
    tsInstrum = stringAttribute(tsSet, 'INSTRUME')

    if (trim(tsInstrum) .ne. 'EMOS1'&
    .and. trim(tsInstrum) .ne. 'EMOS2'&
    .and. trim(tsInstrum) .ne. 'EPN') then

      write(errstr, *) "Instrument "//trim(tsInstrum)//&
      " not supported."
call error('badInstrum')
    end if

    call CAL_setState(instrument=instId(trim(tsInstrum)))

!........................................................................
! Open the RATE table of the timeseries and perform some checks:

    tsTab = table(tsSet, 'RATE')
    tsNumBins = numberOfRows(tsTab)
    if (tsNumBins .le. 0) then 
      write(errstr, *) "The timeseries contains no data."
call error('tsNoData')
    end if

    if (hasAttribute(tsTab, 'NUMBAND')) then
      if (int16Attribute(tsTab, 'NUMBAND') .ne. 1) then
        write(errstr, *) "The timeseries contains data in more &
        &than 1 energy band."
call error('tsTooManyEbands')
      end if
    end if

    tsEmin = real32Attribute(tsTab, 'E_MIN')
    tsEmax = real32Attribute(tsTab, 'E_MAX')
!**** OGIP says the name should be EUNIT (although it's optional).
    if (hasAttribute(tsTab, 'E_UNITS')) then
      tsEunit = stringAttribute(tsTab, 'E_UNITS')
      if (index(tsEunit, 'CHAN') .gt. 0) then ! try to find out whether
      ! it is PHA or PI channels:
        eUnitLabel = label(attribute(tsTab, 'E_UNITS')) !****** 
!**** this is not a good way to get this info. Another keyword?
        if (index(eUnitLabel, 'PI') .gt. 0) then
          call CAL_getEbounds(EBchannels, Emin, Emax)
          if (tsEmax .gt. size(Emin) - 1) then
call error('tsBadChanRange', 'The maximum PI channel of the time series &
&is greater than that returned by CAL_getEbounds.')
          end if
          tsEmin = real(Emin(int(tsEmin)), single)
          tsEmax = real(Emax(int(tsEmax)), single)

        else
call error('tsAmbigEunit', "Energy bounds not in PI channels - can't deduce the energy bounds of the time series in eV.")
        end if
      end if
      tsEunit = 'eV'
    else
      tsEunit = 'keV'
    end if

    if (trim(tsEunit) .eq. 'keV') then
      tsEmin = tsEmin * 1000.0
      tsEmax = tsEmax * 1000.0
    else if (trim(tsEunit) .ne. 'eV') then
      write(errstr, *) "Unrecognised E_UNITS keyword "//trim(tsEunit)//&
      " in the timeseries."
call error('tsBadEunit')
    end if

    tsTimeUnit = stringAttribute(tsTab, 'TIMEUNIT')
    tsMJDref = real64Attribute(tsTab, 'MJDREF')

    if (hasColumn(tsTab, 'TIME')) then
      timeCol = column(tsTab, 'TIME')
      timeColUnits = units(timeCol)

! The following is a workaround until JB fixes emevents.
if (trim(timeColUnits) .eq. 'SECONDS') then
  call warning('deprecatedUnit', "OGIP standards require that the TIME column have units 's' not 'SECONDS'.")
  timeColUnits = 's'
end if
      if (trim(timeColUnits) .ne. ''&
     .and. trim(timeColUnits) .ne. trim(tsTimeUnit)) then
        write(errstr, *) "The timeseries has inconsistent time units."
call error('tsBadTunits')
      end if

      if (hasAttribute(tsTab, 'TIMEZERI')&
      .and. hasAttribute(tsTab, 'TIMEZERF')) then
        tsTimeZerI = int32Attribute(tsTab, 'TIMEZERI')
        tsTimeZerF = real64Attribute(tsTab, 'TIMEZERF')
        tsTimeZero = tsTimeZerI + tsTimeZerF ! type conversion to real64

      else if (hasAttribute(tsTab, 'TIMEZERO')) then
        tsTimeZero = real64Attribute(tsTab, 'TIMEZERO')

      else
        tsTimeZero = 0.0d0
      end if

      time => real64data(timeCol)
      tsTime1stBin = time(1 - 1)
      call release(timeCol)

      if (.not. hasAttribute(tsTab, 'TIMEDEL')) then
        write(errstr, *) "The timeseries has a TIME column, but no TIMEDEL keyword. For the moment this is not supported."
call error('tsBadTimeStyle')
      end if
!**** This may be temporary. Later I may adjust the CR etc subs so they calculate bin times using the TIME column when this is available.

      tsBinTimeDel = real64Attribute(tsTab, 'TIMEDEL')

    ! No TIME column; but...
    else if (hasAttribute(tsTab, 'TIMEDEL')) then
      tsBinTimeDel = real64Attribute(tsTab, 'TIMEDEL')

      if (hasAttribute(tsTab, 'TIMEZERI')&
      .and. hasAttribute(tsTab, 'TIMEZERF')) then
        tsTimeZerI = int32Attribute(tsTab, 'TIMEZERI')
        tsTimeZerF = real64Attribute(tsTab, 'TIMEZERF')
        tsTimeZero = tsTimeZerI + tsTimeZerF ! type conversion to real64

      else
        tsTimeZero = real64Attribute(tsTab, 'TIMEZERO')
      end if

      tsTime1stBin = tsTimeZero

    else
      write(errstr, *) "The timeseries has neither a TIME &
      &column nor a TIMEDEL keyword."
call error('tsNoTimes')
    end if

    if (withBkgTsSet .and. hasColumn(tsTab, 'BACKV')) then
      write(errstr, *) "The timeseries already has a BACKV &
      &column. This data will be discarded."

      call warning('tsHasBackv')
    end if

    call release(tsSet)

  end subroutine checkTSheader

  !---------------------------------------------------------------------
  subroutine initOutSet(numTsBins)
  !---------------------------------------------------------------------

    implicit none

    integer(kind=int32), intent(in) ::&
      numTsBins

    type(DataSetT) ::&
      outSet, srcTsSet
    type(TableT)   ::&
      outTab, srcTsTab
    type(ColumnT)  ::&
      srcTsCountsCol,&
      srcTsErrorCol,&
      outCountsCol,&
      outErrorCol,&
      outRateCol,&
      outRerrorCol,&
      outFracexpCol,&
      outBkgCountsCol,&
      outBkgErrorCol,&
      outBkgRateCol,&
      outBkgRerrorCol,&
      outBkgFracexpCol
    integer(kind=int32) ::&
      bin
    integer(kind=int8),  dimension(:), pointer ::&
      srcTsCounts8,&
      srcTsError8
    integer(kind=int16), dimension(:), pointer ::&
      srcTsCounts16,&
      srcTsError16
    integer(kind=int32), dimension(:), pointer ::&
      outCounts,&
      srcTsCounts32,&
      srcTsError32
    real (kind=single),  dimension(:), pointer ::&
      outError,&
      outBkgRate,&
      outBkgRerror

    !-------------------------------------------------------------------
    ! Start:

!........................................................................
! Create the output data set and copy over header keywords, RATE table
! COUNTS and ERROR columns and RATE table keywords from the source dataset:

    call message(AppMsg, Verbose, 'Copying keywords and columns to output TS.')

    srcTsSet = dataSet(srcTsSetName, READ)
    srcTsTab = table(srcTsSet, 'RATE')
    outSet = dataSet(outSetName, CREATE)
    call setAttributes(attributable(outSet), attributable(srcTsSet))
    outTab = addTable(outSet, 'RATE', numTsBins)
    call setAttributes(attributable(outTab), attributable(srcTsTab))
    call deleteAttribute(outTab, 'HDUCLAS2')
    call deleteAttribute(outTab, 'HDUCLAS3')
    call setAttribute(outTab, 'HDUCLAS3', 'RATE'&
    , 'Data is in the form of counts per sec.')

    call dssDelete(block(outSet, 'RATE')) ! Gets rid of DSS attributes
    ! carried over from the source RATE header. We can't include a DSS
    ! in the output file because in general it contains both a src and bkg
    ! TS, each of which has at least a different selection region.

    outCountsCol     = addColumn(outTab, 'COUNTS',   Integer32, 'counts')
    outErrorCol      = addColumn(outTab, 'ERROR',    Real32,    'counts')
    outRateCol       = addColumn(outTab, 'RATE',     Real32,    'counts/s')
    outRerrorCol     = addColumn(outTab, 'RERROR',   Real32,    'counts/s')
    outFracexpCol    = addColumn(outTab, 'FRACEXP',  Real32)
    outBkgCountsCol  = addColumn(outTab, 'BCOUNTS',  Integer32, 'counts')
    outBkgErrorCol   = addColumn(outTab, 'BERROR',   Real32,    'counts')
    outBkgRateCol    = addColumn(outTab, 'BRATE',    Real32,    'counts/s')
    outBkgRerrorCol  = addColumn(outTab, 'BRERROR',  Real32,    'counts/s')
    outBkgFracexpCol = addColumn(outTab, 'BFRACEXP', Real32)
! If we were being really picky (as we have been elsewhere in the code -
! shutup, who said that??!) we would close one before adding another.
! But TS aren't bigger than a few k... I hope...

    outBkgRate   => real32Data(outBkgRateCol)
    outBkgRerror => real32Data(outBkgRerrorCol)
    outBkgRate   = 0.0
    outBkgRerror = 0.0
!****** this zeroing is done because, if no bkg ts is supplied, and the src ts has no BACKV, BACKE columns, at present these columns remain in the output and are not filled. They should in fact be deleted from the output.

!outTimeCol  = addColumn(outTab, 'TIME',  Real64,    's')
!outTime => real64Data(outTimeCol)
!srcTimeCol = column(srcTsTab, 'TIME')
!srcTime => real64Data(srcTimeCol)
!outTime = srcTime
!call release(srcTimeCol)

    outCounts => int32Data(outCountsCol)

    srcTsCountsCol = column(srcTsTab, 'COUNTS')
    if (columnDataType(srcTsCountsCol) .eq. INTEGER8) then
      srcTsCounts8 => int8Data(srcTsCountsCol)
      outCounts = srcTsCounts8 ! type conversion to int32.
    else if (columnDataType(srcTsCountsCol) .eq. INTEGER16) then
      srcTsCounts16 => int16Data(srcTsCountsCol)
      outCounts = srcTsCounts16 ! type conversion to int32.
    else
      srcTsCounts32 => int32Data(srcTsCountsCol)
      outCounts = srcTsCounts32
    end if
!************** shouldn't you call dataType rather than columnDataType???

    do bin = 1, numTsBins
      if (outCounts(bin - 1) .lt. 0) then
        write(errstr, *) "COUNTS value < 0 detected in the &
        &RATE extension of the source timeseries."
call error('srcTsCountsLeZero')
      end if
    end do

    call release(outCountsCol)

    outError => real32Data(outErrorCol)
    if (hasColumn(srcTsTab, 'ERROR')) then
      srcTsErrorCol = column(srcTsTab, 'ERROR')
      if (columnDataType(srcTsErrorCol) .eq. INTEGER8) then
        srcTsError8 => int8Data(srcTsErrorCol)
        outError = srcTsError8 ! type conversion to real32
      else if (columnDataType(srcTsErrorCol) .eq. INTEGER16) then
        srcTsError16 => int16Data(srcTsErrorCol)
        outError = srcTsError16 ! type conversion to real32
      else
        srcTsError32 => int32Data(srcTsErrorCol)
        outError = srcTsError32 ! type conversion to real32
      end if

      call release(srcTsErrorCol)

    else ! check for POISSERR keyword:
      if (.not. hasAttribute(srcTsTab, 'POISSERR')) then
        write(errstr, *) "There is neither a POISSERR keyword or an &
        &ERROR column in the RATE extension of dataset "//&
        trim(srcTsSetName)//". Poisson uncertainties assumed anyway."

        call warning('srcTsNoPoisserrKwd')

      else if (.not. booleanAttribute(srcTsTab, 'POISSERR')) then
        write(errstr, *) "The RATE extension header of dataset "&
        //trim(srcTsSetName)//" has the POISSERR keyword set to F, &
        &but there is no ERROR column. Don't know how to deal with &
        &uncertainties."
call error('srcTsBadUncertType')
      end if

      outCountsCol = column(outTab, 'COUNTS')
      outCounts => int32Data(outCountsCol)
      outError = sqrt(real(outCounts, single)) ! vector op;
    end if

    call release(srcTsSet)

    if (applyCorrections) then
      call setAttribute(outTab, 'DEADAPP', .true., '')
      call setAttribute(outTab, 'VIGNAPP', .true., '')
    else
      call setAttribute(outTab, 'DEADAPP', .false., '')
      call setAttribute(outTab, 'VIGNAPP', .false., '')
    end if

    call release(outSet)

  end subroutine initOutSet

  !---------------------------------------------------------------------
  subroutine ccdNumbering(instrument, ccdnr2ccd, ccdnr2node, ccd2ccdnr, numCcds, numNodes)
  !---------------------------------------------------------------------

    implicit none

!.................
! Arguments in:

    character(len=20), intent(in) ::&
      instrument

!.................
! Arguments modified:

    integer(kind=int8), intent(in out) ::&
      numCcds,&
      numNodes
    integer(kind=int8), dimension(:,:), intent(in out) ::&
      ccd2ccdnr
    integer(kind=int8), dimension(:), intent(in out) ::&
      ccdnr2ccd,&
      ccdnr2node

!.................
! Other variables:

    integer(kind=int8) ::&
      ccd,&
      node,&
      ccdnr

    !-------------------------------------------------------------------
    ! Start:

    if (trim(instrument) .eq. 'EMOS1'&
    .or. trim(instrument) .eq. 'EMOS2') then
      do ccd = 1, 7
        do node = 1, 2
          ccdnr = 10 * (node - 1) + ccd
          ccd2ccdnr(ccd, node) = ccdnr
          ccdnr2ccd(ccdnr) = ccd
          ccdnr2node(ccdnr) = node
        end do
      end do
      numCcds = 7
      numNodes = 2

    else ! EPN.
      do ccd = 1, 12
        ccdnr = ccd
        ccd2ccdnr(ccd, 1) = ccd
        ccdnr2ccd(ccdnr) = ccd
        ccdnr2node(ccdnr) = 1
      end do
      numCcds = 12
      numNodes = 1

    end if

  end subroutine ccdNumbering

  !---------------------------------------------------------------------
  subroutine evlistCheckAndSetup(&
!    exposuExists,&
!    patternIsCanonical,&
!    patternWasSelected,&
    ccdMode,&
    windowMinRawX,&
    windowMinRawY,&
    windowMaxRawX,&
    windowMaxRawY,&
    windowRawDX,&
    windowRawDY,&
    Tshift,&
    Tdead,&
    pixelWidthXMm,&
    pixelWidthYMm,&
    pixelAreaMm2,&
    mmToDetUnits,&
    mmToArcsec,&
    focalLengthMm,&
    radiusFovDET,&
    chipMinRawX,&
    chipMaxRawX)
  !---------------------------------------------------------------------

    implicit none

!.................
! Arguments modified:

!    logical, dimension(:,:), intent(in out) ::&
!      exposuExists
    integer(kind=int32), dimension(:,:), intent(in out) ::&
      ccdMode
    integer(kind=int16), dimension(:,:), intent(in out) ::&
      windowMinRawX,&
      windowMinRawY,&
      windowMaxRawX,&
      windowMaxRawY,&
      windowRawDX,&
      windowRawDY
    real(kind=double), dimension(:,:), intent(in out) ::&
      Tshift,&
      Tdead
    real(kind=single), intent(in out) ::&
      pixelAreaMm2,&
      mmToDetUnits
    real(kind=double), intent(in out) ::&
      pixelWidthXMm,&
      pixelWidthYMm,&
      mmToArcsec,&
      focalLengthMm,&
      radiusFovDET
    integer(kind=int16), intent(in out) ::&
      chipMinRawX,&
      chipMaxRawX

!.................
! Other vars:

    character(len=20) ::&
      eventInstrument,&
      eventTimeUnit,&
      filterStr
    type(DataSetT) ::&
      eventSet
    type(TableT) ::&
      exposuTab,&
      eventTab
    type(BlockT) ::&
      exposuBlock
    real(kind=double) ::&
      eventMJDref,&
!      pixelX,&
!      pixelY,&
      radiusFovDeg
    integer ::&
      ccd,&
      node,&
      ccdnr,&
      patternID
    logical ::&
      noExposuExtensions
    character(len=8) ::&
      exposuTabName
    real(kind=single) ::&
      nominalTint,&
      nominalTframe
    type(ModeParameterDataType) ::&
      modeParams
    integer(int8), dimension(:,:,:), pointer ::&
      patterns

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Checking event list.')

!print *, eventSetName
    call message(AppMsg, Verbose, '  Opening event list.')
    eventSet = dataSet(eventSetName, READ)

    call message(AppMsg, Verbose, '  Checking header attributes.')
    ! check that it has same INSTRUME, TIMEUNIT and MJDREF kwds as the
    ! source timeseries.
    eventInstrument = stringAttribute(eventSet, 'INSTRUME')
    if (trim(eventInstrument) .ne. trim(srcInstrument)) then
      write(errstr, *) "INSTRUME keyword in "//trim(eventSetName)//" doesn't &
      &match that of the source timeseries."
call error('evlistWrongInstrume')
    end if

    filterStr = stringAttribute(eventSet, 'FILTER')
!print *, filterStr, '****'
!call error('','')
!    if (trim(eventInstrument) .ne. trim(srcInstrument)) then
!    end if
    call CAL_setState(filterId = filterId(filterStr))
!iprint *, 'set cal'
    eventTab = table(eventSet, 'EVENTS')

    eventTimeUnit = stringAttribute(eventTab, 'TIMEUNIT')
    if (trim(eventTimeUnit) .ne. trim(srcTimeUnit)) then
      write(errstr, *) "TIMEUNIT keyword in "//trim(eventSetName)//" doesn't &
      &match that of the source timeseries."
call error('evlistWrongTimeunit')
    end if

    eventMJDref = real64Attribute(eventTab, 'MJDREF')
    if (eventMJDref .ne. srcMJDref) then
      write(errstr, *) "Value of MJDREF keyword in "&
      //trim(eventSetName)//" doesn't match that of the source timeseries."
call error('evlistWrongMJDref')
    end if

    call message(AppMsg, Verbose, '  Reading attributes from evlist &
    &EXPOSU extensions.')

    noExposuExtensions   = .true.  ! default
    neededColumnsPresent = .false. ! array op, default
    exposuExists         = .false. ! array op, default
    eventTimesRandomized = .false. ! array op, default

    do ccd = 1, numCcds
      do node = 1, numNodes
        ccdnr = ccd2ccdnr(ccd, node)
        write(exposuTabName, '(A, I2.2)') 'EXPOSU', ccdnr
        if (hasBlock(eventSet, exposuTabName)) then
          exposuBlock = block(eventSet, exposuTabName)
          call CAL_setState(exposuBlock)
          call CAL_getState(ccdModeId=ccdMode(ccd, node))
          windowMinRawX(ccd, node) = int16attribute(exposuBlock, 'WINDOWX0')
          windowMinRawY(ccd, node) = int16attribute(exposuBlock, 'WINDOWY0')
          windowRawDX(ccd, node) = int16attribute(exposuBlock, 'WINDOWDX')
          windowRawDY(ccd, node) = int16attribute(exposuBlock, 'WINDOWDY')
          windowMaxRawX(ccd, node) = windowMinRawX(ccd, node)&
          + windowRawDX(ccd, node) - 1
          windowMaxRawY(ccd, node) = windowMinRawY(ccd, node)&
          + windowRawDY(ccd, node) - 1

          exposuTab = table(eventSet, exposuTabName)
          if (hasColumn(exposuTab, 'FRACEXP')&
          .and. hasColumn(exposuTab, 'TIMEDEL')&
          .and. hasColumn(exposuTab, 'TIME')) then
            neededColumnsPresent(ccd, node) = .true.
          end if

          call CAL_getModeParameters(modeParams)
          nominalTint = modeParams%integrationTime / 1000.0
          nominalTframe = modeParams%frameTime / 1000.0
          ! Nominal because Tint (and therefore Tframe) can vary by some
          ! amount. The actual values will be calculated on a
          ! frame-by-frame basis from the EXPOSUnn extensions of the
          ! whole-camera event list.

          Tshift(ccd, node) = modeParams%smearedFraction&
          * nominalTint
!**** CHECK! SF may in fact be defined as Tshift / (Tint + Tshift), in which
! case you will want
!
!          Tshift(ccd, node) = nominalTint * (smearedFraction&
!          / (1 - smearedFraction))
!
! instead.

          Tdead(ccd, node) = nominalTframe&
          * (1 - modeParams%dutyCycle)

          if (nominalTframe .ge. srcTsBinTimeDel) then
            write(errstr, *) "The frame duration of ", nominalTframe,&
            " s in the "//trim(exposuTabName)//" extension of "&
            //trim(eventSetName)//" is longer than the bin duration of ",&
            srcTsBinTimeDel," s in the time series."
            call warning('frameLength2Long')
          end if

          exposuExists(ccd, node) = .true.
          noExposuExtensions = .false.

          if (hasAttribute(exposuTab, 'RAND_TIM')) then
            if (booleanAttribute(exposuTab, 'RAND_TIM')) then
              eventTimesRandomized(ccd, node) = .true.
            end if
          end if
        end if
      end do ! next node
    end do ! next ccd
    call release(eventSet)

    if (noExposuExtensions) then
      write(errstr, *) "There are no EXPOSUnn extensions in "&
      //trim(eventSetName)//". Can't do exposure corrections."
call error('noExposuExts')
    end if

    call message(AppMsg, Verbose, '  Miscellaneous checks.')
    do ccd = 1, numCcds
      do node = 1, numNodes
      if (.not. exposuExists(ccd, node)) cycle
        if (.not.&
            ((ccdMode(ccd, node) .eq. PRIME_FULL_WINDOW)&
        .or. (ccdMode(ccd, node) .eq. PRIME_PARTIAL_RFS)&
        .or. (ccdMode(ccd, node) .eq. PRIME_PARTIAL_W2)&
        .or. (ccdMode(ccd, node) .eq. PRIME_PARTIAL_W3)&
        .or. (ccdMode(ccd, node) .eq. PRIME_PARTIAL_W4)&
        .or. (ccdMode(ccd, node) .eq. PRIME_PARTIAL_W5)&
        .or. (ccdMode(ccd, node) .eq. PRIME_PARTIAL_W6)&
!        .or. (ccdMode(ccd, node) .eq. EPN_PRIME_FULL_WINDOW)&
        .or. (ccdMode(ccd, node) .eq. PRIME_FULL_WINDOW_EXTENDED)&
        .or. (ccdMode(ccd, node) .eq. PRIME_LARGE_WINDOW)&
        .or. (ccdMode(ccd, node) .eq. PRIME_SMALL_WINDOW))) then
          write(errstr, *) "Instrument mode", ccdMode(ccd, node),&
          " not supported."
call error('badInstrumeMode')
        end if
      end do ! next node
    end do ! next ccd

    ! The next few quantities are not ccd- or node-specific:
    call CAL_getMiscellaneousData('MM_PER_PIXEL_X', pixelWidthXMm)
    call CAL_getMiscellaneousData('MM_PER_PIXEL_Y', pixelWidthYMm)
    call CAL_getMiscellaneousData('FOCAL_LENGTH', focalLengthMm)
    call CAL_getMiscellaneousData('FOV_RADIUS', radiusFovDeg)

    pixelAreaMm2 = pixelWidthXMm * pixelWidthYMm ! total area of 1 pixel
    ! in mm^2. Implicit type conversion double -> single.

    mmToArcsec = RadToDeg * (3600.0 / focalLengthMm)

    mmToDetUnits = mmToArcsec / detUnitArcsec
    ! implicit type conversion double -> single.

    radiusFovDET = radiusFovDeg * (3600.0 / detUnitArcsec)

    if (trim(srcInstrument) .eq. 'EMOS1'&
    .or. trim(srcInstrument) .eq. 'EMOS2') then
      chipMinRawX = 1
      chipMaxRawX = 600
    else ! EPN.
      chipMinRawX = 1
      chipMaxRawX = 64
    end if
!***** is there a cal routine that returns these? Doesn't seem to be. CAL_getMiscData doesn't return it.

    call CAL_getEventPatterns(patterns)
    numPatterns = size(patterns, 1)
    allocate(patternWasSelected(0:numPatterns))

!******* The following is a bit of a temporary bodge. Eventually should set
! patternWasSelected to just those patterns actually selected by the user
! (info in dss?).
    do patternID = 0, numPatterns - 1
!      patternIsCanonical(patternID) = .true.
      patternWasSelected(patternID) = .true.
    end do

  end subroutine evlistCheckAndSetup

  !---------------------------------------------------------------------
  subroutine addBkgColToOutSet(rebinNecessary, numBkgTsBins)
  !---------------------------------------------------------------------

    implicit none

    logical, intent(in) ::&
      rebinNecessary
    integer(int32), intent(in) ::&
      numBkgTsBins
!    integer(kind=int32), intent(in) ::&
!      numTsBins

    type(DataSetT) ::&
      outSet, bkgTsSet
    type(TableT)   ::&
      outTab, bkgTsTab
    type(ColumnT)  ::&
      bkgTsCountsCol,&
      bkgTsErrorCol,&
!      outCountsCol,&
!      outErrorCol,&
!      outRateCol,&
!      outRerrorCol,&
!      outFracexpCol,&
      outBkgCountsCol,&
      outBkgErrorCol
!,&
!      outBkgRateCol,&
!      outBkgRerrorCol,&
!      outBkgFracexpCol
    integer(kind=int32) ::&
      bin
    integer(kind=int8),  dimension(:), pointer ::&
      bkgTsCounts8,&
      bkgTsError8
    integer(kind=int16), dimension(:), pointer ::&
      bkgTsCounts16,&
      bkgTsError16
    integer(kind=int32), dimension(:), pointer ::&
      outBkgCounts,&
      bkgTsCounts32,&
      bkgTsError32
    real (kind=single),  dimension(:), pointer ::&
      outBkgError, bkgTsCounts, bkgTsError
!,&
!      outRate,&
!      outRerror
    real(double) ::&
      min,&
      max,&
      bkgMin,&
      bkgMax

    !-------------------------------------------------------------------
    ! Start:

!........................................................................
! Copy bkg COUNTS & ERROR columns to the output dataset:

    call message(AppMsg, Verbose, 'Copying bkg columns to output TS.')

    allocate(bkgTsCounts(0:numBkgTsBins - 1))
    bkgTsSet = dataSet(bkgTsSetName, READ)
    bkgTsTab = table(bkgTsSet, 'RATE')
    bkgTsCountsCol = column(bkgTsTab, 'COUNTS')
    if (columnDataType(bkgTsCountsCol) .eq. INTEGER8) then
      bkgTsCounts8 => int8Data(bkgTsCountsCol)
      bkgTsCounts = bkgTsCounts8 ! type conversion to int32
    else if (columnDataType(bkgTsCountsCol) .eq. INTEGER16) then
      bkgTsCounts16 => int16Data(bkgTsCountsCol)
      bkgTsCounts = bkgTsCounts16 ! type conversion to int32
    else
      bkgTsCounts32 => int32Data(bkgTsCountsCol)
      bkgTsCounts = bkgTsCounts32
    end if
    call release(bkgTsCountsCol)

    do bin = 1, numBkgTsBins
      if (bkgTsCounts(bin - 1) .lt. 0) then
        write(errstr, *) "COUNTS value < 0 detected in the &
        &RATE extension of the background timeseries."
call error('bkgTsCountsLeZero')
      end if
    end do

    if (rebinNecessary) then
      call message(AppMsg, Verbose, '  Rebinning background TS.')

      min = srcTsTime1stBin - srcTsBinTimeDel / 2.0
      max = min + numTsBins * srcTsBinTimeDel
      bkgMin = srcTsTime1stBin - srcTsBinTimeDel / 2.0
      bkgMax = min + numTsBins * srcTsBinTimeDel
      call rebinTS(bkgTsCounts, numTsBins, numBkgTsBins, min, max, bkgMin, bkgMax)
    end if
!print *, 'rebinned bkg'

    outSet = dataSet(outSetName, MODIFY)
    outTab = table(outSet, 'RATE')
    call setAttribute(outTab, 'HDUCLAS2', 'TOTAL'&
    , 'Contains both source and background data.')
    outBkgCountsCol = column(outTab, 'BCOUNTS')
    outBkgCounts => int32Data(outBkgCountsCol)
    outBkgCounts = bkgTsCounts
    call release(outBkgCountsCol)
    deallocate(bkgTsCounts)
!print *, 'put bkg into outset'

    if (hasColumn(bkgTsTab, 'ERROR')) then
      allocate(bkgTsError(0:numBkgTsBins - 1))
      bkgTsErrorCol = column(bkgTsTab, 'ERROR')
      if (columnDataType(bkgTsErrorCol) .eq. INTEGER8) then
        bkgTsError8 => int8Data(bkgTsErrorCol)
        bkgTsError = bkgTsError8 ! type conversion to real32
      else if (columnDataType(bkgTsErrorCol) .eq. INTEGER16) then
        bkgTsError16 => int16Data(bkgTsErrorCol)
        bkgTsError = bkgTsError16 ! type conversion to real32
      else
        bkgTsError32 => int32Data(bkgTsErrorCol)
        bkgTsError = bkgTsError32 ! type conversion to real32
      end if
      call release(bkgTsErrorCol)
!print *, 'got bkg errors'

      if (rebinNecessary) then
        call rebinTS(bkgTsError, numTsBins, numBkgTsBins, min, max, bkgMin&
        , bkgMax)
      end if
!print *, 'rebinned bkg errors'

      outBkgErrorCol = column(outTab, 'BERROR')
      outBkgError => real32Data(outBkgErrorCol)
      outBkgError = bkgTsError
!      call release(outBkgErrorCol)
      deallocate(bkgTsError)
!print *, 'put bkg errors into outset'

    else ! check for POISSERR keyword:
      if (.not. hasAttribute(bkgTsTab, 'POISSERR')) then
        write(errstr, *) "There is neither a POISSERR keyword or &
        &an ERROR column in the RATE extension of dataset "&
        //trim(bkgTsSetName)//". Poisson uncertainties assumed anyway."

        call warning('bkgTsNoPoisserrKwd')

      else if (.not. booleanAttribute(bkgTsTab, 'POISSERR')) then
        write(errstr, *) "The RATE extension header of dataset "&
        //trim(bkgTsSetName)//" has the POISSERR keyword set to F, &
        &but there is no ERROR column. Don't know how to deal with &
        &uncertainties."
call error('bkgTsBadUncertType')
      end if

      outBkgCountsCol = column(outTab, 'BCOUNTS')
      outBkgCounts => int32Data(outBkgCountsCol)
      outBkgErrorCol = column(outTab, 'BERROR')
      outBkgError => real32Data(outBkgErrorCol)
      outBkgError = sqrt(real(outBkgCounts, single)) ! vector op;
    end if

    call release(bkgTsSet)
    call release(outSet)

  end subroutine addBkgColToOutSet

  !---------------------------------------------------------------------
  subroutine calcExposureCorrections(&
    isSource,&
    withImg,&
    inSetName,&
    specSetName,&
    srcRA,&
    srcDec,&
    imageModel,&
    useSpecSet,&
    regionAreaMm2,&
    regionAreaSr,&
    weightedEffAreaCm2)
  !---------------------------------------------------------------------

    implicit none

!...........
! Arguments in:

    logical, intent(in) ::&
      isSource,&
      withImg
    character(len=80), intent(in) ::&
      inSetName,&
      specSetName
    real, intent(in) ::&
      srcRA,&
      srcDec

!...........
! Arguments modified:

    character(len=8), intent(in out) ::&
      imageModel
    logical, intent(in out) ::&
      useSpecSet

!...........
! Arguments out:

    real(kind=single), intent(out) ::&
      regionAreaMm2,&
      regionAreaSr,&
      weightedEffAreaCm2

!...........
! Other arguments:

    character(len=8) ::&
      countsName,&
      rateName,&
      countsErrorName,&
      rateErrorName,&
      fracexpName,&
!      pixelListTabName,&
      tdxColName
    type(DataSetT) ::&
      outSet,&
      pixelListSet,&
      tempSpecSet,&
      tdxSet
    type(TableT) ::&
      outTab,&
      tdxTab
    type(ColumnT) ::&
      countsCol,&
      rateCol,&
      countsErrorCol,&
      rateErrorCol,&
      fracexpCol,&
      tdxCol
    integer(kind=int32), dimension(:), pointer ::&
      counts
    integer ::&
      bin
    integer(int16) ::&
      firstChan, lastChan
    real(kind=single), dimension(:), pointer ::&
      rate,&
      countsError,&
      rateError,&
      tdx,&
      fracexp
    character(len=80) ::&
      selectionExpression
    real(kind=single) ::&
      srcCC2X,&
      srcCC2Y,&
      srcTheta,&
      srcPhi
    real(kind=single), dimension(0:0) ::&
      dummyX,&
      dummyY,&
      dummyZ
    real(kind=single), dimension(:), pointer ::&
      dummyTheta,&
      dummyPhi
!    real(kind=single) ::&
!      boxMinDETX,&
!      boxMinDETY,&
!      boxMaxDETX,&
!      boxMaxDETY	! Set by calculateBoundingBox();
!			! used in calculateRegionArea().
!
!    real(kind=single) ::&
!      boxMinCC2X,&
!      boxMinCC2Y,&
!      boxMaxCC2X,&
!      boxMaxCC2Y	! Set by calculateBoundingBox();
!			! used in createPixelLists().

    real(kind=single), dimension(:,:,:), allocatable ::&
      regionYExtentFrac
    integer(kind=int8) ::&
      numCcdsInRegion,&
      ccd,&
      node,&
      nccd,&
      nnode,&
      canonicalCcd,&
      canonicalNode,&
      ccdnr
    real(kind=single), dimension(12,2) ::&
      aveOoteFraction = 0.0
    logical ::&
      mustAverageTdfs,&
      instrumentMismatch
    real(kind=single), dimension(:), allocatable ::&
      product,&
      avTdx
    logical, dimension(:), allocatable ::&
      atLeastOneNonZero
    real(kind=single) ::&
      numerator,&
      denominator,&
      spatial
    real(kind=single), dimension(12,2) ::&
      eventFraction = 0.0
    integer(kind=int32), dimension(12,2) ::&
      numPixels
!    real(kind=single), dimension(:,:), pointer ::&
!      PSFmap
!    integer :: i, j
!    real :: sum
!    real(kind=double) :: mapPixelWidthMm
real :: realDummy

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Entering subroutine calcExposureCorrections.')

    ! Construct outSet column names:
    if (isSource) then
      countsName      = 'COUNTS'
      countsErrorName = 'ERROR'
      rateName        = 'RATE'
      rateErrorName   = 'RERROR'
      fracexpName     = 'FRACEXP'
    else
      countsName      = 'BCOUNTS'
      countsErrorName = 'BERROR'
      rateName        = 'BRATE'
      rateErrorName   = 'BRERROR'
      fracexpName     = 'BFRACEXP'
    end if

!........................................................................
! Convert COUNTS to RATE and adjust ERROR to match. Note: srcTsBinTimeDel
! .gt. 0.0 should already have been tested in the calling routine.

    outSet = dataSet(outSetName, MODIFY)
    outTab = table(outSet, 'RATE')

    countsCol = column(outTab, countsName, READ)
    rateCol = column(outTab, rateName)
    counts => int32Data(countsCol)
    rate   => real32Data(rateCol)

    rate = counts / srcTsBinTimeDel ! vector op; type conversion to real32

    call release(countsCol)
    call release(rateCol)

    countsErrorCol = column(outTab, countsErrorName, READ)
    rateErrorCol = column(outTab, rateErrorName)
    countsError => real32Data(countsErrorCol)
    rateError   => real32Data(rateErrorCol)

    rateError = countsError / srcTsBinTimeDel ! vector op

    call release(outSet)

!........................................................................
! If relevant, transform source RA and dec to TELCOORD system:
!***** maybe move this codelet to calculateBoundingBox() where its other half resides?

    if (.not. withImg .and. imageModel .eq. 'point'&
    .and. .not. getPosFromRegn) then
! eventually, calculate srcCC2X and srcCC2Y from srcRA, srcDec. For the moment:
      write(errstr, *) "Sorry, input of source RA and dec is &
      &not presently supported."
call error('notSupported')
realDummy = srcRA  ! Just to avoid
realDummy = srcDec ! compiler warnings
    end if

!........................................................................
!........................................................................
! In the next few calls collection area event lists are constructed in a
! temporary file. What we want to end up with is one event list for every
! ccd/node that occupies some part of the collection region, each list
! having 1 event for each pixel of its ccd/node that lies within the region.
! The lists will have RAWX, RAWY, DETX, DETY columns. An additional master
! event list will also be created. This will contain all the events (and
! no more) that are in the individual lists and will therefore have an
! additional column CCDNR. It is necessary to have this master list because
! subroutine calcSpatial10a() can't work efficiently otherwise.
!
! The work is done in several steps, as follows:
!
! (i) The region filter in the DSS of the relevant timeseries is used to
! initialise the selectlib filter.
! (ii) If possible, an estimate is made of the size of a box that just
! encloses the region. (This is to avoid the use of unnecessarily hefty
! chunks of memory in the step (iii).)(where it was not possible to
! estimate a box, a box enclosing the field of view defined by the filter
! wheel is created.)
! (iii) For each ccd/node:
! 	- A temporary event list is constructed which has one event for
!       each pixel of its ccd/node that lies within the region bounding box.
!	- This event list is filtered by the region.
!	- The BADPIXnn extensions of the whole-camera event list (to be
! distinguished from the temporary event list mentioned above) are used to
! filter out bad pixels from the temp event lists for each ccd/node.
!
!........................................................................
! (i) Open DSS and calculate the filtering expression from the region spec
! in the DSS of the TS:

    call getSelectionExpression(inSetName, selectionExpression)

!........................................................................
! (ii) Calculate (if possible) the bounds of a square enclosing the region.

    call calculateBoundingBox(inSetName)

    if (getPosFromRegn) then
      srcCC2X = (boxMinCC2X + boxMaxCC2X) / 2.0
      srcCC2Y = (boxMinCC2Y + boxMaxCC2Y) / 2.0
    end if

!........................................................................
! (iii) Make and filter the event lists:

    pixelListSet = dataSet('tempeventset.fits', TEMP)

    call CAL_setState(accuracyLevel=ACCURACY_MEDIUM)
    call CAL_setState(randomize=.false.)
    ! The 'ACCURACY_MEDIUM' is the
    ! accuracy level (mainly of importance for the PSF calculation) and
    ! the 'false' turns off randomisation of integer pixel coordinates
    ! in the conversion from PIXCOORD1 to CAMCOORD2. I don't want this
    ! randomisation because I am not doing coordinate conversion on
    ! actual events, but on a fake event list, 1 per pixel, designed to
    ! facilitate calculation of the spatial exposure factors.

    allocate(regionYExtentFrac(12,2,chipMinRawX:chipMaxRawX))

    call createPixelLists(&
      pixelListSet,&
      selectionExpression,&
      srcCC2X,&
      srcCC2Y,&
      imageModel,&
      ccdInRegion,&
      regionYExtentFrac,&
      numCcdsInRegion,&
      numPixels&
    )

!........................................................................
! Calculate the total area of the region:

    call calculateRegionArea(&
      selectionExpression,&
      regionAreaMm2&
    )

    regionAreaSr = regionAreaMm2 / focalLengthMm / focalLengthMm

!........................................................................
! Whether a model or measured spectrum is used, in either case it is stored in the temporary data set 'tempspecset.fits', with columns as follows:
!
!	BCTRCHAN	- the bin centre in PI channels
!	BWIDCHAN	- the bin width in PI channels
!	BINCTREV	- the bin centre in eV
!	BINWIDEV	- the bin width in eV.
!	RATE		- the count rate
!	EFFAREA0	- the effective area
!
! For a measured spectrum, copying its count values to this temporary file has three advantages: it allows channels to be discarded which have energies outside the range of the source time series; it allows rebinning to a smaller number of bigger channels; and it funnels the two possible types of data (RATE vs COUNTS) into one. When the spatial exposure corrections are calculated, integrals over energy are approximated by a sum over a series of energy samples, these samples comprising the BINCTREV vector.
!
! This temporary spectrum is constructed next:

    call checkCalGetEB
    if (useSpecSet) then
      call checkSpectrum(specSetName, firstChan, lastChan)
      call loadInternalSpectrum(specSetName, firstChan, lastChan, tempSpecSet)
      call calcEffArea(tempSpecSet, weightedEffAreaCm2)
      if (rebinSpec) then
        call rebinSpectrum(specSetName, tempSpecSet)
      end if
    else
      call makeModelSpectrum(tempSpecSet)
      call calcEffArea(tempSpecSet, weightedEffAreaCm2)
    end if

!    call checkAndRebinSpectrum(&
!      specSetName,&
!      rebinSpecMode,&
!      specBinDiv,&
!      useSpecSet,&
!      rebinSpec,&
!      tempSpecSet&
!    )

!........................................................................
! Calculate the effective area as a function of energy. This is stored in
! the temp spec file. Why? Uh, well... it's just a handy place to have it.

!    call calcEffArea(tempSpecSet, weightedEffAreaCm2)

    call loadCoarseCalArrays

!........................................................................
! Calculate the time-dependent exposure factor tdx(total_num_ccds, 1-2,
! numTsBins) and also calculate framesInBin(total_num_ccds, 1-2, numTsBins),
! an array needed for the binning-noise correction. These are both stored
! in the temporary table TDXTAB.

    call calcGtiCorrection(&
      inSetName,&
      treatGti,&
      tdxSet&
    )

    call calcOtherTimeDepCorrns(&
      tdxSet,&
      inSetName,&
      treatDeadTime,&
      treatCosmicRays,&
      aveOoteFraction&
    )

!........................................................................
! Calculate srcTheta, srcPhi:

    dummyX(0) = srcCC2X
    dummyY(0) = srcCC2Y
    dummyZ(0) = 0.0d0
    call CAL_camCoord2ToTelCoord(dummyX, dummyY, dummyZ, dummyTheta, dummyPhi)
    srcTheta = dummyTheta(0)
    srcPhi = dummyPhi(0)
    call CAL_releaseMemory(dummyTheta, dummyPhi)

!........................................................................
!........................................................................
! The main routine for calculating FRACEXP.

    call message(AppMsg, Verbose, '  Calculating spatial corrections.')

!***** put the following in a sub??

    ! Do some error checks first:
    mustAverageTdfs = .true. ! The default value.
!print *, mustAverageTdfs
    if (numCcdsInRegion .gt. 1 .and. useSpecSet) then
!print *, 'should not be here!'
      ccdLoop: do ccd = 1, numCcds
        do node = 1, numNodes
          if (ccdInRegion(ccd, node)) then
            canonicalCcd = ccd
            canonicalNode = node
      exit ccdLoop
          end if
        end do
      end do ccdLoop

      instrumentMismatch = .false. ! default
      do node = canonicalNode + 1, numNodes
        if (ccdInRegion(ccd, node) .and. ccdMode(canonicalCcd, node)&
        .ne. ccdMode(canonicalCcd, canonicalNode)) then
          instrumentMismatch = .true.
        end if
      end do

      do ccd = canonicalCcd + 1, numCcds
        do node = 1, numNodes
          if (ccdInRegion(ccd, node) .and. ccdMode(ccd, node)&
          .ne. ccdMode(canonicalCcd, canonicalNode)) then
            instrumentMismatch = .true.
          end if
        end do
      end do

      if (instrumentMismatch) then
! If image is supplied, don't have to average the tdfs; but at present,
! supply of an image isn't supported.

        write(errstr, *) "The CCDs within the region don't all &
        &have the same instrument mode. In this case you won't get an &
        &accurate result if you use a measured spectrum."

        call warning('ccdDiffMode')
      end if

      allocate(product(0:numTsBins-1), atLeastOneNonZero(numTsBins))
      atLeastOneNonZero = .false. ! vector op; default
      product = 1.0               ! vector op; default
      tdxTab = table(tdxSet, 'TDXTAB')
      do ccd = 1, numCcds
        do node = 1, numNodes
        if (.not. ccdInRegion(ccd, node)) cycle
          ccdnr = ccd2ccdnr(ccd, node)
          write(tdxColName, '(A, I2.2)') 'TDXCOL', ccdnr
          tdxCol = column(tdxTab, tdxColName)
          tdx => real32data(tdxCol)
          product = product * tdx ! vector op
          do bin = 1, numTsBins
            if (tdx(bin - 1) .ne. 0.0) then
              atLeastOneNonZero(bin) = .true.
            end if
          end do ! next bin
          call release(tdxCol)
        end do
      end do
      call release(tdxTab)

      do bin = 1, numTsBins
        if (product(bin - 1) .eq. 0.0 .and. atLeastOneNonZero(bin)) then
        ! one (but not all) of the ccds is out of GTI, therefore the GTIs
        ! of the ccds aren't aligned.
! If image is supplied, don't have to average the tdfs; but at present,
! supply of an image isn't supported.

          write(errstr, *) "The CCDs within the region have &
          &non-aligned GTIs. In this case you won't get an accurate &
          &result if you use a measured spectrum."

          call warning('ccdDiffGtis')
      exit
        end if
      end do ! next bin
      deallocate(product, atLeastOneNonZero)
    end if ! numCcdsInRegion .gt. 1 .and. useSpecSet

!print *, mustAverageTdfs

!........................................................................
! Now work out which version of the correction equations to use:

    outSet = dataSet(outSetName, MODIFY)
    outTab = table(outSet, 'RATE')
    fracexpCol = column(outTab, fracexpName)
    fracexp => real32data(fracexpCol)
    fracexp = 0.0 ! default. *********Make sure this is really what you want.

!print *, 'nccd=', numCcdsInRegion, '; useSpecSet=', useSpecSet

    if (numCcdsInRegion .eq. 1) then
      ccdLoop2: do nccd = 1, numCcds
        do nnode = 1, numNodes
          if (ccdInRegion(nccd, nnode)) then
            ccd = nccd
            node = nnode
      exit ccdLoop2
          end if
        end do
      end do ccdLoop2

!      ccdnr = ccd2ccdnr(ccd, node)
!      write(pixelListTabName, '(A, I2.2)') 'EVENTS', ccdnr

      if (useSpecSet) then
        call calcSpatial00(&
          tempSpecSet,&
          pixelListSet,&
          ccd,&
          node,&
          imageModel,&
          srcTheta,&
          srcPhi,&
          srcCC2X,&
          srcCC2Y,&
          regionAreaMm2,&
          treatVignet,&
          treatBorderEvents,&
!          treatSelPatterns,&
          treatFilterTrans,&
!          treatQuantumEff,&
          numerator,&
          denominator,&
          eventFraction&
        )

        if (denominator .eq. 0) then
          write(errstr, *) "A zero-valued denominator was returned by &
          &subroutine calcSpatial00. Can't calculate fracexp."

call error('denomZero00')
        end if

        spatial = numerator / denominator
!print *, 'spatial, numerator, denominator', spatial, numerator, denominator
!print *, ' '

        tdxTab = table(tdxSet, 'TDXTAB')
        ccdnr = ccd2ccdnr(ccd, node)
        write(tdxColName, '(A, I2.2)') 'TDXCOL', ccdnr
        tdxCol = column(tdxTab, tdxColName)
        tdx => real32data(tdxCol)
        fracexp = tdx * spatial ! vector op.
!do i = 0, numberOfRows(tdxTab) - 1
!print *, fracexp(i), tdx(i)
!end do
!print *, tdxColName
!print *, ' '
        call release(tdxTab)

      else ! use model spectrum.
!print *, ccd, node
!call error('','')

        call calcSpatial01(&
          tempSpecSet,&
          pixelListSet,&
          ccd,&
          node,&
          imageModel,&
          srcTheta,&
          srcPhi,&
          srcCC2X,&
          srcCC2Y,&
          regionAreaMm2,&
          treatVignet,&
          treatBorderEvents,&
!          treatSelPatterns,&
          treatFilterTrans,&
!          treatQuantumEff,&
          numerator,&
          denominator,&
          eventFraction&
        )

        if (denominator .eq. 0) then
          write(errstr, *) "A zero-valued denominator was returned by &
          &subroutine calcSpatial01. Can't calculate fracexp."

call error('denomZero01')
        end if

        spatial = numerator / denominator
!print *, 'numerator, denominator, numerator/denominator', numerator, denominator, spatial

        tdxTab = table(tdxSet, 'TDXTAB')
        ccdnr = ccd2ccdnr(ccd, node)
        write(tdxColName, '(A, I2.2)') 'TDXCOL', ccdnr
        tdxCol = column(tdxTab, tdxColName)
        tdx => real32data(tdxCol)
        fracexp = tdx * spatial ! vector op
!do i = 0, numberOfRows(tdxTab) - 1
!print *, i, tdx(i), fracexp(i)
!end do
        call release(tdxTab)
      end if
!call error('','')

    else ! numCcdsInRegion must be > 1.
!print *, 'here if >1 ccd'
      if (useSpecSet) then
!print *, '>1 ccd, use real spec? Should not be here!'
        if (mustAverageTdfs) then
          ! Average the tdx across all CCDs within the region:
          allocate(avTdx(0:numTsBins-1))
          avTdx = 0.0 ! vector op; default
          tdxTab = table(tdxSet, 'TDXTAB')
          do ccd = 1, numCcds
            do node = 1, numNodes
            if (.not. ccdInRegion(ccd, node)) cycle
              ccdnr = ccd2ccdnr(ccd, node)
              write(tdxColName, '(A, I2.2)') 'TDXCOL', ccdnr
              tdxCol = column(tdxTab, tdxColName)
              tdx => real32data(tdxCol)
              avTdx = avTdx + tdx ! vector op
              call release(tdxCol)
            end do ! next node
          end do ! next ccd
          call release(tdxTab)
          avTdx = avTdx / real(numCcdsInRegion, single) ! vector op

          call calcSpatial10a(&
            tempSpecSet,&
            pixelListSet,&
            imageModel,&
            srcTheta,&
            srcPhi,&
            srcCC2X,&
            srcCC2Y,&
            regionAreaMm2,&
            numPixels,&
            treatVignet,&
            treatBorderEvents,&
!            treatSelPatterns,&
            treatFilterTrans,&
!            treatQuantumEff,&
            numerator,&
            denominator,&
            eventFraction&
          )

          if (denominator .eq. 0) then
            write(errstr, *) "A zero-valued denominator was returned &
            &by subroutine calcSpatial10. Can't calculate fracexp."

call error('denomZero10')
          end if

          spatial = numerator / denominator
          fracexp = avTdx * spatial ! vector op
          deallocate(avTdx)

        else ! mustAverageTdfs = .false.??? Shouldn't happen in current program.
!print *, '>1 ccd, use real spec, no av tdfs? Should not be here! -> calcFracexp10b'
          write(errstr, *) "This mode is currently not supported."
call error('notSupported')
!          call calcFracexp10b() ! Doesn't exist at present. Maybe later.
        end if ! mustAverageTdfs.
      else ! use model spectrum.
!print *, '>1 ccd, use mod spec -> calcFracexp11'
        call calcFracexp11(&
          tempSpecSet,&
          tdxSet,&
          pixelListSet,&
          imageModel,&
          srcTheta,&
          srcPhi,&
          srcCC2X,&
          srcCC2Y,&
          treatVignet,&
          treatBorderEvents,&
!          treatSelPatterns,&
          treatFilterTrans,&
!          treatQuantumEff,&
          fracexp,&
          eventFraction&
        )

      end if
    end if

    call release(outSet)
    call release(tempSpecSet)
    call release(pixelListSet)

    call deallocateCoarseCalArrays

!........................................................................
! Correct for OOTEs:

    if (treatOotes) then
      call calculateOotes(&
        regionYExtentFrac,&
        aveOoteFraction,&
        rateName,&
        rateErrorName,&
        treatOotes&
      )
    end if

    deallocate(regionYExtentFrac)

!........................................................................
! Correct for binning noise.

    call correctBinningNoise(&
      tdxSet,&
      eventFraction,&
      rateName,&
      rateErrorName,&
      fracexpName&
    )

    call release(tdxSet)

!........................................................................
! Apply exposure corrections:

    if (applyCorrections) then
      call message(AppMsg, Verbose, '  Applying exposure corrections.')

      outSet = dataSet(outSetName, MODIFY)
      outTab = table(outSet, 'RATE')
      rateCol = column(outTab, rateName)
      rateErrorCol = column(outTab, rateErrorName)
      fracexpCol = column(outTab, fracexpName)
      rate      => real32data(rateCol)
      rateError => real32data(rateErrorCol)
      fracexp   => real32data(fracexpCol)

      do bin = 1, numTsBins
        if (fracexp(bin - 1) .gt. 0.0) then
          rate(bin - 1) = rate(bin - 1) / fracexp(bin - 1)
          rateError(bin - 1) = rateError(bin - 1) / fracexp(bin - 1)
        else
          call setToNull(rateCol, bin - 1)
          call setToNull(rateErrorCol, bin - 1)
        end if
      end do ! next bin.

      call release(outSet)
    end if ! applyCorrections.

  end subroutine calcExposureCorrections

  !---------------------------------------------------------------------
  subroutine getSelectionExpression(inSetName, selectionExpression)
  !---------------------------------------------------------------------

    ! The DSS of a timeseries contains (in general) several components,
    ! which are OR-ed together to produce the filtering expression that
    ! was used to produce the parent event list. Each component can be
    ! thought of as a volume in an abstract space of which the columns of
    ! the event list form the axes. A component is made up by ANDing
    ! together 1 or more filters, each filter comprising a set of range
    ! criteria for a single 'axis' (event list column). Typical filter
    ! elements for timeseries are as follows:
    !
    !	Type:		Axes:		Remarks:
    !---------------------------------------------------------
    !	Range		CCDNR		Eg, 3.5 < CCDNR < 4.5.
    !	GTI		TIME
    !	Region		DETX, DETY	Expected to be identical for
    !					all components.
    !---------------------------------------------------------
    !
    ! NOTE!! DSS components are _not_ the same as region components,
    ! although the concepts are similar.

    implicit none

!.................
! Arguments coming in:

    character(len=80), intent(in) ::&
      inSetName

!.................
! Arguments going out:

    character(len=80), intent(out) ::&
      selectionExpression

!.................
! Others:

    type(DataSetT) ::&
      inSet
    type(BlockT) ::&
      rateBlock
    type(DssT) ::&
      dss
    integer ::&
!      numFilt,&
!      filterNum,&
      numComp
    logical ::&
      regionExists
    type(DScompT) ::&
      dssComp
    type(DSfilterT) ::&
!      filter,&
      regionFilter
    character(len=80) ::&
      regionSetBlockSpec
    type(TableT) ::&
      regionTable
    integer ::&
      numRegionRows

    character(len=3), parameter ::&
!      colFilteredName = 'POS'
      colFilteredName = 'DET'
!*** should be 'DET' but there is a bug in Dean's DSS code.

!........................................................................
! Start:

    call message(AppMsg, Verbose, '  Getting selectionExpression.')

    inSet = dataSet(inSetName, READ)
    rateBlock = block(inSet, 'RATE')
    dss = dataSubspace(rateBlock)
    numComp = dssComponents(dss)
!print *, 'num dss comps is', numComp
    regionExists = .false. ! default.
    if (numComp .gt. 0) then
      dssComp = dssComponent(dss, 0) ! The region filters in the other
      ! components (if present) should all be identical, because the
      ! components differ only in the CCD they relate to.

      if (dssHasFilter(dssComp, colFilteredName, REGIONBLOCKFILTERT)) then
        regionFilter = dssFilter(dssComp, colFilteredName, REGIONBLOCKFILTERT)

        regionSetBlockSpec = dssFilterBlock(regionFilter)

        regionTable = table(inSet, regionSetBlockSpec)
        numRegionRows = numberOfRows(regionTable)
        if (numRegionRows .gt. 0) then
          selectionExpression = selectExpression(regionFilter)
          call initFilter(selectionExpression)
          regionExists = .true.
        end if
      end if
    end if

    call release(inSet)

    if (.not. regionExists) then
      write(errstr, *) "No region filter was found in the &
      &DSS of dataset "//trim(inSetName)//". Exposure corrections can't be &
      &performed."
call error('DssNoRegionFilter')
    end if

  end subroutine getSelectionExpression

  !---------------------------------------------------------------------
  subroutine calculateBoundingBox(inSetName)
  !---------------------------------------------------------------------
  ! The routine works as follows: If there is only 1 component, then it must
  ! by definition consist of the intersection (AND) of several simpler shapes
  ! which are the elements of the component (see McDowell and Rots' FITS region
  ! definition document). IF at least 1 of these shapes is either circular or
  ! annular, then the whole component is contained within the circular or
  ! annular element with the smallest outer radius.

    implicit none

!.................
! Arguments coming in:

!    logical, intent(in) ::&
!      getPosFromRegn
    character(len=80), intent(in) ::&
      inSetName

!.................
! Arguments modified:

!    real(kind=single), intent(in out) ::&
!      srcCC2X,&
!      srcCC2Y
!    character(len=8), intent(in out) ::&
!      imageModel

!.................
! Arguments going out:

!    real(kind=single), intent(out) ::&
!      boxMinDETX,&
!      boxMinDETY,&
!      boxMaxDETX,&
!      boxMaxDETY	! used in calculateRegionArea().
!
!    real(kind=single), intent(out) ::&
!      boxMinCC2X,&
!      boxMinCC2Y,&
!      boxMaxCC2X,&
!      boxMaxCC2Y	! used in createPixelLists().

!.................
! Others:

    type (DataSetT) ::&
      inSet
    type(BlockT) ::&
      rateBlock
    type(DssT) ::&
      dss
    type(DScompT) ::&
      dssComp
    type(DSfilterT) ::&
      regionFilter
    character(len=80) ::&
      regionSetBlockSpec
    type(TableT) ::&
      regionTable
    integer ::&
      numDimensions
    logical ::&
      canDoCompactBox,&
      regionFormIsSimple,&
      atLeastOneCircOrAnn
    type(ColumnT) ::&
      compCol,&
      rCol,&
      shapeCol,&
      xCol,&
      yCol
    integer(kind=int8), dimension(:), pointer ::&
      comp
    integer(kind=int32) ::&
      element,&
      numRegionRows,&
      elementWithMinRad,&
      startElement
    character(16), allocatable, dimension(:) ::&
      shape
    logical, allocatable, dimension(:) ::&
      isCircOrAnn
    real(kind=single), allocatable, dimension(:) ::&
      outerRadius
    real(kind=single), dimension(:), pointer ::&
      radius1dim,&
      elementDETX,&
      elementDETY
    real(kind=single), dimension(:,:), pointer ::&
      radius2dim
    real(kind=single) ::&
      minRadiusDET

    character(len=3), parameter ::&
!      colFilteredName = 'POS'
      colFilteredName = 'DET'
!*** should be 'DET' but there is a bug in Dean's DSS code.
    character(len=3), parameter ::&
      colPrefix = 'DET'

!........................................................................
! Get the region filter from the TS DSS and do some checks:

    call message(AppMsg, Verbose, '  Calculating bounding box.')

    inSet = dataSet(inSetName, READ)
    rateBlock = block(inSet, 'RATE')
    dss = dataSubspace(rateBlock)

    dssComp = dssComponent(dss, 0) ! The region filters in the other
    ! components (if present) should all be identical, because the
    ! components differ only in the CCD they relate to. Note: have
    ! already tested (in the routine getSelectionExpression) to see that
    ! the DSS has at least 1 component.

    regionFilter = dssFilter(dssComp, colFilteredName, REGIONBLOCKFILTERT)

    regionSetBlockSpec = dssFilterBlock(regionFilter)
    regionTable = table(inSet, regionSetBlockSpec)
    numRegionRows = numberOfRows(regionTable) ! Implicit type conversion
    ! from default integer to int32.

    regionFormIsSimple = .true. ! default.
    if (.not. (hasColumn(regionTable, 'SHAPE')&
    .and. hasColumn(regionTable, colPrefix//'X')&
    .and. hasColumn(regionTable, colPrefix//'Y')&
    .and. hasColumn(regionTable, 'R')&
    .and. hasColumn(regionTable, 'COMPONENT'))) then
      regionFormIsSimple = .false.
    end if
!print *, regionFormIsSimple

    if (regionFormIsSimple) then
      compCol = column(regionTable, 'COMPONENT', READ)
      comp => int8Data(compCol)
      do element = 2, numRegionRows
        if (comp(element - 1) .ne. comp(1 - 1)) then ! there is more than one
        ! component in the region filter. Could probably handle this
        ! situation but we'll schedule a dump out for the time being and
        ! see if it ever arises.

          regionFormIsSimple = .false.
      exit
        end if
      end do
      call release(compCol)
    end if
!print *, regionFormIsSimple

    if (regionFormIsSimple) then
      rCol = column(regionTable, 'R', READ)
      numDimensions = numberOfDimensions(rCol)
      if (numDimensions .gt. 2) then
        regionFormIsSimple = .false.
      end if
    end if
!print *, regionFormIsSimple

    if (.not. regionFormIsSimple) then
      write(errstr, *) "The region filter table in the DSS &
      &of dataset "//trim(inSetName)//" is too complicated to allow estimation &
      &of a small bounding box. A box will be calculated but unfortunately &
      &it must contain the whole field of view."

      call warning('dssToughRegion')
    end if
!call error('','')

!........................................................................
! Try to estimate a radius that contains the whole region:

    canDoCompactBox = .false. ! default.
    if (regionFormIsSimple) then
      rCol = column(regionTable, 'R', READ)
      shapeCol = column(regionTable, 'SHAPE', READ)
      allocate(shape(numRegionRows),&
      isCircOrAnn(numRegionRows), outerRadius(numRegionRows))

      do element = 1, numRegionRows
        shape(element) = stringCell(shapeCol, element - 1)
      end do

      atLeastOneCircOrAnn = .false. ! default
      isCircOrAnn = .false. ! vector op; default
      outerRadius = 0.0     ! vector op; default
      if (numDimensions .eq. 1) then
!print *, '1 dim'
        radius1dim => real32Data(rCol)
        do element = 1, numRegionRows
!print *, element, shape(element)
          if (index(trim(shape(element)), 'CIRCLE') .eq. 1) then
!print *, 'recognised circle'
            atLeastOneCircOrAnn = .true.
            outerRadius(element) = radius1dim(element - 1)
            isCircOrAnn(element) = .true.
          end if
        end do

      else ! numDimensions = 2
!print *, '2 dim'
        radius2dim => real32VectorData(rCol)
        do element = 1, numRegionRows
!print *, element, shape(element)
          if (index(trim(shape(element)), 'CIRCLE') .eq. 1) then
!print *, 'recognised circle'
            atLeastOneCircOrAnn = .true.
            outerRadius(element) = radius2dim(element - 1, 1)
            isCircOrAnn(element) = .true.
          else if (index(trim(shape(element)), 'ANNULUS') .eq. 1) then
            atLeastOneCircOrAnn = .true.
            outerRadius(element) = radius2dim(element - 1, 2)
            isCircOrAnn(element) = .true.
          end if
        end do
      end if ! numDimensions .eq. 1

      call release(shapeCol)
      call release(rCol)

      if (.not. atLeastOneCircOrAnn) then
        write(errstr, *) "The region filter table in the &
        &DSS of dataset "//trim(inSetName)//" contains no circles or annuli, &
        &therefore is too complicated to allow estimation of a small bounding &
        &box. A box will be calculated but unfortunately &
        &it must contain the whole field of view."

        call warning('RegionNoCircles')
      else
        do element = 1, numRegionRows
          if (isCircOrAnn(element)) then
            elementWithMinRad = element
            startElement = element + 1
        exit
          end if
        end do

        do element = startElement, numRegionRows
          if (outerRadius(element) .lt. minRadiusDet) then
            elementWithMinRad = element
          end if
        end do
        minRadiusDET = outerRadius(elementWithMinRad)
 
        canDoCompactBox = .true.
      end if
    end if
!print *, canDoCompactBox
!call error('','')

!........................................................................
! Calculate the box coordinates in both DET and CC2 systems (different only
! in their units):

    if (canDoCompactBox) then
      xCol = column(regionTable, colPrefix//'X', READ)
      yCol = column(regionTable, colPrefix//'Y', READ)
      elementDETX => real32Data(xCol)
      elementDETY => real32Data(yCol)
      boxMinDETX = elementDETX(elementWithMinRad - 1) - minRadiusDET
      boxMaxDETX = elementDETX(elementWithMinRad - 1) + minRadiusDET
      boxMinDETY = elementDETY(elementWithMinRad - 1) - minRadiusDET
      boxMaxDETY = elementDETY(elementWithMinRad - 1) + minRadiusDET
      boxMinCC2X = boxMinDETX / mmToDetUnits
      boxMaxCC2X = boxMaxDETX / mmToDetUnits
      boxMinCC2Y = boxMinDETY / mmToDetUnits
      boxMaxCC2Y = boxMaxDETY / mmToDetUnits

!      if (getPosFromRegn) then
!        srcCC2X = elementDETX(elementWithMinRad - 1) / mmToDetUnits
!        srcCC2Y = elementDETY(elementWithMinRad - 1) / mmToDetUnits
!      end if
    else
      boxMinDETX = - radiusFovDET
      boxMaxDETX = radiusFovDET
      boxMinDETY = - radiusFovDET
      boxMaxDETY = radiusFovDET
      boxMinCC2X = boxMinDETX / mmToDetUnits
      boxMaxCC2X = boxMaxDETX / mmToDetUnits
      boxMinCC2Y = boxMinDETY / mmToDetUnits
      boxMaxCC2Y = boxMaxDETY / mmToDetUnits

!      if (getPosFromRegn) then
!        write(errstr, *) "Can't get centre of source from &
!        &region. Uniform source model assumed."
!        call warning('noPointSrcPos')
!        imageModel = 'uniform'
!!***** set some values for srcCC2X,Y? Don't like to return undefined values.
!srcCC2X = 0.0
!srcCC2Y = 0.0
!      end if
    end if

    call release(inSet)

  end subroutine calculateBoundingBox

  !---------------------------------------------------------------------
  subroutine createPixelLists(&
      pixelListSet,&
      selectionExpression,&
!      boxMinCC2X,&
!      boxMinCC2Y,&
!      boxMaxCC2X,&
!      boxMaxCC2Y,&
      srcCC2X,&
      srcCC2Y,&
      imageModel,&
      ccdInRegion,&
      regionYExtentFrac,&
      numCcdsInRegion,&
      numPixels)
  !---------------------------------------------------------------------

    implicit none

!.................
! Arguments coming in:

    character(len=8), intent(in) ::&
      imageModel
    type(DataSetT), intent(in) ::&
      pixelListSet
    character(len=80), intent(in) ::&
      selectionExpression
!    real(kind=single), intent(in) ::&
!      boxMinCC2X,&
!      boxMinCC2Y,&
!      boxMaxCC2X,&
!      boxMaxCC2Y
    real(kind=single), intent(in) ::&
      srcCC2X,&
      srcCC2Y

!.................
! Arguments modified:

    logical, dimension(:,:), intent(in out) ::&
      ccdInRegion

!.................
! Arguments going out:

    real(kind=single), dimension(:,:,:), intent(out) ::&
      regionYExtentFrac
    integer(kind=int8), intent(out) ::&
      numCcdsInRegion
    integer(kind=int32), dimension(:,:), intent(out) ::&
      numPixels

!.................
! Others:

    type(DataSetT) ::&
      eventSet
    type(TableT) ::&
      pixelListTab,&
      badpixTab,&
      masterListTab
    type(ColumnT) ::&
      pixelListRawXcol,&
      pixelListRawYcol,&
      pixelListDETXcol,&
      pixelListDETYcol,&
      pixelListFlagCol,&
      badpixRawXcol,&
      badpixRawYcol,&
      badpixYextentcol,&
      masterListRawXcol,&
      masterListRawYcol,&
      masterListDETXcol,&
      masterListDETYcol,&
      masterListCcdnrcol
    integer ::&
      numBadpix,&
      badpix
    integer(kind=int8) ::&
      ccd,&
      node,&
      ccdnr
    integer(kind=int16) ::&
      nx,&
      ny,&
      badpixX,&
      badpixY,&
      colum
    integer(kind=int16), dimension(:), allocatable ::&
      regionYExtentPixels
    integer(kind=int32) ::&
      numRows,&
      event,&
      numEvents,&
      totalNumRows,&
      rowStart,&
      row,&
      masterRow
    integer(kind=int16), dimension(3) ::&
      dumx,&
      dumy
    integer(kind=int16) ::&
tempInteger
    integer(kind=int16), dimension(12,2) ::&
      boxMinRawX,&
      boxMinRawY,&
      boxMaxRawX,&
      boxMaxRawY
    character(8) ::&
      pixelListTabName,&
      badpixTabName
    real(kind=single), dimension(:), pointer ::&
      mmx,&
      mmy,&
      mmz,&
      cc2X,&
      cc2Y,&
      cc2Z,&
      detX,&
      detY,&
      masterListDetX,&
      masterListDetY,&
      pixelListDetX,&
      pixelListDetY
    real(kind=single) ::&
      mmx0,&
      aa,&
      bb,&
      mmy0,&
      cc,&
      dd,&
      det,&
      srcDETX,&
      srcDETY,&
      squareHalfSideDET,&
      psfMinDETX,&
      psfMaxDETX,&
      psfMinDETY,&
      psfMaxDETY
    integer(kind=int8), dimension(:), pointer ::&
      pixelListFlag,&
      masterListCcdnr
    integer(kind=int16), dimension(:), pointer ::&
      pixelListRawX,&
      pixelListRawY,&
      badpixRawX,&
      badpixRawY,&
      badpixYextent,&
      masterListRawX,&
      masterListRawY
    character(len=101) ::&
      psfSquareFilterStr
    integer(kind=int32) ::&
      coordSysType

!........................................................................
! Start:

    call message(AppMsg, Verbose, '  Making region evlist.')

    srcDETX = srcCC2X * mmToDetUnits
    srcDETY = srcCC2Y * mmToDetUnits
    squareHalfSideDET = (psfMapHalfsizePixels + 0.5) * psfMapPixelWidthArcsec&
    / detUnitArcsec
    psfMinDETX = srcDETX - squareHalfSideDET
    psfMaxDETX = srcDETX + squareHalfSideDET
    psfMinDETY = srcDETY - squareHalfSideDET
    psfMaxDETY = srcDETY + squareHalfSideDET
    write(psfSquareFilterStr,&
    '(A, E11.5, A, E11.5, A, E11.5, A, E11.5, A)')&
           '(DETX < ', psfMaxDETX, ')&
    & .and. (DETX > ', psfMinDETX, ')&
    & .and. (DETY < ', psfMaxDETY, ')&
    & .and. (DETY > ', psfMinDETY, ')'

    call message(AppMsg, Verbose, '    Region filtering.')
    do ccd = 1, numCcds
      do node = 1, numNodes
      if (.not. exposuExists(ccd, node)) cycle
!print *, '|||| ccd, node:', ccd, node

!........................................................................
!

        call CAL_setState(ccdChipId=int(ccd, int32))
!print *, 'set cal ccd'
        if (numNodes .gt. 1) call CAL_setState(ccdNodeId=int(node - 1, int32))
        ! cal returns a warning if an attempt is made to set ccdNodeId for pn.
!print *, 'set cal node'

        ! Convert the coordinates of the corners of the bounding box from
        ! the CAMCOORD2 to PIXCOORD1 (ie, RAWXY) system. This is done by
        ! working backwards: first a set of three dummy pixels is generated,
        ! then these are converted to CAMCOORD2; the transfer matrix and
        ! offsets are as follows:
        !
        !    /   \   /        \ /      \   /    \
        !    |mmx|   | aa  bb | |dummyx|   |mmx0|
        !    |   | = |        | |      | + |    |
        !    |mmy|   | cc  dd | |dummyy|   |mmy0|
        !    \   /   \        / \      /   \    /
        !    
        ! These numbers are computed by comparing the result with the starting
        ! coordinates.

!        dumx(1) = 0
!        dumx(2) = 1
!        dumx(3) = 1
!        dumy(1) = 0
!        dumy(2) = 1
!        dumy(3) = -1
!        coordSysType=CAMCOORD2
!        call CAL_rawXY2mm(dumx, dumy, mmx, mmy, mmz, coordSysType)
!        mmx0 = mmx(0)
!        aa = (mmx(1) + mmx(2)) / 2.0 - mmx0
!        bb = (mmx(1) - mmx(2)) / 2.0
!        mmy0 = mmy(0)
!        cc = (mmy(1) + mmy(2)) / 2.0 - mmy0
!        dd = (mmy(1) - mmy(2)) / 2.0
dumx(1) = 0
dumx(2) = 1
dumx(3) = 0
dumy(1) = 0
dumy(2) = 0
dumy(3) = 1
coordSysType=CAMCOORD2
call CAL_rawXY2mm(dumx, dumy, mmx, mmy, mmz, coordSysType)
mmx0 = mmx(0)
aa = mmx(1) - mmx0
bb = mmx(2) - mmx0
mmy0 = mmy(0)
cc = mmy(1) - mmy0
dd = mmy(2) - mmy0

!print *, mmx(0), mmx(1), mmx(2)
!print *, mmy(0), mmy(1), mmy(2)
!print *, aa, bb, cc, dd

        call CAL_releaseMemory(mmx, mmy, mmz)
        det = aa * dd - bb * cc

        boxMinRawX(ccd, node) =     floor((dd * (boxMinCC2X - mmx0)&
        - bb * (boxMinCC2Y - mmy0)) / det, int16)
        boxMaxRawX(ccd, node) = 1 + floor((dd * (boxMaxCC2X - mmx0)&
        - bb * (boxMaxCC2Y - mmy0)) / det, int16)
        boxMinRawY(ccd, node) =     floor((aa * (boxMinCC2Y - mmy0)&
        - cc * (boxMinCC2X - mmx0)) / det, int16)
        boxMaxRawY(ccd, node) = 1 + floor((aa * (boxMaxCC2Y - mmy0)&
        - cc * (boxMaxCC2X - mmx0)) / det, int16)

!print *, det
!print *, 'Very fresh RAWX/Y bounds of box:'
!print *, boxMinRawX(ccd, node), boxMaxRawX(ccd, node)
!print *, boxMinRawY(ccd, node), boxMaxRawY(ccd, node)
!print *, ' '
!print *, 'CC2X/Y bounds of box:'
!print *, boxMinCC2X, boxMaxCC2X
!print *, boxMinCC2Y, boxMaxCC2Y
!print *, ' '

	! Since there may be rotations of 180 deg, the max and min may be
        ! inverted. Correct for this:
	if (boxMinRawX(ccd, node) .gt. boxMaxRawX(ccd, node)) then
          tempInteger = boxMinRawX(ccd, node)
          boxMinRawX(ccd, node) = boxMaxRawX(ccd, node)
          boxMaxRawX(ccd, node) = tempInteger
	end if
	if (boxMinRawY(ccd, node) .gt. boxMaxRawY(ccd, node)) then
          tempInteger = boxMinRawY(ccd, node)
          boxMinRawY(ccd, node) = boxMaxRawY(ccd, node)
          boxMaxRawY(ccd, node) = tempInteger
	end if
!**** This is a clumsy way of doing it. Better to keep centre x,y + radius, convert these to pixcoord, then calc the initial min and max rawx and rawy.

        ! Truncate bounding box to edges of the data window on the CCD:

        boxMinRawX(ccd, node)&
        = max(boxMinRawX(ccd, node), windowMinRawX(ccd, node))
        boxMaxRawX(ccd, node)&
        = min(boxMaxRawX(ccd, node), windowMaxRawX(ccd, node))
        boxMinRawY(ccd, node)&
        = max(boxMinRawY(ccd, node), windowMinRawY(ccd, node))
        boxMaxRawY(ccd, node)&
        = min(boxMaxRawY(ccd, node), windowMaxRawY(ccd, node))

	if (boxMinRawX(ccd, node) .gt. boxMaxRawX(ccd, node)&
        .or. boxMinRawY(ccd, node) .gt. boxMaxRawY(ccd, node)) then
        ! then the region doesn't overlap this CCD/node at all. Forget it.
      cycle ! next node.
        end if

!print *, 'RAWX/Y bounds of box:'
!print *, boxMinRawX(ccd, node), boxMaxRawX(ccd, node)
!print *, boxMinRawY(ccd, node), boxMaxRawY(ccd, node)

        numRows = (1 + boxMaxRawX(ccd, node) - boxMinRawX(ccd, node))&
        * (1 + boxMaxRawY(ccd, node) - boxMinRawY(ccd, node))
        ! Implicit type conversion int16 -> int32
!print *, 'num rows =', numRows

        ccdnr = ccd2ccdnr(ccd, node)
        write(pixelListTabName, '(A, I2.2)') 'EVENTS', ccdnr
        pixelListTab = addTable(pixelListSet, pixelListTabName, numRows)
        pixelListRawXcol = addColumn(pixelListTab, 'RAWX', Integer16)
        pixelListRawYcol = addColumn(pixelListTab, 'RAWY', Integer16)
        pixelListDETXcol = addColumn(pixelListTab, 'DETX', Real32)
        pixelListDETYcol = addColumn(pixelListTab, 'DETY', Real32)
        pixelListRawX => int16data(pixelListRawXcol)
        pixelListRawY => int16data(pixelListRawYcol)
        detX => real32data(pixelListDETXcol)
        detY => real32data(pixelListDETYcol)

        event = 1
        do nx = boxMinRawX(ccd, node), boxMaxRawX(ccd, node)
          do ny = boxMinRawY(ccd, node), boxMaxRawY(ccd, node)
            pixelListRawX(event - 1) = nx
            pixelListRawY(event - 1) = ny
            event = event + 1
          end do ! next ny
        end do ! next nx

        coordSysType=CAMCOORD2
!print *, '>>'
        call CAL_rawXY2mm(pixelListRawX, pixelListRawY, cc2X, cc2Y, cc2Z&
        , coordSysType)
!print *, ccd, node, cc2X(0), cc2Y(0)
        detX = cc2X * mmToDetUnits
        detY = cc2Y * mmToDetUnits

!print *, 'a'
        call CAL_releaseMemory(cc2X, cc2Y, cc2Z)
!print *, 'b'
        call release(pixelListDETYcol)
        call release(pixelListDETXcol)
        call release(pixelListRawYcol)
        call release(pixelListRawXcol)
!print *, 'c'

        ! (iv) Now filter it with the region:
        call initFilter(selectionExpression)
!print *, 'DETX/Y bounds of box:'
!print *, boxMinCC2X * mmToDetUnits, boxMaxCC2X * mmToDetUnits
!print *, boxMinCC2Y * mmToDetUnits, boxMaxCC2Y * mmToDetUnits
!print *, ' '
!print *, 'Region sel expr:', selectionExpression
!print *, 'numEvents before region filter =', numEvents
        numEvents = doFilter(pixelListTab)
!print *, 'numEvents after region filter =', numEvents

        if (imageModel .eq. 'point') then ! filter by the square array that
        ! the PSF will be returned in:
          call initFilter(psfSquareFilterStr)
          numEvents = doFilter(pixelListTab)
!print *, psfSquareFilterStr
!print *, numEvents
        end if

!print *, 'a'
        call release(pixelListTab)
!print *, 'b'
      end do ! next node
!print *, 'c'
    end do ! next ccd
!print *, 'd'
!call error('', '')

!........................................................................
! Filter out bad pixels:

    ccdInRegion = .false. ! array op, default
!print *, 'e'
    allocate(regionYExtentPixels(chipMinRawX:chipMaxRawX))
!print *, 'f'
    eventSet = dataSet(eventSetName, READ)
    call message(AppMsg, Verbose, '    Filtering bad pixels.')
    do ccd = 1, numCcds
      do node = 1, numNodes
      if (.not. exposuExists(ccd, node)) cycle
!print *, 'ccd, node:', ccd, node
        regionYExtentPixels = 0 ! vector op; default.

        ccdnr = ccd2ccdnr(ccd, node)
        write(pixelListTabName, '(A, I2.2)') 'EVENTS', ccdnr
      if (.not. hasBlock(pixelListSet, pixelListTabName)) cycle

        pixelListTab = table(pixelListSet, pixelListTabName)
        numEvents = numberOfRows(pixelListTab)
        ! implicit type conversion to int32. Normally I have set the
        ! variable that contains the return from numberOfRows() as
        ! integer, no kind defined, but my guess is that this is int16
        ! and a whole-camera 1-event-per-pixel event list for MOS has
        ! over 2 million rows, quite a bit bigger than ~32000!

      if (numEvents .le. 0) cycle

        write(badpixTabName, '(A, I2.2)') 'BADPIX', ccdnr
        if (hasBlock(eventSet, badpixTabName)) then
          pixelListFlagCol = addColumn(pixelListTab, 'FLAG', Integer8)
          pixelListFlag => int8data(pixelListFlagCol)
          pixelListFlag = 0 ! vector op; default

          pixelListRawXcol = column(pixelListTab, 'RAWX')
          pixelListRawYcol = column(pixelListTab, 'RAWY')
          pixelListRawX => int16data(pixelListRawXcol)
          pixelListRawY => int16data(pixelListRawYcol)

          badpixTab = table(eventSet, badpixTabName)

          badpixRawXcol    = column(badpixTab, 'RAWX')
          badpixRawYcol    = column(badpixTab, 'RAWY')
          badpixYextentcol = column(badpixTab, 'YEXTENT')
          badpixRawX    => int16data(badpixRawXcol)
          badpixRawY    => int16data(badpixRawYcol)
          badpixYextent => int16data(badpixYextentcol)

          numBadpix = numberOfRows(badpixTab)
          do badpix = 1, numBadpix
            badpixX = badpixRawX(badpix - 1)
            do ny = 1, badpixYextent(badpix - 1)
              badpixY = (badpixRawY(badpix - 1) + ny - 1)

              if (badpixX .ge. boxMinRawX(ccd, node)&
              .and. badpixX .le. boxMaxRawX(ccd, node)&
              .and. badpixY .ge. boxMinRawY(ccd, node)&
              .and. badpixY .le. boxMaxRawY(ccd, node)) then
                do event = 1, numEvents
                  if (pixelListRawX(event - 1) .eq. badpixX&
                  .and. pixelListRawY(event - 1) .eq. badpixY) then
                    pixelListFlag(event - 1) = 1
                  end if
                end do ! next event
              end if ! badpix in bounding box
            end do ! next ny
          end do ! next badpix

          call release(badpixTab)
          call release(pixelListRawYcol)
          call release(pixelListRawXcol)
          call release(pixelListFlagCol)

          call initFilter('FLAG .eq. 0')
          numEvents = doFilter(pixelListTab)

        end if ! there is a BADPIXnn table in the event list.

        numPixels(ccd, node) = numEvents

        if (numEvents .gt. 0) then
          ccdInRegion(ccd, node) = .true.

          pixelListRawXcol = column(pixelListTab, 'RAWX')
          pixelListRawX => int16data(pixelListRawXcol)
          do event = 1, numEvents
            regionYExtentPixels(pixelListRawX(event - 1))&
            = regionYExtentPixels(pixelListRawX(event - 1)) + 1
          end do ! next event
          call release(pixelListRawXcol)
          do colum = windowMinRawX(ccd, node), windowMaxRawX(ccd, node)
            regionYExtentFrac(ccd, node, colum)&
            = regionYExtentPixels(colum) / real(windowRawDY(ccd, node))
          end do ! colum misspelt because of conflict with DAL symbol.
        end if

        call deleteColumn(pixelListTab, 'FLAG')
        call release(pixelListTab)
      end do ! next node
    end do ! next ccd
    call release(eventSet)
    deallocate(regionYExtentPixels)

! Should really look at the event flag filtering in the dss and screen out
! pixels next to bad pixels and ccd/node edges if appropriate.

    numCcdsInRegion = 0 ! default
!print *, 'THIRD LOOP:'
    do ccd = 1, numCcds
      do node = 1, numNodes
        if (ccdInRegion(ccd, node)) then
!print *, 'ccd, node:', ccd, node
          numCcdsInRegion = numCcdsInRegion + 1
        end if
      end do ! next node
    end do ! next ccd
    if (numCcdsInRegion .le. 0) then
call error('noCcdsInRegion', "The accumulation region does not overlap &
&any CCD.")
    end if

!........................................................................

    ! Now make a master table in pixelListSet containing _all_ the events,
    ! with an additional column CCDNR. This table is for the routine
    ! calcSpatial10a(). If f90 had such things as hashes, or arrays of
    ! pointers, I wouldn't have had to do this. But can't think of any
    ! other moderately efficient way of doing it.

    totalNumRows = 0
    call message(AppMsg, Verbose, '    Making master table.')
    do ccd = 1, numCcds
      do node = 1, numNodes
      if (.not. ccdInRegion(ccd, node)) cycle
!print *, 'ccd, node:', ccd, node
        ccdnr = ccd2ccdnr(ccd, node)
        write(pixelListTabName, '(A, I2.2)') 'EVENTS', ccdnr
        pixelListTab = table(pixelListSet, pixelListTabName)

        totalNumRows = totalNumRows + numberOfRows(pixelListTab)
        ! implicit type conversion to int32. Normally I have set the
        ! variable that contains the return from numberOfRows() as
        ! integer, no kind defined, but my guess is that this is int16
        ! and a whole-camera 1-event-per-pixel event list for MOS has
        ! over 2 million rows, quite a bit bigger than ~32000!

      end do ! next node
    end do ! next ccd

    masterListTab = addTable(pixelListSet, 'MASTER', totalNumRows)
    masterListRawXcol  = addColumn(masterListTab, 'RAWX', Integer16)
    masterListRawYcol  = addColumn(masterListTab, 'RAWY', Integer16)
    masterListDETXcol  = addColumn(masterListTab, 'DETX', Real32)
    masterListDETYcol  = addColumn(masterListTab, 'DETY', Real32)
    masterListCcdnrcol = addColumn(masterListTab, 'CCDNR', Integer8)
    masterListRawX  => int16data(masterListRawXcol)
    masterListRawY  => int16data(masterListRawYcol)
    masterListDetX  => real32data(masterListDETXcol)
    masterListDetY  => real32data(masterListDETYcol)
    masterListCcdnr => int8data(masterListCcdnrcol)

    rowStart = 0
!print *, 'FIFTH LOOP:'
    do ccd = 1, numCcds
      do node = 1, numNodes
      if (.not. ccdInRegion(ccd, node)) cycle
!print *, 'ccd, node:', ccd, node
        ccdnr = ccd2ccdnr(ccd, node)
        write(pixelListTabName, '(A, I2.2)') 'EVENTS', ccdnr

        pixelListTab = table(pixelListSet, pixelListTabName)
        pixelListRawXcol = column(pixelListTab, 'RAWX')
        pixelListRawYcol = column(pixelListTab, 'RAWY')
        pixelListDETXcol = column(pixelListTab, 'DETX')
        pixelListDETYcol = column(pixelListTab, 'DETY')
        pixelListRawX => int16data(pixelListRawXcol)
        pixelListRawY => int16data(pixelListRawYcol)
        pixelListDetX => real32data(pixelListDETXcol)
        pixelListDetY => real32data(pixelListDETYcol)

        numRows = numberOfRows(pixelListTab)
        ! implicit type conversion to int32. Normally I have set the
        ! variable that contains the return from numberOfRows() as
        ! integer, no kind defined, but my guess is that this is int16
        ! and a whole-camera 1-event-per-pixel event list for MOS has
        ! over 2 million rows, quite a bit bigger than ~32000!

        do row = 1, numRows
          masterRow = row + rowStart

          masterListRawX(masterRow - 1) = pixelListRawX(row - 1)
          masterListRawY(masterRow - 1) = pixelListRawY(row - 1)
          masterListDetX(masterRow - 1) = pixelListDetX(row - 1)
          masterListDetY(masterRow - 1) = pixelListDetY(row - 1)
          masterListCcdnr(masterRow - 1) = ccdnr
        end do

        rowStart = rowStart + numRows
        call release(pixelListTab)
      end do ! next node
    end do ! next ccd

boxMinDETX = minval(masterListDetX)
boxMaxDETX = maxval(masterListDetX)
boxMinDETY = minval(masterListDetY)
boxMaxDETY = maxval(masterListDetY)

boxMinCC2X = boxMinDETX / mmToDetUnits
boxMaxCC2X = boxMaxDETX / mmToDetUnits
boxMinCC2Y = boxMinDETY / mmToDetUnits
boxMaxCC2Y = boxMaxDETY / mmToDetUnits

    call release(masterListTab)

  end subroutine createPixelLists

  !---------------------------------------------------------------------
  subroutine calculateRegionArea(&
!    boxMinDETX,&
!    boxMinDETY,&
!    boxMaxDETX,&
!    boxMaxDETY,&
    selectionExpression,&
    regionAreaMm2)
  !---------------------------------------------------------------------

! Calculate the total area of the region. This is done by making an event
! list, one event per 'pixel' in a square grid, and filtering this event
! list with the DSS region. The coordinates of the events don't have to be
! related to the real camera and can thus be generated in the DET system
! from the start. The size of the 'pixels' is arbitrary but they should
! probably be about the same size as the real CCD pixels - much smaller
! and the size of the event list could grow too large; much larger and the
! precision of the answer for small regions could drop too low.

    implicit none

!.................
! Arguments coming in:

!    real(kind=single), intent(in) ::&
!      boxMinDETX,&
!      boxMinDETY,&
!      boxMaxDETX,&
!      boxMaxDETY
    character(len=80), intent(in) ::&
      selectionExpression

!.................
! Arguments going out:

    real(kind=single), intent(out) ::&
      regionAreaMm2

!.................
! Others:

    type(DataSetT) ::&
      pixelListSet
    type(TableT) ::&
      totalTab
    type(ColumnT) ::&
      detXcol,&
      detYcol
    real(kind=single) ::&
      squareLengthMm,&
      squareLengthDET,&
      xVal,&
      yVal
    real(kind=single), dimension(:), pointer ::&
      detX,&
      detY
    integer(kind=int16) ::&
      numXvals,&
      numYvals,&
      nx,&
      ny
    integer(kind=int32) ::&
      numRows,&
      event

!........................................................................
! Start:

    call message(AppMsg, Verbose, '  Calculating total area of region.')

    squareLengthMm = 0.04
    squareLengthDET = squareLengthMm * mmToDetUnits

    numXvals = 1 + int((boxMaxDETX - boxMinDETX) / squareLengthDET, int16)
    numYvals = 1 + int((boxMaxDETY - boxMinDETY) / squareLengthDET, int16)

    numRows = numXvals * numYvals ! type conversion to int32.

    pixelListSet = dataSet('tempeventset.fits', TEMP)
    totalTab = addTable(pixelListSet, 'TOTAL', numRows)
    detXcol = addColumn(totalTab, 'DETX', Real32)
    detYcol = addColumn(totalTab, 'DETY', Real32)
    detX => real32data(detXcol)
    detY => real32data(detYcol)

    event = 1
    do nx = 0, numXvals - 1
      xVal = boxMinDETX + (nx / real(numXvals - 1, single))&
      * (boxMaxDETX - boxMinDETX)
      do ny = 0, numYvals - 1
        yVal = boxMinDETY + (ny / real(numYvals - 1, single))&
        * (boxMaxDETY - boxMinDETY)
        detX(event - 1) = xVal
        detY(event - 1) = yVal
        event = event + 1
      end do
    end do

    call release(detYcol)
    call release(detXcol)

    call initFilter(selectionExpression)

    regionAreaMm2 = doFilter(totalTab) * squareLengthMm * squareLengthMm
    ! implicit type conversion integer -> single

    call release(pixelListSet)

  end subroutine calculateRegionArea

  !---------------------------------------------------------------------
  subroutine checkCalGetEB
  !---------------------------------------------------------------------

    implicit none 

    real(kind=double), dimension(:), pointer ::&
      calEmins,&
      calEmaxs
    integer(int16), dimension(:), pointer ::&
      EBchannels
    integer(int16) ::&
      numEBchannels,&
      EBstartChannel,&
      i

    !-------------------------------------------------------------------
    ! Start:

    call CAL_getEbounds(EBchannels, calEmins, calEmaxs)
    numEBchannels = size(EBchannels)

!do i = 0, numEBchannels - 1
!print *, EBchannels(i), calEmins(i), calEmaxs(i)
!end do

    EBstartChannel = EBchannels(0)
    do i = 2, numEBchannels
      if (EBchannels(i - 1) .ne. EBstartChannel - 1 + i) then
        write(errstr, *) "Non-consecutive array of channels &
        &returned by CAL_getEbounds."

call error('EBbadChanOrder')
    exit
      end if
    end do ! next i

    if (calEmins(0) .gt. andedEmin) then
      write(errstr, *) "The minimum energy returned by CAL_getEbounds &
      &is greater than the minimum energy of the source timeseries."
      call warning('tsBadEmin')
      andedEmin = calEmins(0)
    end if

    if (calEmaxs(numEBchannels - 1) .lt. andedEmax) then
      write(errstr, *) "The maximum energy returned by CAL_getEbounds &
      &is less than the maximum energy of the source timeseries."
      call warning('tsBadEmax')
      andedEmax = calEmaxs(numEBchannels - 1)
    end if

    call CAL_releaseMemory(EBchannels)
    call CAL_releaseMemory(calEmins)
    call CAL_releaseMemory(calEmaxs)

  end subroutine checkCalGetEB

  !---------------------------------------------------------------------
  subroutine checkSpectrum(specSetName, firstChan, lastChan)
  !---------------------------------------------------------------------

    implicit none 

    character(*), intent(in) ::&
      specSetName
    integer(int16), intent(out) ::&
      firstChan,&
      lastChan

    !.................
    ! Others:

    type(DataSetT) ::&
      specSet
    type(TableT) ::&
      specTab
    type(ColumnT) ::&
      specChannelCol
    integer ::&
      numSpecRows
    character(len=20) ::&
      specChanType
    real(kind=double), dimension(:), pointer ::&
      calEmins,&
      calEmaxs
    integer(int16), dimension(:), pointer ::&
      specChannels,&
      EBchannels
    integer(int16) ::&
      numEBchannels,&
      specFirstChan,&
      specLastChan,&
      EBstartChannel,&
      EBstartIndex,&
      EBstopIndex,&
!      i,&
      index

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, '  Checking spectrum.')

    specSet = dataSet(specSetName, READ)
    specTab = table(specSet, 'SPECTRUM')
    numSpecRows = numberOfRows(specTab)

    if (numSpecRows .lt. 1) then
call error('specEmpty', 'There is no data in the spectrum '&
//trim(specSetName)//'.')
    end if

    specChanType = stringAttribute(specTab, 'CHANTYPE')
    if (trim(specChanType) .ne. 'PI') then
call error('specChanTypeNotPi', 'The channel type of the spectrum '&
//trim(specSetName)//' is not PI.')
    end if

    specChannelCol = column(specTab, 'CHANNEL')
    specChannels => int16data(specChannelCol)
    specFirstChan = specChannels(0)
    specLastChan  = specChannels(numSpecRows - 1)

    call release(specSet)

    call CAL_getEbounds(EBchannels, calEmins, calEmaxs)
    numEBchannels = size(EBchannels)
    EBstartChannel = EBchannels(0)

!........................................................................
! Check the energy bounds of the spectrum:

    firstChan = max(specFirstChan, EBchannels(0))
    lastChan  = min(specLastChan,  EBchannels(numEBchannels - 1))

!print *, tempSpecFirstChan, tempSpecLastChan, EBchannels(0), size(EBchannels), calEmaxs(size(EBchannels)-1)
!print *, Emin(EBstartIndex), srcTsEmin
!print *, Emax(EBstopIndex), srcTsEmax
    EBstartIndex = firstChan - EBchannels(0)
    EBstopIndex  = lastChan  - EBchannels(0)
!    if (calEmins(EBstartIndex) .gt. srcTsEmin&
!    .or. calEmaxs(EBstopIndex) .lt. srcTsEmax) then
!      write(errstr, *) "The energy range in "//trim(specSetName)//&
!      " does not entirely overlap that of the source timeseries."
!
!!      call CAL_releaseMemory(EBchannels)
!!      call CAL_releaseMemory(calEmins)
!!      call CAL_releaseMemory(calEmaxs)
!!print *, EBstartIndex, EBstopIndex
!!print *, calEmins(EBstartIndex), calEmaxs(EBstopIndex)
!!print *, srcTsEmin, srcTsEmax
!call error('specBadErange')
!    end if

    if (calEmins(EBstartIndex) .gt. andedEmin) then
      write(errstr, *) "The minimum energy of "//trim(specSetName)//&
      " is greater than that of the source timeseries."

      call warning('specBadEmin')
      andedEmin = calEmins(EBstartIndex)
    end if

    if (calEmaxs(EBstopIndex) .lt. andedEmax) then
      write(errstr, *) "The maximum energy of "//trim(specSetName)//&
      " is less than that of the source timeseries."

      call warning('specBadEmax')
      andedEmax = calEmaxs(EBstopIndex)
    end if

    ! The averages are used below because these are the values that will appear in column BINCTREV of the internal spectrum. The energy range of the coarse cal arrays is from andedEmin to andedEmax; when the spatial exposure corrections are calculated later on, the energy integral is approximated by a sum over a series of energy samples, these samples comprising the BINCTREV vector: thus the possibility of having an energy value outside the range of the coarse cal arrays is avoided. Matters are even safer if a model spectrum is employed or if the measured spectrum is rebinned, because in both cases the energy boundaries of the spectrum (smallest energy of the first channel to the largest energy of the last) are set exactly to andedEmin to andedEmax.
    do index = 0, size(EBchannels) - 1
      if ((calEmins(index) + calEmaxs(index)) / 2.0 .ge. andedEmin) then
        firstChan = EBchannels(index)
    exit
      end if
    end do
    do index = size(EBchannels) - 1, 0, -1
      if ((calEmins(index) + calEmaxs(index)) / 2.0 .le. andedEmax) then
        lastChan = EBchannels(index)
    exit
      end if
    end do

    call CAL_releaseMemory(EBchannels)
    call CAL_releaseMemory(calEmins)
    call CAL_releaseMemory(calEmaxs)

  end subroutine checkSpectrum

  !---------------------------------------------------------------------
  subroutine loadInternalSpectrum(specSetName, firstChan, lastChan, internalSpecSet)
  !---------------------------------------------------------------------

    implicit none 

    character(*), intent(in) ::&
      specSetName
    integer(int16), intent(in) ::&
      firstChan,&
      lastChan
    type(DataSetT), intent(out) ::&
      internalSpecSet

    integer(int16) ::&
      numRows,&
      row,&
      index
    type(DataSetT) ::&
!      internalSpecSet,&
      specSet
    type(TableT) ::&
      internalSpecTab,&
      specTab
    type(ColumnT) ::&
      internalSpecBCTReVcol,&
      internalSpecBWIDeVcol,&
      internalSpecBCTRchanCol,&
      internalSpecBWIDchanCol,&
      internalSpecRateCol,&
      specChannelCol,&
      specCountsCol,&
      specRateCol
    real(single), dimension(:), pointer ::&
      internalSpecBCTReV,&
      internalSpecBWIDeV,&
      internalSpecBCTRchan,&
      internalSpecBWIDchan,&
      internalSpecRate,&
      specRate
    integer(int16), dimension(:), pointer ::&
      specChannel,&
      EbChannels
    integer(int32), dimension(:), pointer ::&
      specCounts
    real(kind=double), dimension(:), pointer ::&
      EbEmins,&
      EbEmaxs

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, '  Loading measured spectrum.')

!........................................................................
! Make the temporary internal spectrum file:

    call CAL_getEbounds(EbChannels, EbEmins, EbEmaxs)

    internalSpecSet = dataSet('tempspecset.fits', TEMP)
    numRows = 1 + lastChan - firstChan
    internalSpecTab = addTable(internalSpecSet, 'SPECTRUM', int(numRows))

    ! Write the BINCTREV column of the temp spec file:
    internalSpecBCTReVcol = addColumn(internalSpecTab, 'BINCTREV', Real32)
    internalSpecBCTReV => real32Data(internalSpecBCTReVcol)
    do row = 0, numRows - 1
      index = row + firstChan - EbChannels(0)
      internalSpecBCTReV(row) = (EbEmins(index) + EbEmaxs(index)) / 2.0
    end do ! next row
    call release(internalSpecBCTReVcol)

    ! Write the BINWIDEV column of the temp spec file:
    internalSpecBWIDeVcol = addColumn(internalSpecTab, 'BINWIDEV', Real32)
    internalSpecBWIDeV => real32Data(internalSpecBWIDeVcol)
    do row = 0, numRows - 1
      index = row + firstChan - EbChannels(0)
      internalSpecBWIDeV(row) = EbEmaxs(index) - EbEmins(index)
    end do ! next row
    call release(internalSpecBWIDeVcol)

    ! Write the BCTRCHAN column of the temp spec file:
    internalSpecBCTRchanCol = addColumn(internalSpecTab, 'BCTRCHAN', Real32)
    internalSpecBCTRchan => real32Data(internalSpecBCTRchanCol)
    do row = 0, numRows - 1
      internalSpecBCTRchan(row) = real(firstChan + row, single)
    end do ! next row
    call release(internalSpecBCTRchanCol)

    ! Write the BWIDCHAN column of the temp spec file:
    internalSpecBWIDchanCol = addColumn(internalSpecTab, 'BWIDCHAN', Real32)
    internalSpecBWIDchan => real32Data(internalSpecBWIDchanCol)
    internalSpecBWIDchan = 1.0 ! vector op
    call release(internalSpecBWIDchanCol)

    specSet = dataSet(specSetName, READ)
    specTab = table(specSet, 'SPECTRUM')
    specChannelCol = column(specTab, 'CHANNEL')
    specChannel => int16Data(specChannelCol)

    ! Write the RATE column of the temp spec file:
    internalSpecRateCol = addColumn(internalSpecTab, 'RATE', Real32)
    internalSpecRate => real32Data(internalSpecRateCol)
    if (hasColumn(specTab, 'COUNTS')) then
      specCountsCol = column(specTab, 'COUNTS')
      specCounts => int32Data(specCountsCol)
      do row = 0, numRows - 1
        internalSpecRate(row)&
        = real(specCounts(row + firstChan - specChannel(0)), single)
      end do ! next row

    else ! RATE
      specRateCol = column(specTab, 'RATE')
      specRate => real32Data(specRateCol)
      do row = 0, numRows - 1
        internalSpecRate(row) = specRate(row + firstChan - specChannel(0))
      end do ! next row
    end if

    call release(internalSpecTab)
    call release(specTab)
    call CAL_releaseMemory(EbChannels)
    call CAL_releaseMemory(EbEmins)
    call CAL_releaseMemory(EbEmaxs)

  end subroutine loadInternalSpectrum

  !---------------------------------------------------------------------
  subroutine rebinSpectrum(specSetName, mainSpecSet)
  !---------------------------------------------------------------------
  ! mainSpecSet should have the following columns:
  !
  !	BCTRCHAN	- the bin centre in PI channels
  !	BWIDCHAN	- the bin width in PI channels
  !	BINCTREV	- the bin centre in eV
  !	BINWIDEV	- the bin width in eV.
  !	RATE		- the count rate
  !	EFFAREA0	- the effective area
  !
  ! The following routine regards the channels of the input spectrum
  ! as bins over a real-valued coordinate p, such that channel C
  ! extends from p = C - 0.5 to C + 0.5. In the output (ie, rebinned)
  ! spectrum the 'channels' may no longer have these pseudo-integer
  ! values.

    implicit none

    character(*), intent(in) ::&
      specSetName
    type(DataSetT), intent(inout) ::&
      mainSpecSet

    real(single) ::&
      frac, outEeVstart,&
      outEeVfinis,&
      outChannelStart,&
      outChannelFinis
    integer(int16) ::&
      numMainSpecRows,&
      row,&
      outRow
    type(TableT) ::&
      mainSpecTab
    type(ColumnT) ::&
      mainSpecBCTReVcol,&
      mainSpecBWIDeVcol,&
      mainSpecBCTRchanCol,&
      mainSpecBWIDchanCol,&
      mainSpecRateCol,&
      mainSpecEACol
    real(single), dimension(:), pointer ::&
      mainSpecBCTReV,&
      mainSpecBWIDeV,&
      mainSpecBCTRchan,&
      mainSpecBWIDchan,&
      mainSpecRate,&
      mainSpecEA
    real(single), dimension(0:numRBspecRows - 1) ::&
      outSpecBCTReV,&
      outSpecBWIDeV,&
      outSpecBCTRchan,&
      outSpecBWIDchan,&
      outSpecRate,&
      outSpecEA

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, '  Rebinning spectrum.')
!srcTsEmin
! At this point the internal spectrum table should have andedEmin larger than the average energy of the first row and less than or equal to the average energy of the second row, and andedEmax less than the average energy of the last row and greater than or equal to the average energy of the second-last row. It is rebinned such that the least energy of the first row of the rebinned spectrum is aligned or equal to andedEmin and the maximum energy of the last row of the rebinned spectrum is aligned with or equal to andedEmax.

!........................................................................
! Prepare for rebinning the measured spectrum:

    mainSpecTab = table(mainSpecSet, 'SPECTRUM')
    numMainSpecRows = numberOfRows(mainSpecTab)

    if (trim(rebinSpecMode) .eq. 'tonumber'&
    .and. numRBspecRows .ge. numMainSpecRows) then
      write(errstr, *) "Number of spectrum bins requested &
      &is greater than the number of channels in "//trim(specSetName)&
      //". Rebinning not done."

      call warning('tooManySpecBins')
  return
    end if

    if (trim(rebinSpecMode) .eq. 'divideby') then
      numRBspecRows = int(numMainSpecRows / specBinDiv, int16) ! The parameter
      ! limits should prevent specBinDiv being less than 1, and its
      ! existence should have been tested in the calling routine.

      if (numRBspecRows .ge. numMainSpecRows) then ! equality can happen
      ! if numMainSpecRows is large and specBinDiv is just a little greater
      ! than 1. 'Greater than' shouldn't be able to happen, but one never knows.

  return ! no need to rebin spectrum.
      end if
    end if

    if (numRBspecRows .lt. leastNSpecBins) then
      numRBspecRows = leastNSpecBins
      write(errstr, *) "Requested number of spectrum bins &
      &is smaller than the minimum value of", leastNSpecBins, ". This &
      &minimum number has been used instead."

      call warning('tooFewSpecBins')
    end if

!........................................................................
! Now rebin the measured spectrum:

    mainSpecBCTReVcol = column(mainSpecTab, 'BINCTREV')
    mainSpecBCTReV => real32Data(mainSpecBCTReVcol)

    mainSpecBWIDeVcol = column(mainSpecTab, 'BINWIDEV')
    mainSpecBWIDeV => real32Data(mainSpecBWIDeVcol)

    mainSpecBCTRchanCol = column(mainSpecTab, 'BCTRCHAN')
    mainSpecBCTRchan => real32Data(mainSpecBCTRchanCol)

!    mainSpecBWIDchanCol = column(mainSpecTab, 'BWIDCHAN')
!    mainSpecBWIDchan => real32Data(mainSpecBWIDchanCol)

    mainSpecRateCol = column(mainSpecTab, 'RATE')
    mainSpecRate => real32Data(mainSpecRateCol)

    mainSpecEACol = column(mainSpecTab, 'EFFAREA0')
    mainSpecEA => real32Data(mainSpecEACol)

    ! First divide up the energy range among the new bins:
!    outSpecBWIDeV = (srcTsEmax - srcTsEmin) / numRBspecRows ! vector op
    outSpecBWIDeV = (andedEmax - andedEmin) / numRBspecRows ! vector op
    do row = 0, numRBspecRows - 1
      frac = (row + 0.5) / real(numRBspecRows)
!      outSpecBCTReV(row) = srcTsEmin * (1.0 - frac) + srcTsEmax * frac
      outSpecBCTReV(row) = andedEmin * (1.0 - frac) + andedEmax * frac
    end do
!***** later, might want to divide up the energy range logarithmically.

    ! A couple of checks which ought to be passed ok if the earlier code is doing what it should:
    outEeVstart = outSpecBCTReV(0) - outSpecBWIDeV(0) / 2.0
    if (outEeVstart .lt. (mainSpecBCTReV(0) - mainSpecBWIDeV(0) / 2.0)&
    .or. outEeVstart .gt. (mainSpecBCTReV(numMainSpecRows - 1)&
    + mainSpecBWIDeV(numMainSpecRows - 1) / 2.0)) then
call error('', '') ! shouldn't be able to happen with the code the way it is.
    end if
    outEeVfinis = outSpecBCTReV(numRBspecRows - 1) + outSpecBWIDeV(numRBspecRows - 1) / 2.0
    if (outEeVfinis .lt. (mainSpecBCTReV(0) - mainSpecBWIDeV(0) / 2.0)&
    .or. outEeVfinis .gt. (mainSpecBCTReV(numMainSpecRows - 1)&
    + mainSpecBWIDeV(numMainSpecRows - 1) / 2.0)) then
call error('', '') ! shouldn't be able to happen with the code the way it is.
    end if

    ! Now interpolate to get the channel info:
!    outChannelStart = getChannelValue(outEeVstart, mainSpecBCTRchan, mainSpecBCTReV, mainSpecBWIDeV, numMainSpecRows)
    outChannelStart = interpolate(outEeVstart, mainSpecBCTReV, mainSpecBWIDeV&
    , mainSpecBCTRchan, mainSpecBWIDchan, numMainSpecRows)

    do outRow = 0, numRBspecRows - 1
      outEeVfinis = outSpecBCTReV(outRow) + outSpecBWIDeV(outRow) / 2.0
!      outChannelFinis = getChannelValue(outEeVfinis, mainSpecBCTRchan, mainSpecBCTReV, mainSpecBWIDeV, numMainSpecRows)
      outChannelFinis = interpolate(outEeVfinis, mainSpecBCTReV&
      , mainSpecBWIDeV, mainSpecBCTRchan, mainSpecBWIDchan, numMainSpecRows)

      outSpecBCTRchan(outRow) = (outChannelFinis + outChannelStart) / 2.0
      outSpecBWIDchan(outRow) = outChannelFinis - outChannelStart
      outChannelStart = outChannelFinis
    end do

    ! Actual rebinning:
    call rebin(mainSpecBCTReV, mainSpecBWIDeV, mainSpecRate, numMainSpecRows&
    , outSpecBCTReV, outSpecBWIDeV, outSpecRate, numRBspecRows)
    call rebin(mainSpecBCTReV, mainSpecBWIDeV, mainSpecEA,   numMainSpecRows&
    , outSpecBCTReV, outSpecBWIDeV, outSpecEA,   numRBspecRows)

    ! Destroy main spectrum:
    call release(mainSpecSet)

    ! Now recreate it and load in the vectors of outSpec:

    mainSpecSet = dataSet('tempspecset.fits', TEMP)
    mainSpecTab = addTable(mainSpecSet, 'SPECTRUM', int(numRBspecRows))

    mainSpecBCTReVcol = addColumn(mainSpecTab, 'BINCTREV', Real32)
    mainSpecBCTReV => real32Data(mainSpecBCTReVcol)
    mainSpecBCTReV = outSpecBCTReV
    call release(mainSpecBCTReVcol)

    mainSpecBWIDeVcol = addColumn(mainSpecTab, 'BINWIDEV', Real32)
    mainSpecBWIDeV => real32Data(mainSpecBWIDeVcol)
    mainSpecBWIDeV = outSpecBWIDeV
    call release(mainSpecBWIDeVcol)

    mainSpecBCTRchanCol = addColumn(mainSpecTab, 'BCTRCHAN', Real32)
    mainSpecBCTRchan => real32Data(mainSpecBCTRchanCol)
    mainSpecBCTRchan = outSpecBCTRchan
    call release(mainSpecBCTRchanCol)

    mainSpecBWIDchanCol = addColumn(mainSpecTab, 'BWIDCHAN', Real32)
    mainSpecBWIDchan => real32Data(mainSpecBWIDchanCol)
    mainSpecBWIDchan = outSpecBWIDchan
    call release(mainSpecBWIDchanCol)

    mainSpecRateCol = addColumn(mainSpecTab, 'RATE', Real32)
    mainSpecRate => real32Data(mainSpecRateCol)
    mainSpecRate = outSpecRate
    call release(mainSpecRateCol)

    mainSpecEACol = addColumn(mainSpecTab, 'EFFAREA0', Real32)
    mainSpecEA => real32Data(mainSpecEACol)
    mainSpecEA = outSpecEA
    call release(mainSpecEACol)

    call release(mainSpecTab)

  end subroutine rebinSpectrum

  !---------------------------------------------------------------------
  function interpolate(x, xCentres, xWidths, yCentres, yWidths, numBins)&
  result(y)
  !---------------------------------------------------------------------
  ! Assumes spec channel widths are equal to 1.0.

    implicit none 

    integer(int16), intent(in) ::&
      numBins
    real(single), dimension(numBins), intent(in) ::&
      xCentres, xWidths, yCentres, yWidths
    real(single), intent(in) ::&
      x

    real(single) ::&
      y, xEdges(0:numBins)
    integer(int16) ::&
      bin,&
      foundBin
!    logical ::&
!      binFound

    !-------------------------------------------------------------------
    ! Start:

    ! set up xEdges (no ambiguity with this form of deliniation of the
    ! x bin edges):
    xEdges(0) = xCentres(1) - xWidths(1) / 2.0
    do bin = 1, numBins
      xEdges(bin) = xCentres(bin) + xWidths(bin) / 2.0
    end do
!**** more efficient to do this outside the function call.

    if (x .lt. xEdges(0) .or. x .gt. xEdges(numBins)) then
call error('', '')
    end if

    ! Identify the bin that x occurs in:
    do bin = 1, numBins
      if (x .le. xEdges(bin)) then
        foundBin = bin
    exit
      end if
    end do

!**** again, yEdges would be more efficient.
    y = (yCentres(foundBin) - yWidths(foundBin) / 2.0)&
    + yWidths(foundBin) * (x - xEdges(foundBin - 1)) / xWidths(foundBin)

  end function interpolate

  !---------------------------------------------------------------------
  subroutine rebin(oldXcentres, oldXwidths, oldYvalues, numOldBins&
  , newXcentres, newXwidths, newYvalues, numNewBins)
  !---------------------------------------------------------------------
  ! This is designed to be a pretty general and robust rebinning routine.
  ! Neither the old or new series is constrained to have bins of equal width.

    implicit none 

    integer(int16), intent(in) ::&
      numOldBins,&
      numNewBins
    real(single), dimension(numOldBins), intent(in) ::&
      oldXcentres, oldXwidths, oldYvalues
    real(single), dimension(numNewBins), intent(in) ::&
      newXcentres, newXwidths
    real(single), dimension(numNewBins), intent(out) ::&
      newYvalues

    real(single), dimension(0:numOldBins) ::&
      oldXedge
    real(single), dimension(0:numNewBins) ::&
      newXedge
    integer(int16) ::&
      oldBin, oldStartBin, oldFinisBin, newBin
    logical ::&
      foundOverlappingBin
    real(single) ::&
      overlap

    !-------------------------------------------------------------------
    ! Start:

    ! Temporarily convert the centre/width vectors into edge values (this is
    ! not open to numerical ambiguity in the same way):
    oldXedge(0) = oldXcentres(1) - oldXwidths(1) / 2.0
    do oldBin = 1, numOldBins
      oldXedge(oldBin) = oldXcentres(oldBin) + oldXwidths(oldBin) / 2.0
    end do
    newXedge(0) = newXcentres(1) - newXwidths(1) / 2.0
    do newBin = 1, numNewBins
      newXedge(newBin) = newXcentres(newBin) + newXwidths(newBin) / 2.0
    end do

    newYvalues = 0.0 ! vector op, default.

    oldStartBin = 1
    NEW: do newBin = 1, numNewBins
      ! Find first old bin that overlaps newBin:
      foundOverlappingBin = .false. ! default
      OLD1: do oldBin = oldStartBin, numOldBins
        if (oldXedge(oldBin) .gt. newXedge(newBin - 1)&
        .and. oldXedge(oldBin - 1) .lt. newXedge(newBin)) then
          foundOverlappingBin = .true.
          oldStartBin = oldBin
      exit OLD1
        end if
      end do OLD1

      if (.not. foundOverlappingBin) then ! have either reached the end of the
      ! old series, or not reached the start. Leave new value at default 0.0.
    exit NEW
      end if

      ! Find the last old bin that overlaps newBin:
      foundOverlappingBin = .true. ! default
      oldFinisBin = numOldBins ! default
      OLD2: do oldBin = oldStartBin + 1, numOldBins
        if (oldXedge(oldBin - 1) .ge. newXedge(newBin)) then
          foundOverlappingBin = .false.
          oldFinisBin = oldBin - 1
      exit OLD2
        end if
      end do OLD2

      if (foundOverlappingBin) then ! have reached the end of the
      ! old series without leaving the new bin.
        oldFinisBin = numOldBins
      end if

      ! Now work though all the old bins that overlap the new, and pour in
      ! the correct proportional amounts from them into newBin:
      OLD3: do oldBin = oldStartBin, oldFinisBin
        if (oldXedge(oldBin) .lt. newXedge(newBin)) then
          if (oldXedge(oldBin - 1) .gt. newXedge(newBin - 1)) then !B - all
          ! of old bin goes into new:
            overlap = 1.0

          else !A - upper part of old goes into new:
            overlap = (oldXedge(oldBin) - newXedge(newBin - 1))&
            / oldXwidths(oldBin)
          end if

        else
          if (oldXedge(oldBin - 1) .gt. newXedge(newBin - 1)) then !D - lower
          ! part of old goes into new:
            overlap = (newXedge(newBin) - oldXedge(oldBin - 1))&
            / oldXwidths(oldBin)

          else !C - middle part of old goes into new:
            overlap = newXwidths(newBin) / oldXwidths(oldBin)
          end if
        end if
        newYvalues(newBin) = newYvalues(newBin) + oldYvalues(oldBin) * overlap
      end do OLD3

      oldStartBin = oldFinisBin
    end do NEW

  end subroutine rebin

  !---------------------------------------------------------------------
  subroutine makeModelSpectrum(internalSpecSet)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT), intent(out) ::&
      internalSpecSet

    integer(int16), dimension(:), pointer ::&
      EBchannels
    real(kind=double), dimension(:), pointer ::&
      calEmins,&
      calEmaxs
    integer(int16) ::&
      numEBchannels,&
!      numSpecSamples,&
!      EBstartChannel,&
      i,&
      row
    real(kind=double) ::&
      eVrange
    type(TableT) ::&
      internalSpecTab
    type(ColumnT) ::&
      internalSpecBCTReVcol,&
      internalSpecBWIDeVcol,&
      internalSpecBCTRchanCol,&
      internalSpecBWIDchanCol,&
      internalSpecRateCol
    real(single), dimension(:), pointer ::&
      internalSpecBCTReV,&
      internalSpecBWIDeV,&
      internalSpecBCTRchan,&
      internalSpecBWIDchan,&
      internalSpecRate
    real(single), dimension(:), allocatable ::&
      eVedge
    real(single) ::&
      frac,&
      channelStart,&
      channelFinis,&
      eV
    real(single), dimension(:), allocatable ::&
      calEcentres,&
      calEwidths,&
      dummyWidths

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, '  Loading model spectrum.')

    call CAL_getEbounds(EBchannels, calEmins, calEmaxs)
    numEBchannels = size(EBchannels)

!........................................................................
! Fill up columns of the internal spectrum:

    numSpecSamples = 10
!    eVrange = srcTsEmax - srcTsEmin
    eVrange = andedEmax - andedEmin

    internalSpecSet = dataSet('tempspecset.fits', TEMP)
    internalSpecTab = addTable(internalSpecSet, 'SPECTRUM', numSpecSamples)

    internalSpecRateCol   = addColumn(internalSpecTab, 'RATE', Real32)
    internalSpecBCTReVcol = addColumn(internalSpecTab, 'BINCTREV', Real32)
    internalSpecBWIDeVcol = addColumn(internalSpecTab, 'BINWIDEV', Real32)
    internalSpecBCTRchanCol = addColumn(internalSpecTab, 'BCTRCHAN', Real32)
    internalSpecBWIDchanCol = addColumn(internalSpecTab, 'BWIDCHAN', Real32)
    internalSpecRate   => real32Data(internalSpecRateCol)
    internalSpecBCTReV => real32Data(internalSpecBCTReVcol)
    internalSpecBWIDeV => real32Data(internalSpecBWIDeVcol)
    internalSpecBCTRchan => real32Data(internalSpecBCTRchanCol)
    internalSpecBWIDchan => real32Data(internalSpecBWIDchanCol)

    internalSpecRate = 1.0 / eVrange ! vector op: normalised, flat spectrum.

    internalSpecBWIDeV = eVrange / real(numSpecSamples, single) ! vector op.
!***** logarithmic E samples might be better.

    allocate(eVedge(0:numSpecSamples))
!    eVedge(0) = srcTsEmin
    eVedge(0) = andedEmin
    do row = 0, numSpecSamples - 1
      frac = (row + 0.5) / real(numSpecSamples, single)
!      internalSpecBCTReV(row) = srcTsEmin * (1.0 - frac) + srcTsEmax * frac
      internalSpecBCTReV(row) = andedEmin * (1.0 - frac) + andedEmax * frac
      frac = (row + 1.0) / real(numSpecSamples, single)
!      eVedge(row + 1) = srcTsEmin * (1.0 - frac) + srcTsEmax * frac
      eVedge(row + 1) = andedEmin * (1.0 - frac) + andedEmax * frac
    end do ! next row
!***** logarithmic E samples might be better.

    ! Now interpolate to get the channel info:
    allocate(calEcentres(numEBchannels), calEwidths(numEBchannels), dummyWidths(numEBchannels))
    do i = 0, numEBchannels - 1
      calEcentres(i + 1) = (calEmaxs(i) + calEmins(i)) / 2.0
      calEwidths(i + 1)  =  calEmaxs(i) - calEmins(i)
      dummyWidths(i + 1) = 1.0
    end do

!    channelStart = interpolate(srcTsEmin, calEcentres, calEwidths&
!    , real(EBchannels, single), dummyWidths, numEBchannels)
    channelStart = interpolate(andedEmin, calEcentres, calEwidths&
    , real(EBchannels, single), dummyWidths, numEBchannels)

    do row = 0, numSpecSamples - 1
      eV = internalSpecBCTReV(row) + internalSpecBWIDeV(row) / 2.0
      channelFinis = interpolate(eVedge(row + 1), calEcentres, calEwidths&
      , real(EBchannels, single), dummyWidths, numEBchannels)

      internalSpecBCTRchan(row) = (channelFinis + channelStart) / 2.0
      internalSpecBWIDchan(row) = channelFinis - channelStart
      channelStart = channelFinis
    end do

    call release(internalSpecTab)
    call CAL_releaseMemory(EbChannels)
    call CAL_releaseMemory(calEmins)
    call CAL_releaseMemory(calEmaxs)

  end subroutine makeModelSpectrum

  !---------------------------------------------------------------------
  subroutine calcEffArea(tempSpecSet, weightedEffAreaCm2)
  !---------------------------------------------------------------------

! This subroutine calculates the effective area at mirror centre and stores
! this in the temp spectrum file. This is done because the call
! CAL_getEffectiveArea calculates its own energy bounds. These need to be
! rebinned to those of the spectrum. However another reason to store the EA
! in the (temporary) spectrum file is that the EA only needs to be calculated
! once, so it makes sense to do it here and then store the result in what is
! effectively a look-up table.

   implicit none

!.................
! Arguments coming in:

    type(DataSetT), intent(in) ::&
      tempSpecSet

!.................
! Arguments going out:

    real ::&
      weightedEffAreaCm2

!.................
! Others:

    real(kind=single), dimension(:), pointer ::&
      tempSpecEA,&
      tempSpecEctr,&
      tempSpecEwid
    real(kind=double) ::&
      thetaArcsec,&
      phi
    real(kind=single) ::&
      Elo,&
      Ehi
!    real(kind=single), allocatable, dimension(:) ::&
!      Eave
    type(TableT) ::&
      tempSpecTab
    type(ColumnT) ::&
      tempSpecEACol,&
      tempSpecEctrCol,&
      tempSpecEwidCol
    integer ::&
      row,&
      numTempSpecRows,&
      startRow,&
      stopRow
!    integer(kind=int16) ::&
!      i
    real ::&
      numerator,&
      denominator

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, '  Calculate effective area.')

    thetaArcsec = 0.0d0
    phi = 0.0d0

    tempSpecTab = table(tempSpecSet, 'SPECTRUM')
    numTempSpecRows = numberOfRows(tempSpecTab)
    tempSpecEctrCol = column(tempSpecTab, 'BINCTREV')
    tempSpecEwidCol = column(tempSpecTab, 'BINWIDEV')
    tempSpecEACol   = addColumn(tempSpecTab, 'EFFAREA0', Real32, 'CM^2')
    tempSpecEctr => real32Data(tempSpecEctrCol)
    tempSpecEwid => real32Data(tempSpecEwidCol)
    tempSpecEA   => real32Data(tempSpecEACol)

    do row = 1, numTempSpecRows
      Elo = tempSpecEctr(row - 1) - tempSpecEwid(row - 1) / 2.0
      Ehi = tempSpecEctr(row - 1) + tempSpecEwid(row - 1) / 2.0
      tempSpecEA(row - 1) = real(CAL_getEffectiveArea(dble(Elo),&
      dble(Ehi), thetaArcsec, phi), single)
    end do

    startRow = numTempSpecRows + 1 ! sic
    do row = 1, numTempSpecRows
!      if (tempSpecEctr(row - 1) .gt. srcTsEmin) then
      if (tempSpecEctr(row - 1) .gt. andedEmin) then
        startRow = row
    exit
      end if
    end do

    stopRow = 0         ! sic
    do row = numTempSpecRows, 1, -1
!      if (tempSpecEctr(row - 1) .lt. srcTsEmax) then
      if (tempSpecEctr(row - 1) .lt. andedEmax) then
        stopRow = row
    exit
      end if
    end do

    ! stopRow >= startRow has already been checked.

    numerator = 0.0
    denominator = 0.0
    do row = startRow, stopRow
      numerator   = numerator   + tempSpecEctr(row - 1)&
      * tempSpecEwid(row - 1) * tempSpecEA(row - 1)
      denominator = denominator + tempSpecEctr(row - 1)&
      * tempSpecEwid(row - 1)
    end do
    weightedEffAreaCm2 = numerator / denominator

    call release(tempSpecTab)
    ! DON'T release the temp spec set!! It is a temp file and will
    ! disappear if so - and we need it later.

  end subroutine calcEffArea

  !---------------------------------------------------------------------
  subroutine loadCoarseCalArrays(&
!    boxMinCC2X,&
!    boxMinCC2Y,&
!    boxMaxCC2X,&
!    boxMaxCC2Y&
  )
  !---------------------------------------------------------------------

    implicit none 

!    real(single), intent(in) ::&
!      boxMinCC2X,&
!      boxMinCC2Y,&
!      boxMaxCC2X,&
!      boxMaxCC2Y
    integer(int16) :: axis, xi, yi, zi, index, maxIndex
    real(single) :: frac
    real(single), dimension(:), pointer :: thetaArcsec, phi
    real(single), dimension(:), allocatable :: cc2X, cc2Y, cc2Z
    real(double), dimension(:), allocatable :: eVvector

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, '  Loading coarse cal arrays.')

    cvf%binWidth(1) = 0.42 ! about 10 mos pixels wide.
    cvf%binWidth(2) = 0.42 ! about 10 mos pixels wide.
    cvf%lo(1) = boxMinCC2X - cvf%binWidth(1) / 2.0
    cvf%lo(2) = boxMinCC2Y - cvf%binWidth(2) / 2.0
!    cvf%lo(3) = srcTsEmin
    cvf%lo(3) = andedEmin
    cvf%hi(1) = boxMaxCC2X + cvf%binWidth(1) / 2.0
    cvf%hi(2) = boxMaxCC2Y + cvf%binWidth(2) / 2.0
!    cvf%hi(3) = srcTsEmax
    cvf%hi(3) = andedEmax
    cvf%iLo = (/1, 1, 1/)
    do axis = 1, 2
      cvf%numSamples(axis) = 1 + int((cvf%hi(axis) - cvf%lo(axis))&
      / cvf%binWidth(axis))
      cvf%binWidth(axis) = (cvf%hi(axis) - cvf%lo(axis))&
      / (cvf%numSamples(axis) - 1)
    end do
!    cvf%numSamples(3) = numSpecSamples
cvf%numSamples(3) = numCoarseEnergySamples
    cvf%binWidth(3) = (cvf%hi(3) - cvf%lo(3)) / (cvf%numSamples(3) - 1)

    cvf%iHi = cvf%iLo + cvf%numSamples - 1

    cvf%units = (/'mm', 'mm', 'eV'/)

    maxIndex = cvf%numSamples(2) - 1 + cvf%numSamples(2)&
    * (cvf%numSamples(1) - 1)
    allocate(cc2X(0:maxIndex), cc2Y(0:maxIndex), cc2Z(0:maxIndex))
    do xi = 0, cvf%numSamples(1) - 1
      do yi = 0, cvf%numSamples(2) - 1
        index = yi + cvf%numSamples(2) * xi
        frac = xi / real(cvf%numSamples(1) - 1)
        cc2X(index) = cvf%lo(1) * (1.0 - frac) + cvf%hi(1) * frac
        frac = yi / real(cvf%numSamples(2) - 1)
        cc2Y(index) = cvf%lo(2) * (1.0 - frac) + cvf%hi(2) * frac
      end do
    end do
    cc2Z = 0.0
    call CAL_camCoord2ToTelCoord(cc2X, cc2Y, cc2Z, thetaArcsec, phi)

    allocate(eVvector(0:cvf%numSamples(3) - 1))
    do zi = 0, cvf%numSamples(3) - 1
      frac = zi / real(cvf%numSamples(3) - 1)
      eVvector(zi) = cvf%lo(3) * (1.0 - frac) + cvf%hi(3) * frac
    end do

    allocate(cvf%array(cvf%iLo(1):cvf%iHi(1), cvf%iLo(2):cvf%iHi(2)&
    , cvf%iLo(3):cvf%iHi(3)))

    do xi = 0, cvf%numSamples(1) - 1
      do yi = 0, cvf%numSamples(2) - 1
        index = yi + cvf%numSamples(2) * xi
        do zi = 0, cvf%numSamples(3) - 1
          cvf%array(xi + cvf%iLo(1), yi + cvf%iLo(2), zi + cvf%iLo(3)) = &
          CAL_getVignettingFactor(eVvector(zi)&
          , real(thetaArcsec(index), double), real(phi(index), double))
        end do
      end do
    end do

  end subroutine loadCoarseCalArrays

  !---------------------------------------------------------------------
  subroutine calcGtiCorrection(&
    inSetName,&
    treatGti,&
    tdxSet&
  )
  !---------------------------------------------------------------------

    implicit none

!.................
! Arguments coming in:

    character(len=80), intent(in) ::&
      inSetName

!.................
! Arguments modified:

    logical, intent(in out) ::&
      treatGti

!.................
! Arguments going out:

    type(DataSetT), intent(out) ::&
      tdxSet

!.................
! Others:

    type(DssT) ::&
      dss
    type(DataSetT) ::&
      inSet
    type(TableT) ::&
      tdxTab
    type(BlockT) ::&
      rateBlock
    type(ColumnT) ::&
      tdxCol,&
      FTIBcol,&
      gtiStartCol,&
      gtiStopCol
    real(kind=single), dimension(:), pointer ::&
      tdx
    integer(kind=int16) ::&
      ccd,&
      node,&
      ccdnr
    character(8) ::&
      tdxColName,&
      FTIBcolName
    integer ::&
      lastBin,&
      gtiStartBin,&
      gtiStopBin,&
      bin
    ! the kind of all these depends on that of the global variable
    ! numTsBins.

    integer ::&
      numComp,&
      comp,&
      numGtis,&
      gti,&
      ccdNodeNumber
    real(kind=double) ::&
      startTimeFirstBin,&
      stopTimeLastBin,&
      gtiTimeZero,&
!      timeZeroDiff,&
      gtiStartTime,&
      gtiStopTime,&
      binStopTime,&
      binStartTime,&
      gtiMJDref
    real(kind=double), dimension(:), allocatable ::&
      exposure
    real(kind=double), dimension(:), pointer ::&
      gtiStart,&
      gtiStop
    type(DScompT) ::&
      dssComp
    type(RangeT), dimension(:), pointer ::&
      rangesList
    logical ::&
      rangeStructureOk,&
      gtiStructureOk
    type(DSfilterT) ::&
      gtiFilter
    type(TableT) ::&
      gtiTable
    character(len=20) ::&
      gtiTimeUnit
    character(len=80) ::&
      gtiSetTabName

!........................................................................
! Start:

    tdxSet = dataSet('tempTDXset.fits', TEMP)
    tdxTab = addTable(tdxSet, 'TDXTAB', numTsBins)
    do ccd = 1, numCcds
      do node = 1, numNodes
!**** is it really necessary to have tables for those ccds not within the region?? I have inserted the appropriate 'cycle' command...
      if (.not. ccdInRegion(ccd, node)) cycle
        ccdnr = ccd2ccdnr(ccd, node)
        write(tdxColName, '(A, I2.2)') 'TDXCOL', ccdnr
        tdxCol = addColumn(tdxTab, tdxColName, Real32)
        write(FTIBcolName, '(A, I2.2)') 'FTIBIN', ccdnr
        FTIBcol = addColumn(tdxTab, FTIBcolName, Real32)
        ! This is the total amount of frame time covered by the given bin. It
        ! is a quantity that will be calculated during the CR/dead-time
        ! routine and accessed during the binning-noise correction.

        tdx => real32data(tdxCol)
        tdx = 1.0 ! vector op.
        call release(FTIBcol)
        call release(tdxCol)
      end do
    end do

!........................................................................
! GTIs - do some checks first on the DSS and any GTI filters therein:

    if (treatGti) then
      call message(AppMsg, Verbose, '  GTI correction.')

      ! Open DSS:
      inSet = dataSet(inSetName, READ)
      rateBlock = block(inSet, 'RATE')
      dss = dataSubspace(rateBlock)

      numComp = dssComponents(dss)
      if (numComp .le. 0) then
        write(errstr, *) "No components were found in the &
        &DSS of dataset "//trim(inSetName)//". GTI correction not applied."

        call warning('dssNoComponents')
        treatGti = .false.
        call release(inSet)
      end if
    end if

    if (treatGti) then
      ! NOTE!! The calculations for TS bin times assume
      ! that TIMEZERO in the rate file header is the time of the CENTRE of
      ! the first (ie, number 1) bin. This is consistent with the OGIP
      ! standard.
      lastBin = numTsBins
!      startTimeFirstBin = srcTsTimeZero - 0.5 * srcTsBinTimeDel
!      stopTimeLastBin = srcTsTimeZero + (numTsBins - 0.5) * srcTsBinTimeDel
!      startTimeFirstBin = -0.5 * srcTsBinTimeDel
!      stopTimeLastBin = (numTsBins - 0.5) * srcTsBinTimeDel
      startTimeFirstBin = srcTsTime1stBin - 0.5 * srcTsBinTimeDel
      stopTimeLastBin = srcTsTime1stBin + (numTsBins - 0.5) * srcTsBinTimeDel

      tdxTab = table(tdxSet, 'TDXTAB')
      compLoop: do comp = 0, numComp - 1 ! Should be 1 DSS GTI
      ! component per ccd/node.

        dssComp = dssComponent(dss, comp)
        if (.not. (dssHasFilter(dssComp, 'CCDNR', RANGEFILTERT)&
        .and. dssHasFilter(dssComp, 'TIME', GTIFILTERT))) then
          write(errstr, *) "No range or gti filter in component ",&
          comp, " of the DSS of dataset "//trim(inSetName)//". GTI correction &
          &not applied for this component."

          call warning('dssNoRangeOrGti')
      cycle compLoop
        end if

        rangesList => dssRanges(dssComp, 'CCDNR')
        rangeStructureOk = .false. ! default
        if (rangesList(0)%lowerB .ne. UNDEF&
        .and. rangesList(0)%upperB .ne. UNDEF) then
!          ccdNodeNumber = 1 + floor(rangesList(0)%lower)
!          if ((floor(rangesList(0)%upper) - ccdNodeNumber) .eq. 0) then
!            rangeStructureOk = .true.
!          end if

          ccdNodeNumber = int((rangesList(0)%lower + rangesList(0)%upper) / 2.0)
          if (ccdNodeNumber .ge. rangesList(0)%lower&
          .and. ccdNodeNumber .le. rangesList(0)%upper) then
            rangeStructureOk = .true.
          end if
        end if

        if (.not. rangeStructureOk) then
          write(errstr, *) "Bad range filter structure in &
          &component ", comp, " of the DSS of dataset "//trim(inSetName)//&
          ". GTI correction not applied for this component."

          call warning('dssBadRangeStruct')
      cycle compLoop
        end if

        ccd  = ccdnr2ccd(ccdNodeNumber)
        node = ccdnr2node(ccdNodeNumber)

      if (.not. ccdInRegion(ccd, node)) cycle compLoop

        gtiFilter = dssFilter(dssComp, 'TIME', GTIFILTERT)
        gtiSetTabName = dssFilterBlock(gtiFilter)

        gtiStructureOk = .true. ! default
        if (blockType(block(inSet, gtiSetTabName)) .ne. TABLE_BLOCK) then
          gtiStructureOk = .false.
        end if

        if (gtiStructureOk) then
          gtiTable = table(inSet, gtiSetTabName)
          if (.not. (hasColumn(gtiTable, 'START')&
          .and. hasColumn(gtiTable, 'STOP'))) then
            gtiStructureOk = .false.
          end if
        end if

        if (gtiStructureOk) then
          gtiStartCol = column(gtiTable, 'START', READ)
          gtiStopCol = column(gtiTable, 'STOP', READ)
          gtiStart => real64data(gtiStartCol)
          gtiStop => real64data(gtiStopCol)

          ! Check that the START values are all < STOP values:
          numGtis = numberOfRows(gtiTable)
          do gti = 1, numGtis
            if (gtiStart(gti - 1) .ge. gtiStop(gti - 1)) then
              gtiStructureOk = .false.
              call release(gtiStartCol)
              call release(gtiStopCol)
          exit
            end if
          end do        
        end if

        if (gtiStructureOk) then
          if (.not. (hasAttribute(gtiTable, 'TIMEZERO')&
          .and. hasAttribute(gtiTable, 'TIMEUNIT')&
          .and. hasAttribute(gtiTable, 'MJDREF'))) then
            gtiStructureOk = .false.
          end if
        end if

        if (gtiStructureOk) then
          gtiTimeUnit = stringAttribute(gtiTable, 'TIMEUNIT')
          gtiMJDref = real64Attribute(gtiTable, 'MJDREF')
          if (trim(gtiTimeUnit) .ne. trim(srcTimeUnit)&
          .or. gtiMJDref .ne. srcMJDref) then
            gtiStructureOk = .false.
          end if
        end if

        if (.not. gtiStructureOk) then
          write(errstr, *) "Bad GTI filter structure in &
          &component ", comp, " of the DSS of dataset "//trim(inSetName)//&
          ". GTI correction not applied for this component."

          call warning('dssBadGtiStruct')
          call release(gtiTable)
      cycle compLoop
        end if

!........................................................................
! Ok we've passed all checks with flying colours. Let's calculate the
! correction for this component:

        gtiTimeZero = real64Attribute(gtiTable, 'TIMEZERO')
!        timeZeroDiff = gtiTimeZero - srcTsTimeZero
        allocate(exposure(numTsBins))
        exposure = 0.0
        gtiLoop: do gti = 1, numGtis
          gtiStartTime = gtiTimeZero + gtiStart(gti - 1)
          gtiStopTime = gtiTimeZero + gtiStop(gti - 1)
!          gtiStartTime = timeZeroDiff + gtiStart(gti - 1)
!          gtiStopTime  = timeZeroDiff + gtiStop(gti - 1)

!          gtiStartBin = 1 + int((gtiStartTime - startTimeFirstBin)&
!          / srcTsBinTimeDel)
!          gtiStopBin  = 1 + int((gtiStopTime  - startTimeFirstBin)&
!          / srcTsBinTimeDel)
          gtiStartBin = 1 + floor((gtiStartTime - startTimeFirstBin)&
          / srcTsBinTimeDel)
          gtiStopBin  = 1 + floor((gtiStopTime  - startTimeFirstBin)&
          / srcTsBinTimeDel)

          if (gtiStartBin .lt. 1) then
        if (gtiStopBin .lt. 1) cycle gtiLoop

            gtiStartBin = 1
            gtiStartTime = startTimeFirstBin
          end if

          if (gtiStopBin .gt. lastBin) then
            gtiStopBin = lastBin
            gtiStopTime = stopTimeLastBin
          end if

          if (gtiStopBin .eq. gtiStartBin) then ! GTI begins and ends in
          ! the same bin.
            bin = gtiStartBin
            exposure(bin) = exposure(bin) + (gtiStopTime - gtiStartTime)&
            / srcTsBinTimeDel

          else if (gtiStopBin .lt. gtiStartBin) then ! Impossible!
          ! Complain loudly!
            write(errstr, *) "Program bug? gtiStopBin .lt. &
            &gtiStartBin, gti=", gti, ", ccd=", ccd, "."

            call warning('stopbinLtStartbin')
      cycle compLoop

          else ! gtiStopBin .gt. gtiStartBin.
            ! Calculate fractional exposure in the first bin:
            bin = gtiStartBin
!            binStopTime = srcTsTimeZero + (bin - 0.5) * srcTsBinTimeDel
!            binStopTime = (bin - 0.5) * srcTsBinTimeDel
            binStopTime = srcTsTime1stBin + (bin - 0.5) * srcTsBinTimeDel
            exposure(bin) = exposure(bin) + (binStopTime - gtiStartTime)&
            / srcTsBinTimeDel

            ! Intermediate bins fall wholly within the GTI, so set exposure
            ! to 1 for these:
            do bin = gtiStartBin + 1, gtiStopBin - 1
              exposure(bin) = 1.0d0
            end do

            ! Calculate fractional exposure in the last bin:
            bin = gtiStopBin
!            binStartTime = srcTsTimeZero + (bin - 1.5) * srcTsBinTimeDel
!            binStartTime = (bin - 1.5) * srcTsBinTimeDel
            binStartTime = srcTsTime1stBin + (bin - 1.5) * srcTsBinTimeDel
            exposure(bin) = exposure(bin) + (gtiStopTime - binStartTime)&
            / srcTsBinTimeDel

          end if ! gtiStopBin .eq. gtiStartBin.
        end do gtiLoop
        call release(gtiStartCol)
        call release(gtiStopCol)
        call release(gtiTable)

        ! The way exposure is calculated above, it is possible, due to
        ! numerical imprecision, to get values slightly greater than 1.
        ! The next loop (i) truncates these to 1.0, if found; then
        ! (ii) multiplies tdf(ccd, node, t) by exposure(bin) for each bin.
        do bin = 1, lastBin
          if (exposure(bin) > 1.0) then
            exposure(bin) = 1.0d0
          end if
        end do ! next bin

        ccdnr = ccd2ccdnr(ccd, node)
        write(tdxColName, '(A, I2.2)') 'TDXCOL', ccdnr
        tdxCol = column(tdxTab, tdxColName)
        tdx => real32data(tdxCol)
!        tdx(0:lastBin - 1) = tdx(0:lastBin - 1) * exposure(1:lastBin) ! vector op; implicit type conversion from
        tdx = tdx * exposure ! vector op; implicit type conversion from
        ! double to single.

!do bin = 1, lastBin
!print *, bin, exposure(bin), tdx(bin - 1)
!end do
!print *, ' '
        call release(tdxCol)
        deallocate(exposure)

      end do compLoop

      call release(tdxTab)
      call release(inSet)
      call dssRelease(dss)

    end if ! treat GTIs.
  end subroutine calcGtiCorrection

  !---------------------------------------------------------------------
  subroutine calcOtherTimeDepCorrns(&
    tdxSet,&
    inSetName,&
    treatDeadTime,&
    treatCosmicRays,&
    aveOoteFraction&
  )
  !---------------------------------------------------------------------

!........................................................................
! ******These notes need some correction:
!***Now calculate corrections for dead time and cosmic rays (also load the
! array frameTimeInBin, used later to calculate the binning-noise correction):
! Suppose both types of correction are requested by the user. What happens
! to tdf in various error circumstances? (i) If there is no EXPOSU block, or
! if it is missing columns or keywords, or if the frames are out of sequence,
! or if there is too much jitter in the frame times, then tdf for that ccd
! is untouched (and remains at its default value of 1 if no GTI corrections
! have been done). (ii) If there is no more than 1 frame, tdf is set to 0
! for the whole ccd. (iii) If some bins have no frames in them, tdf is set
! to 0 for those bins and ccd.

! Note that, although there is a treatTimeIndep parameter, the following
! routine is not optional. This is because it pre-calculates quantities
! needed to calculate the binning-noise correction.

    implicit none

!.................
! Arguments coming in:

    type(DataSetT), intent(in) ::&
      tdxSet
    character(len=80), intent(in) ::&
      inSetName
    logical, intent(in) ::&
      treatDeadTime,&
      treatCosmicRays

!.................
! Arguments going out:

    real(kind=single), dimension(:,:), intent(out) ::&
      aveOoteFraction

!.................
! Others:

    type(DataSetT) ::&
      inSet,&
      eventSet
    real(kind=double), dimension(:), allocatable ::&
      frameTimeInBin,&
      aveFracexp,&
      aveFracLivetime
    type(TableT) ::&
      tdxTab,&
      exposuTab
    integer(kind=int16) ::&
      ccd,&
      node,&
      ccdnr
    character(8) ::&
      exposuTabName,&
      tdxColName,&
      FTIBcolName
!    integer(kind=int32) ::&
!      exposuTimeZerI
    real(kind=double) ::&
      exposuTimeZero,&
!      exposuTimeZerF,&
      timeZeroDiff,&
      timeOfPrevFrame,&
      frameDuration,&
      totalTint,&
      binStartTime,&
      binStopTime,&
      frameTime,&
      totalFracexp,&
      totalFractime,&
      Tframe
    type(ColumnT) ::&
      exposuFracexpCol,&
      exposuTimedelCol,&
      exposuTimeCol,&
      tdxCol,&
      FTIBcol
    real(kind=single), dimension(:), pointer ::&
      exposuFracexp,&
      exposuTimedel,&
      tdx,&
      FTIbin
    real(kind=double), dimension(:), pointer ::&
      exposuTime
    integer ::&
      numFrames,&
      frame,&
      numFramesInBin
    integer ::&
      startBin,&
      stopBin,&
      bin
    ! kind depends on numTsBins.

    character(len=20) ::&
      timeColUnits

!........................................................................
! Start:

    call message(AppMsg, Verbose, '  Other time-dependent corrections.')

    inSet = dataSet(inSetName, READ)
    eventSet = dataSet(eventSetName, READ)

    tdxTab = table(tdxSet, 'TDXTAB')
    do ccd = 1, numCcds
      nodeLoop: do node = 1, numNodes
      if (.not. ccdInRegion(ccd, node)) cycle nodeLoop

!........................................................................
! First do a lot of checks of the relevant EXPOSUnn extension.

        if (.not. exposuExists(ccd, node)) then
          write(errstr, *) "Unable to make dead-time, CR or &
          &binning noise corrections for ccd ", ccd, " node ", node, &
          " because there is no "//trim(exposuTabName)//" extension in "&
          //trim(eventSetName)//"."

          call warning('noExposuBlock')
      cycle nodeLoop

        else if (.not. neededColumnsPresent(ccd, node)) then
          write(errstr, *) "Unable to make dead-time, CR or &
          &binning noise corrections for ccd ", ccd, " node ", node, &
          " because the "//trim(exposuTabName)//" extension in "&
          //trim(eventSetName)//&
          " doesn't contain the correct columns."

          call warning('badExposuBlock')
      cycle nodeLoop
        end if

        ccdnr = ccd2ccdnr(ccd, node)
        write(exposuTabName, '(A, I2.2)') 'EXPOSU', ccdnr
        exposuTab = table(eventSet, exposuTabName)
        exposuTimeCol = column(exposuTab, 'TIME')
        timeColUnits = units(exposuTimeCol)
!******** The following is a workaround until JB fixes emevents.
if (trim(timeColUnits) .eq. 'SECONDS') then
 timeColUnits = 's'
end if
        if (trim(timeColUnits) .ne. ''&
        .and. trim(timeColUnits) .ne. trim(srcTimeUnit)) then
          write(errstr, *) "The "//trim(exposuTabName)//" extension in "&
          //trim(eventSetName)//" has time units that are different &
          &from those of the source time series."
call error('evlistBadTunits')
        end if

        exposuFracexpCol = column(exposuTab, 'FRACEXP')
        exposuTimedelCol = column(exposuTab, 'TIMEDEL')
        exposuFracexp => real32data(exposuFracexpCol)
        exposuTimedel => real32data(exposuTimedelCol)
        exposuTime    => real64data(exposuTimeCol)

        numFrames = numberOfRows(exposuTab)

!        if (hasAttribute(exposuTab, 'TIMEZERI')&
!        .and. hasAttribute(exposuTab, 'TIMEZERF')) then
!          exposuTimeZerI = int32Attribute(exposuTab, 'TIMEZERI')
!          exposuTimeZerF = real64Attribute(exposuTab, 'TIMEZERF')
!          exposuTimeZero = exposuTimeZerI + exposuTimeZerF
!          ! type conversion to real64
!
!        else if (hasAttribute(exposuTab, 'TIMEZERO')) then
!          exposuTimeZero = real64Attribute(exposuTab, 'TIMEZERO')
!        else
!!          exposuTimeZero = 0.0d0
exposuTimeZero = exposuTime(0)
!        end if

        timeZeroDiff = exposuTimeZero - srcTsTime1stBin
!print *, 'et0 =', exposuTimeZero, 'ts0 =', srcTsTimeZero

        if (numFrames .gt. 1) then
          ! Check that the frame times are consecutive - the main routine
          ! relies on this fact.
          frame = 1
          timeOfPrevFrame = exposuTime(frame - 1)
          do frame = 2, numFrames
            frameDuration = exposuTime(frame - 1) - timeOfPrevFrame
            if (frameDuration .le. 0.0) then
              write(errstr, *) "Unable to make dead-time, &
              &CR or binning noise corrections for ccd ", ccd, &
              "because frames", frame - 2, "and", frame - 1, "in block "&
              //trim(exposuTabName)//" are not sequential."
!print *, timeOfPrevFrame, exposuTime(frame - 2), exposuTime(frame - 1) 
              call warning('frameTimesNotSeq')
              call release(exposuTab)
      cycle nodeLoop
            end if
            timeOfPrevFrame = exposuTime(frame - 1)
          end do ! next frame
        end if ! numFrames .gt. 1

        ! Calculate the start and stop bins. The start bin is the first
        ! which has a stop time later than the time of the first frame
        ! (or bin number 1, whichever is later); the stop bin is the last
        ! bin that has a start time earlier than the time of the last frame
        ! (or bin number numTsBins, whichever is earlier).
        frame = 1
!        startBin = 1 + nint((exposuTime(frame - 1) + timeZeroDiff)&
        startBin = 1 + nint((exposuTime(frame - 1) - srcTsTime1stBin)&
        / srcTsBinTimeDel)
!print *, exposuTime(1), srcTsTime1stBin, srcTsBinTimeDel, startBin

        if (startBin .lt. 1) then
          startBin = 1
        end if

        frame = numFrames
!        stopBin = 1 + nint((exposuTime(frame - 1) + timeZeroDiff)&
        stopBin = 1 + nint((exposuTime(frame - 1) - srcTsTime1stBin)&
        / srcTsBinTimeDel)

        if (stopBin .gt. numTsBins) then
          stopBin = numTsBins
        end if

        if (startBin .gt. numTsBins) then
          write(errstr, *) "Unable to make dead-time, CR or &
          &binning noise corrections for ccd ", ccd, "because the first &
          &frame starts later than the end of the timeseries."

          call warning('firstFrame2Late')

        else if (startBin .gt. 1) then
          write(errstr, *) "Unable to make complete dead-time, &
          &CR or binning noise corrections for ccd ", ccd, " because the &
          &first frame doesn't occur until bin number ", startBin, "."

          call warning('firstFrameLate')
        end if

        if (stopBin .lt. 1) then
          write(errstr, *) "Unable to make dead-time, CR or &
          &binning noise corrections for ccd ", ccd, "because the last &
          &frame starts earlier than the beginning of the timeseries."

          call warning('lastFrame2Early')

        else if (stopBin .lt. numTsBins) then
          write(errstr, *) "Unable to make complete dead-time, &
          &CR or binning noise corrections for ccd ", ccd, " because the &
          &last frame occurs too early, within bin number ", stopBin, "."

          call warning('lastFrameEarly')
        end if
!**** check that, in the event of failure in one of these little routines,
! that the correct thing happens - ie either the tdf(ccd, node) is all 1 or all
! 0 (or partially 1 or 0), as seems best.

!........................................................................
! Ok checks passed, now calculate the corrections. The fracexp correction is
! simple to calculate: one just averages the values of fracexp whose
! corresponding frame centre times fall within the bin. The other correction
! is a little more tricky. What we want is the average fraction of time the
! ccd/node spent in quiescent-state accumulation or integration of useful
! photons; ie we want to average Tint / Tframe for all frames that fall
! within a given bin. Tint is supposedly equal to the TIMEDEL values, but
! Tframe must be calculated by adding Tdead(ccd,node) and Tshift(ccd, node)
! to Tint.

! **** probably should load nominalTframe(ccd, node) and then check that the
! calculated Tframe is never less than the nominal value.

        ! Main routine:

        totalTint = 0.0

        allocate(frameTimeInBin(0:numTsBins-1),&
                 aveFracexp(0:numTsBins-1),&
                 aveFracLivetime(0:numTsBins-1))
        frameTimeInBin = 0.0d0  ! vector op.
        aveFracexp = 0.0d0      ! vector op.
        aveFracLivetime = 0.0d0 ! vector op.

        frame = 1
        do bin = startBin, stopBin
!          binStartTime = srcTsTimeZero + (bin - 1.5) * srcTsBinTimeDel
!          binStopTime = binStartTime + srcTsBinTimeDel
!          frameTime = exposuTimeZero + exposuTime(frame - 1)
          binStartTime = srcTsTime1stBin + (bin - 1.5) * srcTsBinTimeDel
          binStopTime = binStartTime + srcTsBinTimeDel
!          frameTime = timeZeroDiff + exposuTime(frame - 1)
          frameTime = exposuTime(frame - 1)
!***** could get rid of the variable frameTime and just use exposuTime(frame - 1).
!***** could restore timeZeroDiff if did exposuTime = exposuTime - exposuTime(0) 

          ! Move through the frames until we arrive at the first within the
          ! bin:
          do while (frameTime .lt. binStartTime)
            frame = frame + 1
          if (frame .gt. numFrames) exit
!            frameTime = exposuTimeZero + exposuTime(frame - 1)
!            frameTime = timeZeroDiff + exposuTime(frame - 1)
            frameTime = exposuTime(frame - 1)
          end do

          ! At this point frameTime should be > binStartTime.

          ! Now add up all the values of exposuFracexp and
          ! exposuTimedel/Tframe that lie within the bin:
          totalFracexp = 0.0d0
          totalFractime = 0.0d0
          numFramesInBin = 0
          do while (frameTime .lt. binStopTime)
            totalFracexp = totalFracexp + exposuFracexp(frame - 1)

            Tframe = exposuTimedel(frame - 1) + Tdead(ccd, node)&
            + Tshift(ccd, node)
            totalFractime = totalFractime + exposuTimedel(frame - 1) / Tframe
            totalTint = totalTint + exposuTimedel(frame - 1)
!***** NOTE!! These definitions of Tint, TIMEDEL etc are NOT set in concrete, so be alive to possible changes!!

!**** should also check for Tframe = 0 before doing the division in totalFractime.

            frameTimeInBin(bin - 1) = frameTimeInBin(bin - 1) + Tframe

            numFramesInBin = numFramesInBin + 1
            frame = frame + 1
          if (frame .gt. numFrames) exit
!            frameTime = exposuTimeZero + exposuTime(frame - 1)
!            frameTime = timeZeroDiff + exposuTime(frame - 1)
            frameTime = exposuTime(frame - 1)
          end do

          ! Calculate the average value of fracexp and fractime over all
          ! frames in the bin:
          if (numFramesInBin .gt. 0) then
            aveFracexp(bin - 1) = totalFracexp / real(numFramesInBin, double)
            aveFracLivetime(bin - 1) = totalFractime&
            / real(numFramesInBin, double)
          else
            aveFracexp(bin - 1) = 0.0d0      ! If calc of either of these
            aveFracLivetime(bin - 1) = 0.0d0 ! corrections is enabled, this
            ! will set tdf = 0.0
            ! for that ccd, node and bin. Which is what we want, because if
            ! there are no frames in that bin, there can't (or shouldn't)
            ! be any events in there. Remember that (later on) all bins with
            ! exposure equal to zero will be set to nulls.
          end if

        end do ! next bin

        ccdnr = ccd2ccdnr(ccd, node)
        write(tdxColName, '(A, I2.2)') 'TDXCOL', ccdnr
        tdxCol = column(tdxTab, tdxColName)
        tdx => real32data(tdxCol)
        if (treatDeadTime) then
          tdx = tdx * aveFracLivetime ! vector op.
        end if
        if (treatCosmicRays) then
          tdx = tdx * aveFracexp
        end if
        tdx = min(1.0, tdx)
!**** can I call min with an array like this?

!do bin = 0,numTsBins-1
!print *, bin, aveFracLivetime(bin), aveFracexp(bin), tdx(bin)
!end do
!print *, ' '
        call release(tdxCol)
        deallocate(aveFracLivetime, aveFracexp)

        write(FTIBcolName, '(A, I2.2)') 'FTIBIN', ccdnr
        FTIBcol = column(tdxTab, FTIBcolName, real32)
        FTIbin => real32data(FTIBcol)
        FTIbin = frameTimeInBin ! vector op
        call release(FTIBcol)
        deallocate(frameTimeInBin)

        if ((stopBin - startBin + 1) .gt. 0) then
          aveOoteFraction(ccd, node) = Tshift(ccd, node)&
          / (Tshift(ccd, node) + totalTint / (stopBin - startBin + 1))
        else ! a fairly pathological situation - thus we want the
        ! contribution due to OOTEs to evaluate to zero.
          aveOoteFraction(ccd, node) = 0.0
        end if

        call release(exposuTimedelCol)
        call release(exposuFracexpCol)
        call release(exposuTimeCol)
        call release(exposuTab)
      end do nodeLoop
    end do ! next ccd

    call release(tdxTab)
    call release(inSet)
    call release(eventSet)

  end subroutine calcOtherTimeDepCorrns

  !---------------------------------------------------------------------
  subroutine calcSpatial00(&
      tempSpecSet,&
      pixelListSet,&
      ccd,&
      node,&
      imageModel,&
      srcTheta,&
      srcPhi,&
      srcCC2X,&
      srcCC2Y,&
      regionAreaMm2,&
      treatVignet,&
      treatBorderEvents,&
!      treatSelPatterns,&
      treatFilterTrans,&
!      treatQuantumEff,&
      numerator,&
      denominator,&
      eventFraction)
  !---------------------------------------------------------------------

! For the case that only 1 CCD in region, meas spec supplied.

    implicit none

!.................
! Arguments coming in:

    type(DataSetT), intent(in) ::&
      tempSpecSet,&
      pixelListSet
    integer(kind=int8), intent(in) ::&
      ccd,&
      node
    character(len=8), intent(in) ::&
      imageModel
    real(kind=single), intent(in) ::&
      srcTheta,&
      srcPhi,&
      srcCC2X,&
      srcCC2Y,&
      regionAreaMm2
    logical, intent(in) ::&
      treatVignet,&
      treatBorderEvents,&
!      treatSelPatterns,&
      treatFilterTrans!!!!!!,&
!      treatQuantumEff

!.................
! Arguments going out:

    real(kind=single), intent(out) ::&
      numerator,&
      denominator
    real(kind=single), dimension(:,:), intent(out) ::&
      eventFraction

!.................
! Others:

    type(TableT) ::&
      specTab,&
      eventTab
    integer ::&
      numPbins,&
      startPbin,&
      stopPbin,&
      numBadPbins,&
      pBin,&
      pixel
    integer(kind=int8) ::&
      ccdnr
    type(ColumnT)  ::&
      rateCol,&
      dPcol,&
      Ecol,&
      rawXcol,&
      rawYcol,&
      detXcol,&
      detYcol
    real(kind=single), dimension(:), pointer ::&
      specMeas,&
      dp,&
      E,&
      detX,&
      detY,&
      cc2X,&
      cc2Y,&
      thetaArcsec,&
      phi
    integer(kind=int32) ::&
      numPixels
    real(kind=single) ::&
      expr
!,&
!      dp,&
!      procDuration
    real(kind=single), allocatable, dimension(:) ::&
      dummyZ
    integer(kind=int16), dimension(:), pointer ::&
      rawX,&
      rawY
    real(kind=double), dimension(:), allocatable ::&
eVvector,&
filterTransVector,&
totalQeVector
    real(kind=double) ::&
      eV,&
      localCountRate,&
      filterTrans,&
      totalQe,&
!      fracEventsSel,&
      vigFactor
    real(kind=single), dimension(:,:), pointer ::&
      PSFmap
    character(8) ::&
      pixelListTabName

!real(kind=double) :: psfFraction
logical :: logicalDummy

!........................................................................
! Start:

logicalDummy = treatBorderEvents ! Just to avoid compiler warnings.

    call message(AppMsg, Verbose, '    Only 1 ccd in region, spectrum supplied.')

    eventFraction(ccd, node) = 1.0 ! the default for the rest set 0.0
    ! when this global var is declared.

    call CAL_setState(ccdChipId=int(ccd, int32))
    if (numNodes .gt. 1) call CAL_setState(ccdNodeId=int(node - 1, int32))

    specTab = table(tempSpecSet, 'SPECTRUM')
    numPbins = numberOfRows(specTab)
    rateCol = column(specTab, 'RATE')
    dPcol       = column(specTab, 'BWIDCHAN')
    Ecol        = column(specTab, 'BINCTREV')
    specMeas => real32data(rateCol)
    dp       => real32data(dPcol)
    E        => real32data(Ecol)
!    dp = real32Attribute(specTab, 'BWIDCHAN')

    ccdnr = ccd2ccdnr(ccd, node)
    write(pixelListTabName, '(A, I2.2)') 'EVENTS', ccdnr
    eventTab = table(pixelListSet, pixelListTabName)
    numPixels = int(numberOfRows(eventTab), int32)
    ! implicit type conversion to int32. Normally I have set the
    ! variable that contains the return from numberOfRows() as
    ! integer, no kind defined, but my guess is that this is int16
    ! and a whole-camera 1-event-per-pixel event list for MOS has
    ! over 2 million rows, quite a bit bigger than ~32000!

    rawXcol = column(eventTab, 'RAWX')
    rawYcol = column(eventTab, 'RAWY')
    detXcol = column(eventTab, 'DETX')
    detYcol = column(eventTab, 'DETY')
    rawX => int16data(rawXcol)
    rawY => int16data(rawYcol)
    detX => real32data(detXcol)
    detY => real32data(detYcol)
    allocate(cc2X(0:numPixels - 1), cc2Y(0:numPixels - 1))
    cc2X = detX / mmToDetUnits ! vector op.
    cc2Y = detY / mmToDetUnits ! vector op.
    call release(detXcol)
    call release(detYcol)
    allocate(dummyZ(0:numPixels - 1))
    dummyZ = 0.0 ! vector op.
    call CAL_camCoord2ToTelCoord(cc2X, cc2Y, dummyZ, thetaArcsec, phi)
    deallocate(dummyZ)

    startPbin = numPbins + 1 ! sic
    do pBin = 1, numPbins
!      if (E(pBin - 1) .gt. srcTsEmin) then
      if (E(pBin - 1) .gt. andedEmin) then
        startPbin = pBin
    exit
      end if
    end do
    stopPbin = 0         ! sic
    do pBin = numPbins, 1, -1
!      if (E(pBin - 1) .lt. srcTsEmax) then
      if (E(pBin - 1) .lt. andedEmax) then
        stopPbin = pBin
    exit
      end if
    end do
!    if (stopPbin .lt. startPbin) then
!      write(errstr, *) "There are no spectral bins within the energy &
!      &range of the timeseries."
!call error('noSpecBins00')
!    end if

    numerator = 0.0
    denominator = 0.0

    allocate(eVvector(startPbin:stopPbin), totalQeVector(startPbin:stopPbin)&
    , filterTransVector(startPbin:stopPbin))

! For the time being, assume that filterTrans and totalQe are spatially uniform. (As of sas 5.1 alpha this is true.) Just get a vector of their values at the specified energies but at the position of the first pixel in the list:

    do pBin = startPbin, stopPbin
      eVvector(pBin) = dble(E(pBin - 1))
    end do

    if (treatFilterTrans) then
      do pBin = startPbin, stopPbin
        filterTransVector(pBin) = CAL_getFilterTransmission(eVvector(pBin)&
        , 0.0d0, 0.0d0)
if (filterTransVector(pBin) .lt. 0.0) then
  call warning('negativeFilTrans', 'Negative values of filter transmission.')
  filterTransVector(pBin) = 0.0d0
end if
if (filterTransVector(pBin) .gt. 1.0) then
  call warning('badFilTrans', 'Values of filter transmission greater than 1.')
  filterTransVector(pBin) = 1.0d0
end if
!print *, pBin, E(pBin - 1), filterTransVector(pBin)
!if (pBin .gt. 607) call error('','')
      end do
    else
      filterTransVector = 1.0d0
    end if

    call calcQeVector(eVvector(startPbin:stopPbin), rawX(0), rawY(0), totalQeVector(startPbin:stopPbin))

    if (trim(imageModel) .eq. 'point') then ! eqn 33, 34 of SSC-LUX-TN-0053
!      procDuration = 2.42e-3 * numPixels * (1 + stopPbin - startPbin)
!      write(errstr, '(A, F7.1, A)') "This part of the procedure is likely &
!      &to take about", procDuration, " seconds (on a Sun ultra 5)."
!      call message(AppMsg, Verbose)
      call message(AppMsg, Verbose, '      Point-like source.')

      numBadPbins = 0
!print *, stopPbin-startPbin+1, numPixels
      do pBin = startPbin, stopPbin
!print *, 'energy bin', pBin, '/', stopPbin-startPbin+1
        eV = dble(E(pBin - 1))

        localCountRate = 0.0

!        call CAL_getPSFmap(eV, dble(srcTheta), dble(srcPhi),&
!        psfMapHalfsizePixels, psfMapPixelWidthArcsec,&
!        localCountRate, PSFmap)

call my_getPSFmap(psfMapHalfsizePixels, psfMapPixelWidthArcsec,&
PSFmap, psfNormalisation) ! ***** until the CAL routine is implemented. NOTE!! Don't really need the first 2 args for my sub, they are global parameters.

        filterTrans = filterTransVector(pBin)
        totalQe = totalQeVector(pBin)
        expr = 0.0
        do pixel = 1, numPixels
!          if (treatFilterTrans) then
!            filterTrans = CAL_getFilterTransmission(eV,&
!            dble(thetaArcsec(pixel - 1)), dble(phi(pixel - 1)))
!          else
!            filterTrans = 1.0d0
!          end if
!
!          call calcQe(treatQuantumEff, treatSelPatterns, eV,&
!          rawX(pixel - 1), rawY(pixel - 1), totalQe, fracEventsSel)

          expr = expr + filterTrans * totalQe&
          * psf(srcCC2X, srcCC2Y, cc2X(pixel - 1), cc2Y(pixel - 1), PSFmap)
        end do ! next pixel

        if (treatVignet) then
          vigFactor = CAL_getVignettingFactor(eV, dble(srcTheta), dble(srcPhi))
          vigFactor = min(vigFactor, 1.0d0) ! because vigFactor in the ccf
          ! can be slightly > 1.0.
        else
          vigFactor = 1.0d0
        end if

        expr = expr * pixelAreaMm2 * vigFactor / psfNormalisation

!        if (expr .eq. 0.0) then
        if (expr .lt. 0.001) then
          denominator = denominator + specMeas(pBin - 1) * dp(pBin - 1)
          numBadPbins = numBadPbins + 1
        else
          if (expr .gt. 1.0 .and. expr .lt. 1.1) then ! can happen if
          ! filterTrans, totalQe, fracEventsSel and vigFactor are all = 1.0.
          ! In this situation expr is just equal to an estimate of the
          ! fraction of psf contained within the region on this ccd/node.
          ! Since the estimation is fairly crude, this may evaluate to
          ! slightly over 1.
            expr = 1.0
          else if (expr .gt. 1.1) then
            write(errstr, *) "A wrong value was found for the psf &
            &normalisation in subroutine calcSpatial00."

call error('badPsfNorm00')
          end if
          denominator = denominator + specMeas(pBin - 1) * dp(pBin - 1) / expr
        end if
        numerator = numerator + specMeas(pBin - 1) * dp(pBin - 1)
!print *, pBin, E(pBin - 1), filterTrans, totalQe, vigFactor, expr, specMeas(pBin - 1), dp(pBin - 1), denominator, numerator
!if (pBin .gt. 400) call error('','')
      end do ! next pBin

      if (numBadPbins .gt. 0) then
        write(errstr, *) "Some energy bins had very low spatial exposure &
        &factor. This reduces the accuracy of the fracexp calculation."

        call warning('zeroExpFactor')
      end if

    else if (trim(imageModel) .eq. 'uniform') then ! eqn 35, 36 of
    ! SSC-LUX-TN-0053
!      procDuration = 2.16e-3 * numPixels * (1 + stopPbin - startPbin)
!      write(errstr, '(A, F7.1, A)') "This part of the procedure is likely &
!      &to take about", procDuration, " seconds (on a Sun ultra 5)."
!      call message(AppMsg, Verbose)
      call message(AppMsg, Verbose, '      Uniform source.')

      numBadPbins = 0

      do pBin = startPbin, stopPbin
print *, 'energy bin', pBin, '/', stopPbin-startPbin+1
        eV = dble(E(pBin - 1))

        filterTrans = filterTransVector(pBin)
        totalQe = totalQeVector(pBin)
        expr = 0.0
        do pixel = 1, numPixels
!          if (treatFilterTrans) then
!            filterTrans = CAL_getFilterTransmission(eV,&
!            dble(thetaArcsec(pixel - 1)), dble(phi(pixel - 1)))
!          else
!            filterTrans = 1.0d0
!          end if
!
!          call calcQe(treatQuantumEff, treatSelPatterns, eV,&
!          rawX(pixel - 1), rawY(pixel - 1), totalQe, fracEventsSel)

          if (treatVignet) then
            vigFactor = interpolateVigFactor(cc2X(pixel - 1)&
            , cc2Y(pixel - 1), eV)

          else
            vigFactor = 1.0d0
          end if

          expr = expr + filterTrans * totalQe * vigFactor
        end do ! next pixel
        if (expr .eq. 0.0) then
          denominator = denominator + specMeas(pBin - 1) * dp(pBin - 1)
          numBadPBins = numBadPbins + 1
        else
          denominator = denominator + specMeas(pBin - 1) * dp(pBin - 1) / expr
        end if
        numerator = numerator + specMeas(pBin - 1) * dp(pBin - 1)
      end do ! next pBin
      denominator = denominator * regionAreaMm2

      if (numBadPbins .gt. 0) then
        write(errstr, *) "Some energy bins had zero spatial exposure factor. &
        &This reduces the accuracy of the fracexp calculation."

        call warning('zeroExpFactor')
      end if

    else ! assume user-supplied or model image - eqn 37, 38 of SSC-LUX-TN-0053
      write(errstr, *) "This mode is currently not supported."
call error('notSupported')
    end if ! point source

    deallocate(cc2X, cc2Y)
    call CAL_releaseMemory(thetaArcsec, phi)
    call release(specTab)
    call release(eventTab)

  end subroutine calcSpatial00

  !---------------------------------------------------------------------
  subroutine calcSpatial01(&
      tempSpecSet,&
      pixelListSet,&
      ccd,&
      node,&
      imageModel,&
      srcTheta,&
      srcPhi,&
      srcCC2X,&
      srcCC2Y,&
      regionAreaMm2,&
      treatVignet,&
      treatBorderEvents,&
!      treatSelPatterns,&
      treatFilterTrans,&
!      treatQuantumEff,&
      numerator,&
      denominator,&
      eventFraction)
  !---------------------------------------------------------------------
! For the case that only 1 CCD in region, model natural spec required.

    implicit none

!.................
! Arguments coming in:

    type(DataSetT), intent(in) ::&
      tempSpecSet,&
      pixelListSet
    integer(kind=int8), intent(in) ::&
      ccd,&
      node
    character(len=8), intent(in) ::&
      imageModel
    real(kind=single), intent(in) ::&
      srcTheta,&
      srcPhi,&
      srcCC2X,&
      srcCC2Y,&
      regionAreaMm2
    logical, intent(in) ::&
      treatVignet,&
      treatBorderEvents,&
!      treatSelPatterns,&
!      treatQuantumEff,&
      treatFilterTrans

!.................
! Arguments going out:

    real(kind=single), intent(out) ::&
      numerator,&
      denominator
    real(kind=single), dimension(:,:), intent(out) ::&
      eventFraction

!.................
! Others:

    type(TableT) ::&
      specTab,&
      eventTab
    integer ::&
      numEbins,&
      Ebin,&
      pixel
    integer(kind=int8) ::&
      ccdnr
    type(ColumnT)  ::&
      rateCol,&
      dEcol,&
      Ecol,&
      ea0Col,&
      rawXcol,&
      rawYcol,&
      detXcol,&
      detYcol
    integer(kind=int32) ::&
      numPixels
    real(kind=single), dimension(:), pointer ::&
      specNat,&
      dE,&
      E,&
      ea0,&
      detX,&
      detY,&
      cc2X,&
      cc2Y,&
      thetaArcsec,&
      phi
    character(8) ::&
      pixelListTabName
    real(kind=single) ::&
      rIntegral
    real(kind=single), allocatable, dimension(:) ::&
      dummyZ
    integer(kind=int16), dimension(:), pointer ::&
      rawX,&
      rawY
    real(kind=double), dimension(:), allocatable ::&
eVvector,&
filterTransVector,&
totalQeVector
    real(kind=double) ::&
      eV,&
      localCountRate,&
      filterTrans,&
      totalQe,&
!      fracEventsSel,&
      vigFactor
    real(kind=single), dimension(:,:), pointer ::&
      PSFmap
logical :: logicalDummy

!........................................................................
! Start:

logicalDummy = treatBorderEvents ! Just to avoid compiler warnings.

    call message(AppMsg, Verbose, '    Only 1 ccd in region, no spectrum supplied.')
!****** release things after use.

    eventFraction(ccd, node) = 1.0 ! the default for the rest set 0.0
    ! when this global var is declared.

    call CAL_setState(ccdChipId=int(ccd, int32))
    if (numNodes .gt. 1) call CAL_setState(ccdNodeId=int(node - 1, int32))

    specTab = table(tempSpecSet, 'SPECTRUM')
    numEbins = numberOfRows(specTab)
!print *, numEbins
!call error('','')
    rateCol = column(specTab, 'RATE')
    dEcol       = column(specTab, 'BINWIDEV')
    Ecol        = column(specTab, 'BINCTREV')
    ea0Col      = column(specTab, 'EFFAREA0')
    specNat  => real32data(rateCol)
    dE       => real32data(dEcol)
    E        => real32data(Ecol)
    ea0      => real32Data(ea0Col)

    ccdnr = ccd2ccdnr(ccd, node)
    write(pixelListTabName, '(A, I2.2)') 'EVENTS', ccdnr
    eventTab = table(pixelListSet, pixelListTabName)
    numPixels = numberOfRows(eventTab)
    ! implicit type conversion to int32. Normally I have set the
    ! variable that contains the return from numberOfRows() as
    ! integer, no kind defined, but my guess is that this is int16
    ! and a whole-camera 1-event-per-pixel event list for MOS has
    ! over 2 million rows, quite a bit bigger than ~32000!

    rawXcol = column(eventTab, 'RAWX')
    rawYcol = column(eventTab, 'RAWY')
    detXcol = column(eventTab, 'DETX')
    detYcol = column(eventTab, 'DETY')
    rawX => int16data(rawXcol)
    rawY => int16data(rawYcol)
    detX => real32data(detXcol)
    detY => real32data(detYcol)
    allocate(cc2X(0:numPixels - 1), cc2Y(0:numPixels - 1))
    cc2X = detX / mmToDetUnits ! vector op.
    cc2Y = detY / mmToDetUnits ! vector op.
    call release(detXcol)
    call release(detYcol)
    allocate(dummyZ(0:numPixels - 1))
    dummyZ = 0.0 ! vector op.
    call CAL_camCoord2ToTelCoord(cc2X, cc2Y, dummyZ, thetaArcsec, phi)
    deallocate(dummyZ)

    numerator = 0.0
    denominator = 0.0

    allocate(eVvector(numEbins), totalQeVector(numEbins)&
    , filterTransVector(numEbins))

! For the time being, assume that filterTrans and totalQe are spatially uniform. (As of sas 5.1 alpha this is true.) Just get a vector of their values at the specified energies but at the position of the first pixel in the list:

    do Ebin = 1, numEbins
      eVvector(Ebin) = dble(E(Ebin - 1))
    end do

    if (treatFilterTrans) then
      do Ebin = 1, numEbins
        filterTransVector(Ebin) = CAL_getFilterTransmission(eVvector(Ebin)&
        , 0.0d0, 0.0d0)
      end do
    else
      filterTransVector = 1.0d0
    end if

    call calcQeVector(eVvector, rawX(0), rawY(0), totalQeVector)

    if (trim(imageModel) .eq. 'point') then ! eqn 22 of SSC-LUX-TN-0053
      call message(AppMsg, Verbose, '      Point-like source.')
      do Ebin = 1, numEbins
print *, 'energy bin', Ebin, '/', numEbins
        eV = dble(E(Ebin - 1))
        localCountRate = 0.0
!        call CAL_getPSFmap(eV, dble(srcTheta), dble(srcPhi)&
!        , psfMapHalfsizePixels, psfMapPixelWidthArcsec, localCountRate&
!        , PSFmap)

call my_getPSFmap(psfMapHalfsizePixels, psfMapPixelWidthArcsec,&
PSFmap, psfNormalisation) ! ***** until the CAL routine is implemented.

        filterTrans = filterTransVector(Ebin)
        totalQe = totalQeVector(Ebin)
!print *, 'filterTrans =', filterTrans
!print *, '    totalQe =', totalQe
!print *, ' '

        rIntegral = 0.0
        do pixel = 1, numPixels
          rIntegral = rIntegral + filterTrans * totalQe&
          * psf(srcCC2X, srcCC2Y, cc2X(pixel - 1), cc2Y(pixel - 1), PSFmap)
!print *, '  ', pixel, psf(srcCC2X, srcCC2Y, cc2X(pixel - 1), cc2Y(pixel - 1), PSFmap)
        end do ! next pixel
!print *, ' '
!print *, '   rIntegral =', rIntegral
!print *, 'pixelAreaMm2 =', pixelAreaMm2

        rIntegral = rIntegral * pixelAreaMm2 / psfNormalisation
!print *, '   rIntegral =', rIntegral

        if (rIntegral .gt. 1.0 .and. rIntegral .lt. 1.1) then ! can happen if
        ! filterTrans, totalQe and fracEventsSel are all = 1.0.
        ! In this situation rIntegral is just equal to an estimate of the
        ! fraction of psf contained within the region on this ccd/node.
        ! Since the estimation is fairly crude, this may evaluate to
        ! slightly over 1.
          rIntegral = 1.0
        else if (rIntegral .gt. 1.1) then
          write(errstr, *) "A wrong value was found for the psf &
          &normalisation in subroutine calcSpatial01."

call error('badPsfNorm01')
        end if

        if (treatVignet) then
          vigFactor = CAL_getVignettingFactor(eV, dble(srcTheta), dble(srcPhi))
          vigFactor = min(vigFactor, 1.0d0) ! because vigFactor in the ccf
          ! can be slightly > 1.0.
        else
          vigFactor = 1.0d0
        end if

!print *, '        vigFactor =', vigFactor
!print *, '    ea0(Ebin - 1) =', ea0(Ebin - 1)
!print *, 'specNat(Ebin - 1) =', specNat(Ebin - 1)
!print *, '     dE(Ebin - 1) =', dE(Ebin - 1)
!print *, ' '
        numerator = numerator + rIntegral * ea0(Ebin - 1)&
        * specNat(Ebin - 1) * vigFactor * dE(Ebin - 1)
        denominator = denominator + ea0(Ebin - 1) * specNat(Ebin - 1)&
        * dE(Ebin - 1)
      end do ! next Ebin

    else if (trim(imageModel) .eq. 'uniform') then ! eqn 24 of SSC-LUX-TN-0053
      call message(AppMsg, Verbose, '      Uniform source.')

      do Ebin = 1, numEbins
print *, 'energy bin', Ebin, '/', numEbins
        eV = dble(E(Ebin - 1))

        filterTrans = filterTransVector(Ebin)
        totalQe = totalQeVector(Ebin)

        rIntegral = 0.0
        do pixel = 1, numPixels
          if (treatVignet) then
            vigFactor = interpolateVigFactor(cc2X(pixel - 1)&
            , cc2Y(pixel - 1), eV)

          else
            vigFactor = 1.0d0
          end if

          rIntegral = rIntegral + filterTrans * totalQe&
          * vigFactor
        end do ! next pixel
        rIntegral = rIntegral * pixelAreaMm2

        numerator = numerator + rIntegral * ea0(Ebin - 1)&
        * specNat(Ebin - 1) * dE(Ebin - 1)
        denominator = denominator + ea0(Ebin - 1) * specNat(Ebin - 1)&
        * dE(Ebin - 1)
      end do ! next Ebin
      denominator = denominator * regionAreaMm2

    else ! assume user-supplied or model image - eqn 25 of SSC-LUX-TN-0053
      write(errstr, *) "This mode is currently not supported."
call error('notSupported')
    end if ! point source

    deallocate(cc2X, cc2Y)
    call CAL_releaseMemory(thetaArcsec, phi)
    call release(specTab)
    call release(eventTab)

  end subroutine calcSpatial01

  !---------------------------------------------------------------------
  subroutine calcSpatial10a(&
      tempSpecSet,&
      pixelListSet,&
      imageModel,&
      srcTheta,&
      srcPhi,&
      srcCC2X,&
      srcCC2Y,&
      regionAreaMm2,&
      numPixels,&
      treatVignet,&
      treatBorderEvents,&
!      treatSelPatterns,&
      treatFilterTrans,&
!      treatQuantumEff,&
      numerator,&
      denominator,&
      eventFraction)
  !---------------------------------------------------------------------
! For the case that > 1 CCD in region, meas spec supplied and tdf averaged.

    implicit none

!.................
! Arguments coming in:

    type(DataSetT), intent(in) ::&
      tempSpecSet,&
      pixelListSet
    character(len=8), intent(in) ::&
      imageModel
    real(kind=single), intent(in) ::&
      srcTheta, srcPhi,&
      srcCC2X,&
      srcCC2Y,&
      regionAreaMm2
    integer(kind=int32), dimension(:,:), intent(in) ::&
      numPixels
    logical, intent(in) ::&
      treatVignet,&
      treatBorderEvents,&
!      treatSelPatterns,&
      treatFilterTrans!!!!!!!!,&
!      treatQuantumEff

!.................
! Arguments going out:

    real(kind=single), intent(out) ::&
      numerator,&
      denominator
    real(kind=single), dimension(:,:), intent(out) ::&
      eventFraction

!.................
! Others:

    type(TableT) ::&
      specTab,&
      eventTab
    integer ::&
      numPbins,&
      startPbin,&
      stopPbin,&
      masterNumPixels,&
      numBadPbins,&
      pBin,&
      rowStart,&
      pixel,&
      row
    integer(kind=int8) ::&
      ccd,&
      node,&
      ccdnr
    type(ColumnT) ::&
      rateCol,&
      dPcol,&
      Ecol,&
      rawXcol,&
      rawYcol,&
      detXcol,&
      detYcol,&
      ccdnrCol
    real(kind=single), dimension(:), pointer ::&
      specMeas,&
      dp,&
      E
    real(kind=single) ::&
      expr,&
!      dp,&
      contribution,&
      totalEventFraction
!,&
!      procDuration
    real(kind=single), dimension(:), pointer ::&
      detX,&
      detY,&
      cc2X,&
      cc2Y,&
      thetaArcsec,&
      phi
    real(kind=single), allocatable, dimension(:) ::&
      dummyZ
    integer(kind=int16), dimension(:), pointer ::&
      rawX,&
      rawY
    integer(kind=int8), dimension(:), pointer ::&
      masterCcdnr 
    real(kind=double), dimension(:), allocatable ::&
eVvector,&
filterTransVector,&
totalQeVector
    real(kind=double) ::&
      eV,&
      localCountRate,&
      filterTrans,&
      totalQe,&
!      fracEventsSel,&
      vigFactor
    real(kind=single), dimension(:,:), pointer ::&
      PSFmap
logical :: logicalDummy

!........................................................................
! Start:

logicalDummy = treatBorderEvents ! Just to avoid compiler warnings.

!call error('notSupported', 'Not yet supported.')
    call message(AppMsg, Verbose, '    Several ccds in region, spectrum supplied.')

!****** release things after use.

    specTab = table(tempSpecSet, 'SPECTRUM')
    numPBins = numberOfRows(specTab)
    rateCol = column(specTab, 'RATE')
    dPcol   = column(specTab, 'BWIDCHAN')
    Ecol    = column(specTab, 'BINCTREV')
    specMeas => real32data(rateCol)
    dp       => real32data(dPcol)
    E        => real32data(Ecol)
!    dp = real32Attribute(specTab, 'BWIDCHAN')

    eventTab = table(pixelListSet, 'MASTER')
    masterNumPixels = numberOfRows(eventTab)
    rawXcol = column(eventTab, 'RAWX')
    rawYcol = column(eventTab, 'RAWY')
    detXcol = column(eventTab, 'DETX')
    detYcol = column(eventTab, 'DETY')
    ccdnrCol = column(eventTab, 'CCDNR')
    rawX => int16data(rawXcol)
    rawY => int16data(rawYcol)
    detX => real32data(detXcol)
    detY => real32data(detYcol)
    masterCcdnr => int8data(ccdnrCol)
    allocate(cc2X(0:masterNumPixels - 1), cc2Y(0:masterNumPixels - 1))
    cc2X = detX / mmToDetUnits ! vector op.
    cc2Y = detY / mmToDetUnits ! vector op.
    call release(detXcol)
    call release(detYcol)
    allocate(dummyZ(0:masterNumPixels - 1))
    dummyZ = 0.0 ! vector op.

    call CAL_setState(ccdChipId=1)
    if (numNodes .gt. 1) call CAL_setState(ccdNodeId=0) ! arbitrary choice because the
    ! next call is ccd-independent:
!***** well then do we need to do it at all!
    call CAL_camCoord2ToTelCoord(cc2X, cc2Y, dummyZ, thetaArcsec, phi)
    deallocate(dummyZ)

    startPbin = numPbins + 1 ! sic
    do pBin = 1, numPbins
!      if (E(pBin - 1) .gt. srcTsEmin) then
      if (E(pBin - 1) .gt. andedEmin) then
        startPbin = pBin
    exit
      end if
    end do
    stopPbin = 0         ! sic
    do pBin = numPbins, 1, -1
!      if (E(pBin - 1) .lt. srcTsEmax) then
      if (E(pBin - 1) .lt. andedEmax) then
        stopPbin = pBin
    exit
      end if
    end do
!    if (stopPbin .lt. startPbin) then
!      write(errstr, *) "There are no spectral bins within the energy &
!      &range of the timeseries."
!call error('noSpecBins10')
!    end if
!    procDuration = 4e-3 * masterNumPixels * (1 + stopPbin - startPbin)
!    write(errstr, '(A, F7.1, A)') "This part of the procedure is likely &
!    &to take about", procDuration, " seconds."
!    call message(AppMsg, Verbose)

    numerator = 0.0
    denominator = 0.0

    allocate(eVvector(startPbin:stopPbin), totalQeVector(startPbin:stopPbin)&
    , filterTransVector(startPbin:stopPbin))

! For the time being, assume that filterTrans and totalQe are spatially uniform. (As of sas 5.1 alpha this is true.) Just get a vector of their values at the specified energies but at the position of the first pixel in the list:

    do pBin = startPbin, stopPbin
      eVvector(pBin) = dble(E(pBin - 1))
    end do

    if (treatFilterTrans) then
      do pBin = startPbin, stopPbin
        filterTransVector(pBin) = CAL_getFilterTransmission(eVvector(pBin)&
        , 0.0d0, 0.0d0)
      end do
    else
      filterTransVector = 1.0d0
    end if

    call calcQeVector(eVvector(startPbin:stopPbin), rawX(0), rawY(0), totalQeVector(startPbin:stopPbin))

    if (trim(imageModel) .eq. 'point') then ! eqn 33, 34 of SSC-LUX-TN-0053
    ! (but sum over CCDS as well)
      call message(AppMsg, Verbose, '      Point-like source.')

      numBadPbins = 0

      do pBin = startPbin, stopPbin
print *, 'energy bin', pBin, '/', stopPbin-startPbin+1
        eV = dble(E(pBin - 1))

        localCountRate = 0.0
!        call CAL_getPSFmap(eV, dble(srcTheta), dble(srcPhi), psfMapHalfsizePixels,&
!        psfMapPixelWidthArcsec, localCountRate, PSFmap)
!!*** watch out that this is not ccd-dependent.*****

call my_getPSFmap(psfMapHalfsizePixels, psfMapPixelWidthArcsec,&
PSFmap, psfNormalisation) ! ***** until the CAL routine is implemented.

        filterTrans = filterTransVector(pBin)
        totalQe = totalQeVector(pBin)
        rowStart = 0
        expr = 0.0
        do ccd = 1, numCcds
          do node = 1, numNodes
          if (.not. ccdInRegion(ccd, node)) cycle
print *, '  ccd/node:', ccd, '/', node
            call CAL_setState(ccdChipId=int(ccd, int32))
            if (numNodes .gt. 1) then
              call CAL_setState(ccdNodeId=int(node - 1, int32))
            end if

            ccdnr = ccd2ccdnr(ccd, node)

            do pixel = 1, numPixels(ccd, node)
              row = pixel + rowStart
              if (masterCcdnr(row - 1) .ne. ccdnr) then
                write(errstr, *) "Something is badly wrong with &
                &the master temporary event table! CCDNR listed doesn't &
                &equal that calculated."

call error('tempEvlistBadCcdnr')
              end if

!              if (treatFilterTrans) then
!                filterTrans = CAL_getFilterTransmission(eV,&
!                dble(thetaArcsec(row - 1)), dble(phi(row - 1)))
!              else
!                filterTrans = 1.0d0
!              end if
!
!              call calcQe(treatQuantumEff, treatSelPatterns, eV,&
!              rawX(row - 1), rawY(row - 1), totalQe, fracEventsSel)

              contribution = filterTrans * totalQe&
              * psf(srcCC2X, srcCC2Y, cc2X(row - 1), cc2Y(row - 1), PSFmap)
              expr = expr + contribution ! integrated over ccd, node
              eventFraction(ccd, node) = eventFraction(ccd, node)&
              + contribution ! integrated over p (it has to be normalised,
              ! this is done late in this subroutine).
            end do ! next pixel

            rowStart = rowStart + numPixels(ccd, node)
          end do ! next node
        end do ! next ccd

        if (treatVignet) then
          vigFactor = CAL_getVignettingFactor(eV, dble(srcTheta), dble(srcPhi))
          vigFactor = min(vigFactor, 1.0d0) ! because vigFactor in the ccf
          ! can be slightly > 1.0.
        else
          vigFactor = 1.0d0
        end if

        expr = expr * pixelAreaMm2 * vigFactor / psfNormalisation
        if (expr .eq. 0.0) then
          denominator = denominator + specMeas(pBin - 1) * dp(pBin - 1)
          numBadPbins = numBadPbins + 1
        else
          if (expr .gt. 1.0 .and. expr .lt. 1.1) then ! can happen if
          ! filterTrans, totalQe, fracEventsSel and vigFactor are all = 1.0.
          ! In this situation expr is just equal to an estimate of the
          ! fraction of psf contained within the region on this ccd/node.
          ! Since the estimation is fairly crude, this may evaluate to
          ! slightly over 1.
            expr = 1.0
          else if (expr .gt. 1.1) then
            write(errstr, *) "A wrong value was found for the psf &
            &normalisation in subroutine calcSpatial10."

call error('badPsfNorm10')
          end if
          denominator = denominator + specMeas(pBin - 1) * dp(pBin - 1) / expr
        end if
        numerator = numerator + specMeas(pBin - 1) * dp(pBin - 1)
      end do ! next pBin

      if (numBadPbins .gt. 0) then
        write(errstr, *) "Some energy bins had zero spatial exposure factor. &
        &This reduces the accuracy of the fracexp calculation."

        call warning('zeroExpFactor')
      end if

      ! Normalise eventFraction:
      totalEventFraction = 0.0
      do ccd = 1, numCcds
        do node = 1, numNodes
          totalEventFraction = totalEventFraction + eventFraction(ccd, node)
        end do ! next node
      end do ! next ccd
      if (totalEventFraction .le. 0.0) then
        do ccd = 1, numCcds
          do node = 1, numNodes
            eventFraction(ccd, node) = eventFraction(ccd, node)&
            / totalEventFraction
          end do ! next node
        end do ! next ccd
      end if

    else if (trim(imageModel) .eq. 'uniform') then ! eqn 35, 36 of
    ! SSC-LUX-TN-0053 (but sum over CCDS as well)
      call message(AppMsg, Verbose, '      Uniform source.')

      numBadPbins = 0

      do pBin = startPbin, stopPbin
print *, 'energy bin', pBin, '/', stopPbin-startPbin+1
        eV = dble(E(pBin - 1))

        rowStart = 0
        filterTrans = filterTransVector(pBin)
        totalQe = totalQeVector(pBin)
        expr = 0.0
        do ccd = 1, numCcds
          do node = 1, numNodes
          if (.not. ccdInRegion(ccd, node)) cycle
print *, '  ccd/node:', ccd, '/', node
            call CAL_setState(ccdChipId=int(ccd, int32))
            if (numNodes .gt. 1) then
              call CAL_setState(ccdNodeId=int(node - 1, int32))
            end if

            ccdnr = ccd2ccdnr(ccd, node)

            do pixel = 1, numPixels(ccd, node)
              row = pixel + rowStart
              if (masterCcdnr(row - 1) .ne. ccdnr) then
                write(errstr, *) "Something is badly wrong with &
                &the master temporary event table! CCDNR listed doesn't &
                &equal that calculated."

call error('pixelListBadCcdnr')
              end if

!              if (treatFilterTrans) then
!                filterTrans = CAL_getFilterTransmission(eV,&
!                dble(thetaArcsec(row - 1)), dble(phi(row - 1)))
!              else
!                filterTrans = 1.0d0
!              end if
!
!              call calcQe(treatQuantumEff, treatSelPatterns, eV,&
!              rawX(row - 1), rawY(row - 1), totalQe, fracEventsSel)

              if (treatVignet) then
!                vigFactor = CAL_getVignettingFactor(eV,&
!                dble(thetaArcsec(row - 1)), dble(phi(row - 1)))
                vigFactor = interpolateVigFactor(cc2X(pixel - 1)&
                , cc2Y(pixel - 1), eV)

              else
                vigFactor = 1.0d0
              end if

              contribution = filterTrans * totalQe * vigFactor
              expr = expr + contribution ! integrated over ccd, node
              eventFraction(ccd, node) = eventFraction(ccd, node)&
              + contribution ! integrated over p (it has to be normalised,
              ! this is done late in this subroutine).
            end do ! next pixel

            rowStart = rowStart + numPixels(ccd, node)
          end do ! next node
        end do ! next ccd
        if (expr .eq. 0.0) then
          denominator = denominator + specMeas(pBin - 1) * dp(pBin - 1)
          numBadPbins = numBadPbins + 1
        else
          denominator = denominator + specMeas(pBin - 1) * dp(pBin - 1) / expr
        end if
        numerator = numerator + specMeas(pBin - 1) * dp(pBin - 1)
      end do ! next pBin
      denominator = denominator * regionAreaMm2

      if (numBadPbins .gt. 0) then
        write(errstr, *) "Some energy bins had zero spatial exposure &
        &factor. This reduces the accuracy of the fracexp calculation."

        call warning('zeroExpFactor')
      end if

      ! Normalise eventFraction:
      totalEventFraction = 0.0
      do ccd = 1, numCcds
        do node = 1, numNodes
          totalEventFraction = totalEventFraction + eventFraction(ccd, node)
        end do ! next node
      end do ! next ccd
      if (totalEventFraction .le. 0.0) then
        do ccd = 1, numCcds
          do node = 1, numNodes
            eventFraction(ccd, node) = eventFraction(ccd, node)&
            / totalEventFraction
          end do ! next node
        end do ! next ccd
      end if

    else ! assume user-supplied or model image - eqn 37, 38 of SSC-LUX-TN-0053  
    !(but sum over CCDS as well)
      write(errstr, *) "This mode is currently not supported."
call error('notSupported')
    end if ! point source

    deallocate(cc2X, cc2Y)
    call CAL_releaseMemory(thetaArcsec, phi)
    call release(specTab)
    call release(eventTab)

  end subroutine calcSpatial10a

  !---------------------------------------------------------------------
  subroutine calcFracexp10b()
  !---------------------------------------------------------------------
!****error! Not yet supported.
  end subroutine calcFracexp10b

  !---------------------------------------------------------------------
  subroutine calcFracexp11(&
      tempSpecSet,&
      tdxSet,&
      pixelListSet,&
      imageModel,&
      srcTheta,&
      srcPhi,&
      srcCC2X,&
      srcCC2Y,&
      treatVignet,&
      treatBorderEvents,&
!      treatSelPatterns,&
      treatFilterTrans,&
!      treatQuantumEff,&
      fracexp,&
      eventFraction)
  !---------------------------------------------------------------------
! For the case that  > 1 CCD in region, but natural spectrum model used.
! Note: equations 26 - 28 have been modified to take into account the fact
! that the spatial variation in tdf is restricted to changes in the
! ccd/node. The upper integral is changed to (in the example of a point
! source, eqn 26):
!
!   sum_over_ccd[ tdf*int_over_E{ ea0*specNat*vig*int_over_r( fbq*psf )}]
!
! The integral over E thus depends only on the ccd/node and doesn't have to
! be evaluated freshly for each t. The denominator is also evaluated only
! once.

    implicit none

!.................
! Arguments coming in:

    type(DataSetT), intent(in) ::&
      tempSpecSet,&
      tdxSet,&
      pixelListSet
    character(len=8), intent(in) ::&
      imageModel
    real(kind=single), intent(in) ::&
      srcTheta,&
      srcPhi,&
      srcCC2X,&
      srcCC2Y
    real(kind=single), dimension(:), pointer ::&
      fracexp
    logical, intent(in) ::&
      treatVignet,&
      treatBorderEvents,&
!      treatSelPatterns,&
      treatFilterTrans!!!!!!,&
!      treatQuantumEff

!.................
! Arguments going out:

    real(kind=single), dimension(:,:), intent(out) ::&
      eventFraction

!.................
! Others:

    type(TableT) ::&
      specTab,&
      eventTab,&
      tdxTab
    integer ::&
      numEBins,&
      EBin,&
      pixel
    integer(kind=int8) ::&
      ccd,&
      node,&
      ccdnr
    integer(kind=int32) ::&
      numPixels
    type(ColumnT)  ::&
      rateCol,&
      dEcol,&
      Ecol,&
      ea0Col,&
      rawXcol,&
      rawYcol,&
      detXcol,&
      detYcol,&
      tdxCol
    character(len=8) ::&
      pixelListTabName,&
      tdxColName
    real(kind=single), dimension(:), pointer ::&
      specNat,&
      dE,&
      E,&
      ea0
    real(kind=single), dimension(12,2) ::&
      rEIntegral=0.0
    real(kind=single) ::&
      denominator,&
      rIntegral,&
      totalREIntegral
    real(kind=single), dimension(:), pointer ::&
      detX,&
      detY,&
      cc2X,&
      cc2Y,&
      thetaArcsec,&
      phi,&
      tdx
    real(kind=single), allocatable, dimension(:) ::&
      dummyZ
    integer(kind=int16), dimension(:), pointer ::&
      rawX,&
      rawY
    real(kind=double), dimension(:), allocatable ::&
eVvector,&
filterTransVector,&
totalQeVector
    real(kind=double) ::&
      eV, localCountRate,&
      filterTrans,&
      totalQe,&
!      fracEventsSel,&
      vigFactor
    real(kind=single), dimension(:,:), pointer ::&
      PSFmap
logical :: logicalDummy

!........................................................................
! Start:

logicalDummy = treatBorderEvents ! Just to avoid compiler warnings.

!call error('notSupported', 'Not yet supported.')
    call message(AppMsg, Verbose, '    Several ccds in region, no spectrum supplied.')

!****** release things after use.

    specTab = table(tempSpecSet, 'SPECTRUM')
    numEBins = numberOfRows(specTab)
    rateCol = column(specTab, 'RATE')
    dEcol   = column(specTab, 'BINWIDEV')
    Ecol    = column(specTab, 'BINCTREV')
    ea0Col  = column(specTab, 'EFFAREA0')
    specNat  => real32data(rateCol)
    dE       => real32data(dEcol)
    E        => real32data(Ecol)
    ea0      => real32Data(ea0Col)

    allocate(eVvector(numEbins), totalQeVector(numEbins)&
    , filterTransVector(numEbins))

! For the time being, assume that filterTrans and totalQe are spatially uniform. (As of sas 5.1 alpha this is true.) Just get a vector of their values at the specified energies but at the position of the first pixel in the list:

    do Ebin = 1, numEbins
      eVvector(Ebin) = dble(E(Ebin - 1))
    end do

    if (treatFilterTrans) then
      do Ebin = 1, numEbins
        filterTransVector(Ebin) = CAL_getFilterTransmission(eVvector(Ebin)&
        , 0.0d0, 0.0d0)
      end do
    else
      filterTransVector = 1.0d0
    end if

    if (trim(imageModel) .eq. 'point') then ! eqn 26 of SSC-LUX-TN-0053
      call message(AppMsg, Verbose, '      Point-like source.')
      denominator = 0.0
      do EBin = 1, numEBins
        denominator = denominator + ea0(EBin - 1) * specNat(EBin - 1)&
        * dE(EBin - 1)
      end do ! next EBin
      if (denominator .eq. 0.0) then
        write(errstr, *) "Denominator zero in subroutine calcFracexp11. &
        &Can't calculate spatial part of exposure."

call error('denomZero11')
      end if

      totalREIntegral = 0.0
      do ccd = 1, numCcds
        do node = 1, numNodes
        if (.not. ccdInRegion(ccd, node)) cycle
print *, '  ccd/node:', ccd, '/', node
          call CAL_setState(ccdChipId=int(ccd, int32))
          if (numNodes .gt. 1) then
            call CAL_setState(ccdNodeId=int(node - 1, int32))
          end if

          ccdnr = ccd2ccdnr(ccd, node)
          write(pixelListTabName, '(A, I2.2)') 'EVENTS', ccdnr
          eventTab = table(pixelListSet, pixelListTabName)
          numPixels = numberOfRows(eventTab)
          ! implicit type conversion to int32. Normally I have set the
          ! variable that contains the return from numberOfRows() as
          ! integer, no kind defined, but my guess is that this is int16
          ! and a whole-camera 1-event-per-pixel event list for MOS has
          ! over 2 million rows, quite a bit bigger than ~32000!

          rawXcol = column(eventTab, 'RAWX')
          rawYcol = column(eventTab, 'RAWY')
          detXcol = column(eventTab, 'DETX')
          detYcol = column(eventTab, 'DETY')
          rawX => int16data(rawXcol)
          rawY => int16data(rawYcol)
          detX => real32data(detXcol)
          detY => real32data(detYcol)
          allocate(cc2X(0:numPixels - 1), cc2Y(0:numPixels - 1))
          cc2X = detX / mmToDetUnits ! vector op.
          cc2Y = detY / mmToDetUnits ! vector op.
          call release(detXcol)
          call release(detYcol)
          allocate(dummyZ(0:numPixels - 1))
          dummyZ = 0.0 ! vector op.
          call CAL_camCoord2ToTelCoord(cc2X, cc2Y, dummyZ, thetaArcsec, phi)
          deallocate(dummyZ)

          call calcQeVector(eVvector, rawX(0), rawY(0), totalQeVector)
!***** all these calcQeVector calls should use the centre pixel. rawX(0), rawY(0) could be any old place.

          do EBin = 1, numEBins
print *, 'energy bin', Ebin, '/', numEbins
            eV = dble(E(Ebin - 1))
            localCountRate = 0.0
!            call CAL_getPSFmap(eV, dble(srcTheta), dble(srcPhi),&
!            psfMapHalfsizePixels, psfMapPixelWidthArcsec,&
!            localCountRate, PSFmap)

call my_getPSFmap(psfMapHalfsizePixels, psfMapPixelWidthArcsec,&
PSFmap, psfNormalisation) ! ***** until the CAL routine is implemented.

            filterTrans = filterTransVector(Ebin)
            totalQe = totalQeVector(Ebin)

            rIntegral = 0.0
            do pixel = 1, numPixels
!              if (treatFilterTrans) then
!                filterTrans = CAL_getFilterTransmission(eV,&
!                dble(thetaArcsec(pixel - 1)), dble(phi(pixel - 1)))
!              else
!                filterTrans = 1.0d0
!              end if
!
!              call calcQe(treatQuantumEff, treatSelPatterns, eV,&
!              rawX(pixel - 1), rawY(pixel - 1), totalQe, fracEventsSel)

              rIntegral = rIntegral + filterTrans * totalQe&
              * psf(srcCC2X, srcCC2Y, cc2X(pixel - 1), cc2Y(pixel - 1), PSFmap)
            end do ! next pixel

            rIntegral = rIntegral * pixelAreaMm2 / psfNormalisation

            if (rIntegral .gt. 1.0 .and. rIntegral .lt. 1.1) then
            ! can happen if filterTrans, totalQe and fracEventsSel
            ! are all = 1.0. In this situation rIntegral is just
            ! equal to an estimate of the fraction of psf contained within
            ! the region on this ccd/node. Since the estimation is fairly
            ! crude, this may evaluate to slightly over 1.
              rIntegral = 1.0
            else if (rIntegral .gt. 1.1) then
              write(errstr, *) "A wrong value was found for the psf &
              &normalisation in subroutine calcFracexp11."

call error('badPsfNorm11')
            end if

            if (treatVignet) then
              vigFactor = CAL_getVignettingFactor(eV, dble(srcTheta),&
              dble(srcPhi))
              vigFactor = min(vigFactor, 1.0d0) ! because vigFactor in the ccf
              ! can be slightly > 1.0.
            else
              vigFactor = 1.0d0
            end if

            rEIntegral(ccd, node) = rEIntegral(ccd, node) + rIntegral&
            * ea0(EBin - 1) * specNat(EBin - 1) * vigFactor * dE(EBin - 1)
          end do ! next EBin

          deallocate(cc2X, cc2Y)
          call release(eventTab)
          call CAL_releaseMemory(thetaArcsec, phi)

          totalREIntegral = totalREIntegral + rEIntegral(ccd, node)
        end do ! next node
      end do ! next ccd

      tdxTab = table(tdxSet, 'TDXTAB')
      do ccd = 1, numCcds
        do node = 1, numNodes
        if (.not. ccdInRegion(ccd, node)) cycle
          ccdnr = ccd2ccdnr(ccd, node)
          write(tdxColName, '(A, I2.2)') 'TDXCOL', ccdnr
          tdxCol = column(tdxTab, tdxColName)
          tdx => real32data(tdxCol)
          fracexp = fracexp + tdx * rEIntegral(ccd, node)
          call release(tdxCol)

          if (totalREIntegral .gt. 0.0) then
            eventFraction(ccd, node) = rEIntegral(ccd, node) / totalREIntegral
          end if
        end do ! next node
      end do ! next ccd
      fracexp = fracexp / denominator ! vector op
      call release(tdxTab)

    else if (trim(imageModel) .eq. 'uniform') then ! eqn 27 of SSC-LUX-TN-0053
      call message(AppMsg, Verbose, '      Uniform source.')
      denominator = 0.0
      do EBin = 1, numEBins
        denominator = denominator + ea0(EBin - 1) * specNat(EBin - 1)&
        * dE(EBin - 1)
      end do ! next EBin
      if (denominator .eq. 0.0) then
        write(errstr, *) "Denominator zero in subroutine calcFracexp11. &
        &Can't calculate spatial part of exposure."

call error('denomZero11')
      end if

      totalREIntegral = 0.0
      do ccd = 1, numCcds
        do node = 1, numNodes
        if (.not. ccdInRegion(ccd, node)) cycle
print *, '  ccd/node:', ccd, '/', node
          call CAL_setState(ccdChipId=int(ccd, int32))
          if (numNodes .gt. 1) then
            call CAL_setState(ccdNodeId=int(node - 1, int32))
          end if

          ccdnr = ccd2ccdnr(ccd, node)
          write(pixelListTabName, '(A, I2.2)') 'EVENTS', ccdnr
          eventTab = table(pixelListSet, pixelListTabName)
          numPixels = numberOfRows(eventTab)
          ! implicit type conversion to int32. Normally I have set the
          ! variable that contains the return from numberOfRows() as
          ! integer, no kind defined, but my guess is that this is int16
          ! and a whole-camera 1-event-per-pixel event list for MOS has
          ! over 2 million rows, quite a bit bigger than ~32000!

          rawXcol = column(eventTab, 'RAWX')
          rawYcol = column(eventTab, 'RAWY')
          detXcol = column(eventTab, 'DETX')
          detYcol = column(eventTab, 'DETY')
          rawX => int16data(rawXcol)
          rawY => int16data(rawYcol)
          detX => real32data(detXcol)
          detY => real32data(detYcol)
          allocate(cc2X(0:numPixels - 1), cc2Y(0:numPixels - 1))
          cc2X = detX / mmToDetUnits ! vector op.
          cc2Y = detY / mmToDetUnits ! vector op.
          call release(detXcol)
          call release(detYcol)
          allocate(dummyZ(0:numPixels - 1))
          dummyZ = 0.0 ! vector op.
          call CAL_camCoord2ToTelCoord(cc2X, cc2Y, dummyZ, thetaArcsec, phi)
          deallocate(dummyZ)

          call calcQeVector(eVvector, rawX(0), rawY(0), totalQeVector)

          do EBin = 1, numEBins
print *, 'energy bin', Ebin, '/', numEbins
            eV = dble(E(Ebin - 1))

            filterTrans = filterTransVector(Ebin)
            totalQe = totalQeVector(Ebin)

            rIntegral = 0.0
            do pixel = 1, numPixels
!              if (treatFilterTrans) then
!                filterTrans = CAL_getFilterTransmission(eV,&
!                dble(thetaArcsec(pixel - 1)), dble(phi(pixel - 1)))
!              else
!                filterTrans = 1.0d0
!              end if
!
!              call calcQe(treatQuantumEff, treatSelPatterns, eV,&
!              rawX(pixel - 1), rawY(pixel - 1), totalQe, fracEventsSel)

              if (treatVignet) then
!                vigFactor = CAL_getVignettingFactor(eV,&
!                dble(thetaArcsec(pixel - 1)), dble(phi(pixel - 1)))
                vigFactor = interpolateVigFactor(cc2X(pixel - 1)&
                , cc2Y(pixel - 1), eV)
              else
                vigFactor = 1.0d0
              end if

              rIntegral = rIntegral + filterTrans * totalQe&
              * vigFactor
            end do ! next pixel
            rIntegral = rIntegral * pixelAreaMm2

            rEIntegral(ccd, node) = rEIntegral(ccd, node)&
            + rIntegral * ea0(EBin - 1) * specNat(EBin - 1) * dE(EBin - 1)
          end do ! next EBin

          deallocate(cc2X, cc2Y)
          call release(eventTab)
          call CAL_releaseMemory(thetaArcsec, phi)

          totalREIntegral = totalREIntegral + rEIntegral(ccd, node)
        end do ! next node
      end do ! next ccd

      tdxTab = table(tdxSet, 'TDXTAB')
      do ccd = 1, numCcds
        do node = 1, numNodes
        if (.not. ccdInRegion(ccd, node)) cycle
          ccdnr = ccd2ccdnr(ccd, node)
          write(tdxColName, '(A, I2.2)') 'TDXCOL', ccdnr
          tdxCol = column(tdxTab, tdxColName)
          tdx => real32data(tdxCol)
          fracexp = fracexp + tdx * rEIntegral(ccd, node)
          call release(tdxCol)

          if (totalREIntegral .gt. 0.0) then
            eventFraction(ccd, node) = rEIntegral(ccd, node) / totalREIntegral
          end if
        end do ! next node
      end do ! next ccd
      fracexp = fracexp / denominator ! vector op
      call release(tdxTab)

    else ! assume user-supplied or model image - eqn 37, 38 of SSC-LUX-TN-0053
      write(errstr, *) "This mode is currently not supported."
call error('notSupported')
    end if ! point source

    call release(specTab)
    call release(eventTab)

  end subroutine calcFracexp11

  !---------------------------------------------------------------------
  subroutine calcQeVector(&
    eVvector,&
    rawX,&
    rawY,&
    totalQeVector&
  )
  !---------------------------------------------------------------------

    implicit none

!.................
! Arguments coming in:

    real(kind=double), intent(in) ::&
      eVvector(:)
    integer(kind=int16), intent(in) ::&
      rawX,&
      rawY

!.................
! Arguments going out:

    real(kind=double), intent(out) ::&
      totalQeVector(:)

!.................
! Others:

    integer ::&
      patternID, energyBin
    real(kind=double) ::&
      eV

!........................................................................
! Start:

    if (treatQuantumEff) then
      totalQeVector = 0.0d0
      do patternID = 0, numPatterns - 1
!print *, 'pattern =', patternID
      if (treatSelPatterns .and. .not. patternWasSelected(patternID)) cycle
        do energyBin = 1, size(eVvector)
          eV = eVvector(energyBin)
          totalQeVector(energyBin) = totalQeVector(energyBin)&
          + CAL_getQuantumEfficiency(eV, rawX, rawY, patternID)
!print *, 'energyBin, eV, totalQeVector:', energyBin, eV, totalQeVector(energyBin)
        end do
      end do

    else ! don't compensate for QE
      totalQeVector = 1.0d0
    end if ! treatQuantumEff
!call error('','')

  end subroutine calcQeVector

  !---------------------------------------------------------------------
!  function interpolateVigFactor(rawX, rawY, eV) result(vigFactor)
function interpolateVigFactor(cc2X, cc2Y, eV) result(vigFactor)
  !---------------------------------------------------------------------

    implicit none 

!    integer(kind=int16), intent(in) ::&
!      rawX,&
!      rawY
real(single), intent(in) ::&
  cc2X,&
  cc2Y
    real(double), intent(in) ::&
      eV
    real(double) ::&
      vigFactor, f(0:1, 0:1, 0:1)
    integer(kind=int16) ::&
      axis, bin(3), xi, yi, zi
    real(single) ::&
      relCoord, inValue(3), frac(3)

    !-------------------------------------------------------------------
    ! Start:

!    if (rawX .gt. cvf%hi(1)&
!    .or. rawX .lt. cvf%lo(1)) then
!call error('outsideXlimits', '')
!    end if
!
!    if (rawY .gt. cvf%hi(2)&
!    .or. rawY .lt. cvf%lo(2)) then
!call error('outsideYlimits', '')
!    end if

if (cc2X .gt. cvf%hi(1) + cvf%binWidth(1) / 10.0&
.or. cc2X .lt. cvf%lo(1) - cvf%binWidth(1) / 10.0) then
call error('outsideXlimits', '')
end if

if (cc2Y .gt. cvf%hi(2) + cvf%binWidth(2) / 10.0&
.or. cc2Y .lt. cvf%lo(2) - cvf%binWidth(2) / 10.0) then
call error('outsideYlimits', '')
end if

    if (eV .gt. cvf%hi(3) + cvf%binWidth(3) / 10.0&
    .or. eV .lt. cvf%lo(3) - cvf%binWidth(3) / 10.0) then
call error('outsideElimits', '')
    end if

!    inValue(1) = real(rawX, single)
!    inValue(2) = real(rawY, single)
inValue(1) = cc2X
inValue(2) = cc2Y
    inValue(3) = real(eV,   single)

    ! Get the bin number corresponding to the coordinate just less than
    ! or equal to the input coordinate:
    do axis = 1, 3
      relCoord = (inValue(axis) - cvf%lo(axis))&
      / cvf%binWidth(axis)

      bin(axis) = cvf%iLo(axis) + floor(relCoord)
      bin(axis) = min(bin(axis), cvf%iHi(axis) - int(1, int16))
      frac(axis) = relCoord - (bin(axis) - cvf%iLo(axis))
    end do

    do xi = 0, 1
      do yi = 0, 1
        do zi = 0, 1
          f(xi, yi, zi)&
          = cvf%array(bin(1) + xi, bin(2) + yi, bin(3) + zi)
        end do
      end do
    end do

    vigFactor = interpolate3d(f, frac)

  end function interpolateVigFactor

  !---------------------------------------------------------------------
  function interpolate3d(f, frac) result(value)
  !---------------------------------------------------------------------
  ! Given a cubical array of 8 values f(), and a position within the cube specified by a vector frac(), this function returns the trilinear interpolated value at the position frac().

    implicit none 

    real(single), intent(in) ::&
      frac(3)
    real(double), intent(in) ::&
      f(0:1, 0:1, 0:1)
    real(double) ::&
      value, fa, fb, fc, fd, fe, ff

    !-------------------------------------------------------------------
    ! Start:

    fa = f(0, 0, 0) + frac(1) * (f(1, 0, 0) - f(0, 0, 0))
    fb = f(0, 1, 0) + frac(1) * (f(1, 1, 0) - f(0, 1, 0))
    fc = fa + frac(2) * (fb - fa)
    fd = f(0, 0, 1) + frac(1) * (f(1, 0, 1) - f(0, 0, 1))
    fe = f(0, 1, 1) + frac(1) * (f(1, 1, 1) - f(0, 0, 1))
    ff = fd + frac(2) * (fe - fd)

    value = fc + frac(3) * (ff - fc)

  end function interpolate3d

  !---------------------------------------------------------------------
  subroutine calcQe(&
     treatQuantumEff,&
      treatSelPatterns,&
      eV,&
      rawX,&
      rawY,&
      totalQe,&
      fracEventsSel)
  !---------------------------------------------------------------------

    implicit none

!.................
! Arguments coming in:

    logical, intent(in) ::&
      treatQuantumEff,&
      treatSelPatterns
    real(kind=double), intent(in) ::&
      eV
    integer(kind=int16), intent(in) ::&
      rawX,&
      rawY

!.................
! Arguments going out:

    real(kind=double), intent(out) ::&
      totalQe,&
      fracEventsSel

!.................
! Others:

    integer ::&
      patternID
    real(kind=double) ::&
      qeOfCanon,&
      qeOfSelected,&
      qeOfPattern

!........................................................................
! Start:

    if (treatQuantumEff) then
      if (treatSelPatterns) then
      ! Apparently CAL_getQE returns the quantum efficiency weighted
      ! by the fraction of events having the specified pattern. Strange
      ! but true...
        qeOfCanon = 0.0
        qeOfSelected = 0.0
        totalQe = 0.0
        do patternID = 0, numPatterns - 1
          qeOfPattern = CAL_getQuantumEfficiency(eV, rawX, rawY, patternID)
          totalQe = totalQe + qeOfPattern
!          if (patternIsCanonical(patternID)) then
            qeOfCanon = qeOfCanon + qeOfPattern
            if (patternWasSelected(patternID)) then
              qeOfSelected = qeOfSelected + qeOfPattern
            end if
!          end if
        end do
        fracEventsSel = qeOfSelected / qeOfCanon

      else ! don't compensate for unselected patterns
        totalQe = 0.0
        do patternID = 0, numPatterns - 1
          qeOfPattern = CAL_getQuantumEfficiency(eV, rawX, rawY, patternID)
          totalQe = totalQe + qeOfPattern
        end do
        fracEventsSel = 1.0
      end if

    else ! don't compensate for QE
      if (treatSelPatterns) then
        qeOfCanon = 0.0
        qeOfSelected = 0.0
        do patternID = 0, numPatterns - 1
!          if (patternIsCanonical(patternID)) then
            qeOfPattern = CAL_getQuantumEfficiency(eV, rawX, rawY, patternID)
            qeOfCanon = qeOfCanon + qeOfPattern
            if (patternWasSelected(patternID)) then
              qeOfSelected = qeOfSelected + qeOfPattern
            end if
!          end if
        end do
        fracEventsSel = qeOfSelected / qeOfCanon

      else ! don't compensate for unselected patterns
        fracEventsSel = 1.0
      end if
      totalQe = 1.0
    end if ! treatQuantumEff

  end subroutine calcQe

  !---------------------------------------------------------------------
  subroutine my_getPSFmap(mapHalfsizePixels, mapPixelWidthArcsec,&
  PSFmap, psfNormalisation)
  !---------------------------------------------------------------------
  ! This sub returns an approximate PSF while the CAL call is not yet
  ! implemented. A Gaussian is returned with approximately the right
  ! FWHM (~5 arcsec). The function is normalised to units of mm, ie the
  ! unbounded integral of the psf wrt the cc2 coord system should equal 1.
  ! However since exposure integrals involving the psf are necessarily
  ! approximated by sums, the normalisation integral is also approximated
  ! by summing all the array elements and returned as psfNormalisation.

    implicit none

!.................
! Arguments coming in:

    real(kind=double), intent(in) ::&
      mapPixelWidthArcsec
    integer(kind=int32), intent(in) ::&
      mapHalfsizePixels

!.................
! Arguments going out:

    real(kind=single), dimension(:,:), pointer ::&
      PSFmap
    real(kind=single), intent(out) ::&
      psfNormalisation

!.................
! Others:

    real(kind=single), parameter ::&
      pi=3.1415926,&
      fwhmArcsec = 5.0
    real(kind=double) ::&
      mapPixelWidthMm
    real(kind=single) ::&
      fwhmMm,&
      sigmaMm,&
      denom2,&
      amp,&
      x,&
      y
    integer(kind=int32) ::&
      size,&
      i,&
      j

!........................................................................

    fwhmMm = fwhmArcsec / mmToArcsec
    mapPixelWidthMm = mapPixelWidthArcsec / mmToArcsec

    sigmaMm = fwhmMm / 2.0 / sqrt(2.0 * log(2.0))
    denom2 = 2.0 * sigmaMm * sigmaMm
    amp = 0.5 / sigmaMm / sigmaMm / pi

    size = 2 * mapHalfsizePixels + 1
    allocate(PSFmap(size, size))

    psfNormalisation = 0.0
    do i = 1, size
      x = (i - mapHalfsizePixels - 1) * mapPixelWidthMm
      do j = 1, size
        y = (j - mapHalfsizePixels - 1) * mapPixelWidthMm
        PSFmap(i, j) = amp * exp(-(x * x + y * y) / denom2)
        psfNormalisation = psfNormalisation + PSFmap(i, j)
      end do
    end do
    psfNormalisation = psfNormalisation * mapPixelWidthMm * mapPixelWidthMm

  end subroutine my_getPSFmap

  !---------------------------------------------------------------------
  function psf(srcCC2X, srcCC2Y, cc2X, cc2Y, PSFmap)
  !---------------------------------------------------------------------
! Function psf calculates the psf as seen at the given pixel position. If the present pixel position is outside the bounds of the map, psf() returns zero.

!******* this could be sped up by sending it the current and source positions in psf map coordinates.

    implicit none

!.................
! Arguments coming in:

    real(kind=single), intent(in) ::&
      srcCC2X,&
      srcCC2Y
    real(kind=single), intent(in) ::&
      cc2X,&
      cc2Y
    real(kind=single), dimension(:,:), pointer ::&
      PSFmap

!.................
! Arguments going out:

!.................
! Others:

    real(kind=single) ::&
      psf
    integer ::&
      xIndex,&
      yIndex
    integer(kind=int32) ::&
      indexOfPSFarrayCentre,&
      widthOfPSFarray

!........................................................................
! Start:

    indexOfPSFarrayCentre = 1 + psfMapHalfsizePixels
    widthOfPSFarray = 2 * psfMapHalfsizePixels + 1
! The PSF array returned by CAL_getPSFmap is a square array of
! dimension(widthOfPsfArray, widthOfPsfArray). It is (supposedly) aligned
! with the CAMCOORD2 system.

    xIndex = indexOfPSFarrayCentre&
    + nint((cc2X - srcCC2X) * mmToArcsec / psfMapPixelWidthArcsec)
    yIndex = indexOfPSFarrayCentre&
    + nint((cc2Y - srcCC2Y) * mmToArcsec / psfMapPixelWidthArcsec)

    if (xIndex .ge. 1 .and. xIndex .le. widthOfPsfArray&
    .and. yIndex .ge. 1 .and. yIndex .le. widthOfPsfArray) then
      psf = PSFmap(xIndex, yIndex)
    else
      psf = 0.0
    end if

  end function psf

  !---------------------------------------------------------------------
  subroutine deallocateCoarseCalArrays
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

    deallocate(cvf%array)

  end subroutine deallocateCoarseCalArrays

  !---------------------------------------------------------------------
  subroutine calculateOotes(&
      regionYExtentFrac,&
      aveOoteFraction,&
      rateName,&
      rateErrorName,&
      treatOotes)
  !---------------------------------------------------------------------

    implicit none

!.................
! Arguments coming in:

    real(kind=single), dimension(:,:,:), intent(in) ::&
      regionYExtentFrac
    real(kind=single), dimension(:,:), intent(in) ::&
      aveOoteFraction
    character(len=8) ::&
      rateName,&
      rateErrorName

!.................
! Arguments modified:

    logical, intent(in out) ::&
      treatOotes

!.................
! Others:
    integer(kind=int8) ::&
      ccd,&
      node
    type(DataSetT) ::&
      eventSet,&
      outSet
    type(TableT) ::&
      eventTab,&
      outTab
    type(ColumnT) ::&
      eventTimeCol,&
      eventRawXcol,&
      eventCcdnrCol,&
      rateCol,&
      rateErrorCol
    real(kind=double) ::&
      eventTimeZero,&
      timeDiff,&
      firstBinStopTime,&
      firstEventTime,&
      lastBinStartTime,&
      lastEventTime,&
      binStartTime,&
      binStopTime,&
      relEventTime,&
      ootesInRegionThisBin
    integer ::&
      numEvents,&
      event
    real(kind=double), dimension(:), pointer ::&
      eventTime
    real(kind=single), dimension(:), pointer ::&
      rate,&
      rateError
    integer ::&
      bin
    integer(kind=int16) ::&
      colum ! intentionally misspelt to avoid conflict with DAL
      ! 'column' symbol.
    integer(kind=int16), dimension(:), pointer ::&
      eventRawX
    integer(kind=int8), dimension(:), pointer ::&
      eventCcdnr
    integer(kind=int16), dimension(:,:,:), allocatable ::&
      eventsInBin

!........................................................................
! Start:

call error('notSupported', 'Sorry, correction of OOTEs is not yet supported.')

    ! Check for bad modes and warn if any found:
    if (treatOotes) then
      call message(AppMsg, Verbose, '  Calculating OOTEs.')

      do ccd = 1, numCcds
        do node = 1, numNodes
          if (ccdInRegion(ccd, node)) then
            if (ccdMode(ccd, node) .ne. PRIME_FULL_WINDOW&
            .and. ccdMode(ccd, node) .ne. PRIME_PARTIAL_RFS&
!            .and. ccdMode(ccd, node) .ne. EPN_PRIME_FULL_WINDOW&
            .and. ccdMode(ccd, node) .ne. PRIME_FULL_WINDOW_EXTENDED)&
            then
              write(errstr, *) "Some of the CCDs within the &
              &region are in modes that preclude complete correction for &
              &OOTEs. You would be better advised to choose a background &
              &region of exactly the same shape as the source region but &
              &displaced in the CCD y direction. If this is done, simple &
              &background subtraction takes care of the OOTEs."

              call warning('poorOoteRepair')
            end if
          end if
        end do ! next node
      end do ! next ccd
    end if ! treatOotes

! NOTE!! This correction could be sped up by a factor of 2 if it was done
! once for both source and background, since both use the same whole-camera
! event list. But this is a quite bit more complicated. So we'll do it here
! to start with.

! It would also be sped up by factors up to 12 or more if I could filter the
! event list on the basis of CCD and columns in the region (well maybe not so
! much, the filtering takes some time presumably!) but the stupid DAL doesn't
! allow you to open a file and modify the memory copy without also modifying
! the disk copy.

! ALSO!! It assumes that the events come in time order.

! FINALLY! OOTE correction should be done BEFORE the binning noise is
! corrected because OOT events are allocated to bins in the same simplistic
! way that evselect does when constructing the original TS - ie, if the
! event time (= time of the centre of its frame) falls within a bin, the
! event is allocated to that bin.

    if (treatOotes) then
      eventSet = dataSet(eventSetName, READ)
      eventTab = table(eventSet, 'EVENTS')
      eventTimeZero = real64Attribute(eventTab, 'TIMEZERO')
      numEvents =  numberOfRows(eventTab)
      timeDiff = eventTimeZero - srcTsTimeZero
      eventTimeCol = column(eventTab, 'TIME')
      eventTime => real64data(eventTimeCol)

      bin = 1
      event = 1
!      firstBinStopTime = srcTsTimeZero + (bin - 0.5) * srcTsBinTimeDel
!      firstEventTime = eventTimeZero + eventTime(event - 1)
      firstBinStopTime = (bin - 0.5) * srcTsBinTimeDel
      firstEventTime = timeDiff + eventTime(event - 1)
      bin = numTsBins
      event = numEvents
!      lastBinStartTime = srcTsTimeZero + (bin - 1.5) * srcTsBinTimeDel
!      lastEventTime = eventTimeZero + eventTime(event - 1)
      lastBinStartTime = (bin - 1.5) * srcTsBinTimeDel
      lastEventTime = timeDiff + eventTime(event - 1)

      if (firstEventTime .ge. firstBinStopTime&
      .or. lastEventTime .lt. lastBinStartTime) then
        write(errstr, *) "The event list "//trim(eventSetName)//" appears to &
        &contain no events whose times fall within the range of the &
        &timeseries. OOTE correction not done."

        call warning('eventBadTimeRange')
        call release(eventSet)
        treatOotes = .false.
      end if
    end if

    if (treatOotes) then
      eventRawXcol = column(eventTab, 'RAWX')
      eventCcdnrCol = column(eventTab, 'CCDNR')
      eventRawX  => int16data(eventRawXcol)
      eventCcdnr => int8data(eventCcdnrCol)

      allocate(eventsInBin(12, 2, chipMinRawX:chipMaxRawX))

      event = 1
      do bin = 1, numTsBins
!        binStartTime = srcTsTimeZero + (bin - 1.5) * srcTsBinTimeDel
!        binStopTime = binStartTime + srcTsBinTimeDel
!        relEventTime = eventTimeZero + eventTime(event - 1)
        binStartTime = (bin - 1.5) * srcTsBinTimeDel
        binStopTime = binStartTime + srcTsBinTimeDel
        relEventTime = timeDiff + eventTime(event - 1)

        ! Move through the events until we arrive at the first within the
        ! bin:
        do while (relEventTime .lt. binStartTime)
          event = event + 1
        if (event .gt. numEvents) exit
!          relEventTime = eventTimeZero + eventTime(event - 1)
          relEventTime = timeDiff + eventTime(event - 1)
        end do

        ! At this point, 'event' should be the row of the first event within
        ! the timeseries, and relEventTime should be the time (relative to
        ! the start of the timeseries) of this event.

        ! Add up all events per column in this bin:
        eventsInBin = 0 ! vector op
        do while (relEventTime .lt. binStopTime)
          ccd = ccdnr2ccd(eventCcdnr(event - 1))
          node = ccdnr2node(eventCcdnr(event - 1))
          colum = eventRawX(event - 1)
          if (ccd .le. 0 .or. node .le. 0) then
            write(errstr, *) "Event ", event, "in the event list "&
            //trim(eventSetName)//" has a CCDNR which appears &
            &inconsistent with the instrument ", srcInstrument, "."

            call warning('eventBadCcdnr')
        cycle
!**** not sure if one can do this with a 'do while'!

          end if

          if (ccdInRegion(ccd, node)) then
!****** we'll assume this implies
! that the datamode isn't timing. This should all be sorted out just before
! the spatial stuff is done.
            if (colum .lt. windowMinRawX(ccd, node)&
            .or. colum .gt. windowMaxRawX(ccd, node)) then
              write(errstr, *) "Event ", event, "in the event list "&
              //trim(eventSetName)//" has a RAWX value which is outside the &
              &specified window."

              call warning('eventBadRawX')
        cycle
            end if

            eventsInBin(ccd, node, colum)&
            = eventsInBin(ccd, node, colum) + 1
          end if

          event = event + 1
        if (event .gt. numEvents) exit
!          relEventTime = eventTimeZero + eventTime(event - 1)
          relEventTime = timeDiff + eventTime(event - 1)
        end do

        call release(eventSet)

        ! Calculate and subtract OOTEs:
        ootesInRegionThisBin = 0.0
        do ccd = 1, numCcds
          do node = 1, numNodes
            if (ccdInRegion(ccd, node)) then
              do colum = windowMinRawX(ccd, node), windowMaxRawX(ccd, node)
                ootesInRegionThisBin = ootesInRegionThisBin&
                + regionYExtentFrac(ccd, node, colum)&
                * eventsInBin(ccd, node, colum) * aveOoteFraction(ccd, node)
              end do ! next colum
            end if
          end do ! next node
        end do ! next ccd
      end do ! next bin

      deallocate(eventsInBin)

      outSet = dataSet(outSetName, MODIFY)
      outTab = table(outSet, 'RATE')
      rateCol = column(outTab, rateName)
      rateErrorCol = column(outTab, rateErrorName)
      rate      => real32data(rateCol)
      rateError => real32data(rateErrorCol)

      rate = rate - ootesInRegionThisBin / srcTsBinTimeDel        ! All these
      rateError = sqrt(rateError * rateError&                     ! are
      + ootesInRegionThisBin / srcTsBinTimeDel / srcTsBinTimeDel) ! vector ops.
!********** Isn't the above assuming that ootesInRegionThisBin is a vector?????
! Implicit type conversion to single.
      call release(outSet)

    end if ! treatOotes

  end subroutine calculateOotes

  !---------------------------------------------------------------------
  subroutine correctBinningNoise(&
      tdxSet,&
      eventFraction,&
      rateName,&
      rateErrorName,&
      fracexpName)
  !---------------------------------------------------------------------

! The problem can be illustrated as follows. Say we have the following
! (graphically illustrated) situation:
!
! Frames:
!   |---------|---------|---------|---------|---------|---------|---------|
!
! Events in frames (timestamped with the time of frame centre):
!
!        6         6         6         6         6         6         6
!
! Bins: |--------- 1 ---------|--------- 2 ---------|--------- 3 ---------| etc
!
! Events in bins: 18                    12                    12
!
! but should be:  13.2                  13.2                  13.2
!
! Each dash here represents a second of time. (We have neglected the usual
! natural noise in the signal to illustrate the problem better.) The
! correction is done by applying the following formula:
!
! true_rate(bin) = meas_rate(bin) * bin_duration&
! / frameTimeInBin(ccd, node, bin).
!
! Matters are slightly more complicated when multiple ccds are involved:
! the total frame time in the bin for each ccd must be weighted by the
! fraction of events recorded from that ccd. The latter numbers could be
! obtained in an empirical fashion from the whole-camera event list
! product (a procedure which suffers from potential imprecision if the
! number of events is small), but instead I have chosen to get these
! fractions semi-analytically during the spatial part of the exposure
! calculation which is done by the previous section of this subroutine.

    implicit none

!.................
! Arguments coming in:

    type(DataSetT), intent(in) ::&
      tdxSet
    real(kind=single), dimension(:,:), intent(in) ::&
      eventFraction
    character(len=8) ::&
      rateName,&
      rateErrorName,&
      fracexpName

!.................
! Arguments going out:

!.................
! Others:

    logical ::&
      doCorrection
    type(DataSetT) ::&
      outSet
    type(TableT) ::&
      outTab,&
      tdxTab
    type(ColumnT) ::&
      rateCol,&
      rateErrorCol,&
      fracexpCol,&
      FTIBcol
    real(kind=single), dimension(:), pointer ::&
      rate,&
      rateError,&
      fracexp,&
      FTIbin
    real(kind=single), dimension(:), allocatable ::&
      denominator
    integer(kind=int8) ::&
      ccd,&
      node,&
      ccdnr
    character(8) ::&
      FTIBcolName
    integer ::&
      bin

!........................................................................
! Start:

!call error('notSupported', 'Sorry, correction of binning noise is not &
!&yet supported. Binning noise can be avoided by calling e*events with &
!&randomizetime=yes.')

!***** choice here if spatial corrections not requested!!??

    doCorrection = .true. ! This parameter is provided in case we want
    ! to make the correction optional at some future time.

    if (doCorrection) then
      call message(AppMsg, Verbose, '  Correcting for binning noise.')

      outSet = dataSet(outSetName, MODIFY)
      outTab = table(outSet, 'RATE')
      rateCol = column(outTab, rateName)
      rateErrorCol = column(outTab, rateErrorName)
      fracexpCol = column(outTab, fracexpName)
      rate      => real32data(rateCol)
      rateError => real32data(rateErrorCol)
      fracexp   => real32data(fracexpCol)

      allocate(denominator(0:numTsBins-1))
      denominator = 0.0 ! vector op.
      tdxTab = table(tdxSet, 'TDXTAB')
      do ccd = 1, numCcds
        do node = 1, numNodes
        if (.not. ccdInRegion(ccd, node)) cycle

          if (eventTimesRandomized(ccd, node)) then
            call warning('eventsRandomized', 'Event times have been &
            &randomized for this ccd/node. Binning noise correction &
            &not applied.')

            denominator = denominator + eventFraction(ccd, node)
            ! vector op.
        cycle
          end if

          ccdnr = ccd2ccdnr(ccd, node)
          write(FTIBcolName, '(A, I2.2)') 'FTIBIN', ccdnr
          FTIBcol = column(tdxTab, FTIBcolName, real32)
          FTIbin => real32data(FTIBcol)

          denominator = denominator + eventFraction(ccd, node) * FTIbin
          ! vector op.

          call release(FTIBcol)
        end do ! next node
      end do ! next ccd
      call release(tdxTab)

      denominator = denominator / srcTsBinTimeDel ! vector op.

      do bin = 1, numTsBins
        if (denominator(bin - 1) .eq. 0.0) then ! must be because there are
        ! no frames in this bin. Exposure is zero and the rate etc must
        ! therefore be set to null. (The latter should be done in the
        ! calling routine, after the present call to correctBinningNoise.)
          fracexp(bin - 1) = 0.0
        else
          rate(bin - 1) = rate(bin - 1) / denominator(bin - 1)
          rateError(bin - 1) = rateError(bin - 1) / denominator(bin - 1)
        end if

      end do ! next bin
      deallocate(denominator)

      call release(outSet)
    end if ! doCorrection

  end subroutine correctBinningNoise

  !---------------------------------------------------------------------
  subroutine compareSrcBkgKwds(&
    srcInstrument,   bkgInstrument,&
    numBkgTsBins,    numTsBins,&
    srcTimeUnit,     bkgTimeUnit,&
    srcMJDref,       bkgMJDref,&
    srcTsTimeZero,   bkgTsTimeZero,&
    srcTsBinTimeDel, bkgTsBinTimeDel,&
    rebinNecessary&
  )
  !---------------------------------------------------------------------

    implicit none

    real(kind=double), intent(in) ::&
      srcMJDref, bkgMJDref, srcTsTimeZero, bkgTsTimeZero, srcTsBinTimeDel, bkgTsBinTimeDel
    character(len=20), intent(in) ::&
      srcInstrument, bkgInstrument,&
      srcTimeUnit, bkgTimeUnit
    integer(kind=int32), intent(in) ::&
      numBkgTsBins, numTsBins
    logical, intent(out) ::&
      rebinNecessary

    !-------------------------------------------------------------------
    ! Start:

    if (trim(srcInstrument) .ne. trim(bkgInstrument)) then
      write(errstr, *) "Source and background timeseries &
      &INSTRUME keywords don't match."
call error('tsDiffInstrumeKwds')
    end if

    if (trim(srcTimeUnit) .ne. trim(bkgTimeUnit)) then
      write(errstr, *) "Source and background timeseries &
      &TIMEUNIT keywords don't match."
call error('tsDiffTimeunitKwds')
    end if

    if (srcMJDref .ne. bkgMJDref) then
      write(errstr, *) "Source and background timeseries &
      &MJDREF keywords don't match."
call error('tsDiffMJDrefKwds')
    end if

    rebinNecessary = .false.
    if (srcTsTimeZero .ne. bkgTsTimeZero) then
      write(errstr, *) "The time of first entry of the TIME column is &
      &different for the source and background timeseries. Background &
      &will be rebinned."
!call error('tsDiffTimezero')
      call warning('tsDiffTimezero')
      rebinNecessary = .true.
    end if

    if (srcTsBinTimeDel .ne. bkgTsBinTimeDel) then
      write(errstr, *) "The source and background timeseries don't &
      &have the same bin width. Background will be rebinned."
!call error('tsDiffBinWidth')
      call warning('tsDiffBinWidth')
      rebinNecessary = .true.
    end if

    if (numBkgTsBins .ne. numTsBins) then 
      write(errstr, *) "The source and background timeseries &
      &have different numbers of bins. Background will be rebinned."
!call error('tsDiffNumBins')
      call warning('tsDiffNumBins')
      rebinNecessary = .true.
    end if

  end subroutine compareSrcBkgKwds

  !---------------------------------------------------------------------
  subroutine subtractCompensatedBkg(srcRegionAreaMm2, bkgRegionAreaMm2)
  !---------------------------------------------------------------------

    implicit none

    real(kind=single), intent(in) ::&
      srcRegionAreaMm2,&
      bkgRegionAreaMm2
      
    type(DataSetT) ::&
      outSet
    type(TableT)   ::&
      outTab
    type(ColumnT)  ::&
      outRateCol,&
      outRerrorCol,&
      outBkgRateCol,&
      outBkgRerrorCol
    integer(kind=int32) ::&
      bin
    real(kind=single),  dimension(:), pointer ::&
      outRate,&
      outRerror,&
      outBkgRate,&
      outBkgRerror

    !-------------------------------------------------------------------
    ! Start:

!........................................................................
!........................................................................
! Normalise bkg to src collection area:

    call message(AppMsg, Verbose, 'Normalising bkg to src collection area.')

    if (srcRegionAreaMm2 .le. 0.0) then
      write(errstr, *) "Source collection area is of zero size."
call error('srcRegionAreaLeZero')
    end if
    if (bkgRegionAreaMm2 .le. 0.0) then
      write(errstr, *) "Background collection area is of zero size."
call error('bkgRegionAreaLeZero')
    end if

    outSet = dataSet(outSetName, MODIFY)
    outTab = Table(outSet, 'RATE')

    outBkgRateCol = column(outTab, 'BRATE', MODIFY)
    outBkgRerrorCol = column(outTab, 'BRERROR', MODIFY)

    outBkgRate => real32data(outBkgRateCol)
    outBkgRerror => real32data(outBkgRerrorCol)

    ! Vector ops:
    outBkgRate = outBkgRate * srcRegionAreaMm2 / bkgRegionAreaMm2
    outBkgRerror = outBkgRerror * srcRegionAreaMm2 / bkgRegionAreaMm2

!........................................................................
! Subtract background from source if this was requested:

    if (subtractBkg) then
      call message(AppMsg, Verbose, 'Subtracting bkg from src TS.')

      outRateCol = column(outTab, 'RATE', MODIFY)
      outRerrorCol = column(outTab, 'RERROR', MODIFY)
      outRate => real32data(outRateCol)
      outRerror => real32data(outRerrorCol)
      do bin = 1, numTsBins
        if (.not. isNull(outRateCol, bin - 1)) then
          if (isNull(outBkgRateCol, bin - 1)) then
            call setToNull(outRateCol, bin - 1)
          else
            outRate(bin - 1) = outRate(bin - 1) - outBkgRate(bin - 1)
            outRerror(bin - 1) = sqrt(outRerror(bin - 1) * outRerror(bin - 1)&
            + outBkgRerror(bin - 1) * outBkgRerror(bin - 1))
          end if
        end if
      end do

      call setAttribute(outTab, 'BACKAPP', .true., '')
    else
      call setAttribute(outTab, 'BACKAPP', .false., '')
    end if

    call release(outSet)

  end subroutine subtractCompensatedBkg

  !---------------------------------------------------------------------
  subroutine rebinTS(vectorB, numBinsA, numBinsB, minA, maxA, minB, maxB)
  !---------------------------------------------------------------------
!rebin(
    implicit none 

    real(kind=single),  dimension(:), pointer ::&
      vectorB, rebinnedVectorB
    integer(int32), intent(in) ::&
      numBinsA, numBinsB
    real(double), intent(in) ::&
      minA, maxA, minB, maxB

    real(kind=single) ::&
      realNumBinsA,&
      realNumBinsB,&
      deltaA,&
      binWidthA,&
      binWidthB,&
      binWidthRatio,&
      frac,&
      overlap,&
      loEdgeBinA,&
      hiEdgeBinA,&
      loEdgeBinB,&
      hiEdgeBinB
    integer(int32) ::&
      binA,&
      binB,&
      startBinA,&
      finisBinA

    !-------------------------------------------------------------------
    ! Start:

    realNumBinsA = real(numBinsA)
    realNumBinsB = real(numBinsB)
    deltaA = maxA - minA
    binWidthA = deltaA / realNumBinsA
    binWidthB = (maxB - minB) / realNumBinsB
    binWidthRatio = binWidthA / binWidthB

    allocate(rebinnedVectorB(0:numBinsA - 1))
    rebinnedVectorB = 0.0
    do binB = 1, numBinsB
      frac = (binB - 1) / realNumBinsB
      loEdgeBinB = minB * (1.0 - frac) + maxB * frac
      startBinA = 1 + floor(((loEdgeBinB - minA) / deltaA) * realNumBinsA)
      startBinA = max(startBinA, 1)

      frac = binB / realNumBinsB
      hiEdgeBinB = minB * (1.0 - frac) + maxB * frac
      finisBinA = 1 + floor(((hiEdgeBinB - minA) / deltaA) * realNumBinsA)
      finisBinA = min(finisBinA, numBinsA)

      do binA = startBinA, finisBinA
        frac = (binA - 1) / realNumBinsA
        loEdgeBinA = minA * (1.0 - frac) + maxA * frac
        frac = binA / realNumBinsA
        hiEdgeBinA = minA * (1.0 - frac) + maxA * frac

        overlap = 0.0
        if (loEdgeBinA .lt. loEdgeBinB) then
          if (hiEdgeBinA .gt. hiEdgeBinB) then
            overlap = 1.0
          else if (hiEdgeBinA .gt. loEdgeBinB) then
            overlap = (hiEdgeBinA - loEdgeBinB) / binWidthB
!          else
          end if
        else if (loEdgeBinA .lt. hiEdgeBinB) then
          if (hiEdgeBinA .gt. hiEdgeBinB) then
            overlap = (hiEdgeBinB - loEdgeBinA) / binWidthB
          else if (hiEdgeBinA .gt. loEdgeBinB) then
            overlap = binWidthRatio
!          else
          end if
!        else
        end if

        rebinnedVectorB(binA - 1) = rebinnedVectorB(binA - 1) + overlap&
        * vectorB(binB - 1)
      end do
    end do
    deallocate(vectorB)
    vectorB => rebinnedVectorB

    ! Put in nulls in stretches of A range where there was incomplete B data available:
    startBinA = ceiling(((minB - minA) / deltaA) * realNumBinsA)
    finisBinA = 1 + floor(((maxB - minA) / deltaA) * realNumBinsA)
    do binA = 1, startBinA
      vectorB(binA - 1) = myRealSingleNull
    end do
    do binA = finisBinA, numBinsA
      vectorB(binA - 1) = myRealSingleNull
    end do

  end subroutine rebinTS

  !---------------------------------------------------------------------
  subroutine copyOverBackvColToOutSet(numTsBins)
  !---------------------------------------------------------------------

    implicit none

    integer(kind=int32), intent(in) ::&
      numTsBins

    type(DataSetT) ::&
      outSet, srcTsSet
    type(TableT)   ::&
      outTab, srcTsTab
    type(ColumnT)  ::&
      outBkgCountsCol,&
      outBkgErrorCol,&
      srcTsBackvCol,&
      srcTsBackeCol
    integer(kind=int8), dimension(:), pointer ::&
      srcTsBackv8,&
      srcTsBacke8
    integer(kind=int16), dimension(:), pointer ::&
      srcTsBackv16,&
      srcTsBacke16
    integer(kind=int32), dimension(:), pointer ::&
      srcTsBackv32,&
      srcTsBacke32,&
      outBkgCounts
    real(kind=single),  dimension(:), pointer ::&
      outBkgError
    integer(kind=int32) ::&
      bin

    !-------------------------------------------------------------------
    ! Start:

      outSet = dataSet(outSetName, MODIFY)
      outTab = Table(outSet, 'RATE')
      srcTsSet = dataSet(srcTsSetName, READ)
      srcTsTab = table(srcTsSet, 'RATE')

      if (.not. hasColumn(srcTsTab, 'BACKV')) then
        call setAttribute(outTab, 'HDUCLAS2', 'NET',&
       'Contains just source data.')
      else

        call setAttribute(outTab, 'HDUCLAS2', 'TOTAL',&
        'Contains both source and background data.')

        outBkgCountsCol = column(outTab, 'BCOUNTS')
        outBkgCounts => int32Data(outBkgCountsCol)
        srcTsBackvCol = column(srcTsTab, 'BACKV')
        if (columnDataType(srcTsBackvCol) .eq. INTEGER8) then
          srcTsBackv8 => int8Data(srcTsBackvCol)
          outBkgCounts = srcTsBackv8 ! type conversion to int32
        else if (columnDataType(srcTsBackvCol) .eq. INTEGER16) then
          srcTsBackv16 => int16Data(srcTsBackvCol)
          outBkgCounts = srcTsBackv16 ! type conversion to int32
        else
          srcTsBackv32 => int32Data(srcTsBackvCol)
          outBkgCounts = srcTsBackv32
        end if

        do bin = 1, numTsBins
          if (outBkgCounts(bin - 1) .lt. 0) then
            write(errstr, *) "BACKV value < 0 detected in the &
            &RATE extension of the source timeseries."
call error('srcTsBackvLeZero')
          end if
        end do

        call release(outBkgCountsCol)

        outBkgErrorCol = column(outTab, 'BERROR')
        outBkgError => real32Data(outBkgErrorCol)
        if (hasColumn(srcTsTab, 'BACKE')) then
          srcTsBackeCol = column(srcTsTab, 'BACKE')
          if (columnDataType(srcTsBackeCol) .eq. INTEGER8) then
            srcTsBacke8 => int8Data(srcTsBackeCol)
            outBkgError = srcTsBacke8 ! type conversion to real32
          else if (columnDataType(srcTsBackeCol) .eq. INTEGER16) then
            srcTsBacke16 => int16Data(srcTsBackeCol)
            outBkgError = srcTsBacke16 ! type conversion to real32
          else
            srcTsBacke32 => int32Data(srcTsBackeCol)
            outBkgError = srcTsBacke32 ! type conversion to real32
          end if

          call release(srcTsBackeCol)

        else ! assume Poissonian uncertainties.
          outBkgCountsCol = column(outTab, 'BCOUNTS')
          outBkgCounts => int32Data(outBkgCountsCol)
          outBkgError = sqrt(real(outBkgCounts, single)) ! vector op;
        end if
      end if

      call release(srcTsSet)
      call release(outSet)

  end subroutine copyOverBackvColToOutSet

  !---------------------------------------------------------------------
  subroutine finaliseOutSet()
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT) ::&
      outSet
    type(TableT)   ::&
      outTab
    type(ColumnT)  ::&
      outRerrorCol,&
      outBkgRateCol,&
      outBkgRerrorCol

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Cleaning up output TS.')

    outSet = dataSet(outSetName, MODIFY)
    outTab = Table(outSet, 'RATE')

    if (hasColumn(outTab, 'COUNTS')) then
      call deleteColumn(outTab, 'COUNTS')
    end if
    if (hasColumn(outTab, 'ERROR')) then
      call deleteColumn(outTab, 'ERROR')
    end if
    ! leave RATE

    outRerrorCol = column(outTab, 'RERROR', MODIFY)
    call rename(outRerrorCol, 'ERROR')
    call release(outRerrorCol)

    if (hasColumn(outTab, 'BCOUNTS')) then
      call deleteColumn(outTab, 'BCOUNTS')
    end if
    if (hasColumn(outTab, 'BERROR')) then
      call deleteColumn(outTab, 'BERROR')
    end if

    outBkgRateCol = column(outTab, 'BRATE', MODIFY)
    call rename(outBkgRateCol, 'BACKV')
    call release(outBkgRateCol)
    outBkgRerrorCol = column(outTab, 'BRERROR', MODIFY)
    call rename(outBkgRerrorCol, 'BACKE')
    call release(outBkgRerrorCol)

    if (hasColumn(outTab, 'BDEADC')) then
      call deleteColumn(outTab, 'BDEADC')
    end if
    if (hasColumn(outTab, 'BVIGNET')) then
      call deleteColumn(outTab, 'BVIGNET')
    end if
    if (hasColumn(outTab, 'BFRACEXP')) then
      call deleteColumn(outTab, 'BFRACEXP')
    end if

!***** replace myRealSingleNull values in all columns by dal nulls.

!****** write pixel area keywords???

    ! Release the output dataset first so the above changes get written,
    ! then reopen to count the columns:
!    call release(outTab)
    call release(outSet)
!    outSet = dataSet(outSetName, READ)
!    outTab = Table(outSet, 'RATE')
!    if (numberOfColumns(outTab) .ne. 5) then
!      write(errstr, *) "Wrong number of columns in output dataset?"
!call error('outTsWrongNumCols')
!    end if
!    call release(outTab)
!    call release(outSet)

  end subroutine finaliseOutSet

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

  end subroutine template

!++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
! No longer used subroutines:

  !---------------------------------------------------------------------
  subroutine checkAndRebinSpectrum(&
    specSetName,&
    rebinSpecMode,&
    specBinDiv,&
    useSpecSet,&
    rebinSpec,&
    tempSpecSet)
  !---------------------------------------------------------------------

! Sort out meas vs model spectrum and rebin if necessary. The basic idea
! of this routine is to create a temporary file that has an extension
! called 'SPEC' with columns as follows:
!	RATE		- the rebinned, normalised count rate
!	BCTRCHAN	- the bin centre in PI channels (not needed in a model)
!	BINCTREV	- the bin centre in eV
!	BINWIDEV	- the bin width in eV.
! plus a keyword
!	BWIDCHAN	- the bin width in PI channels
!
! All values are real32. The channel values don't have to be integer,
! fractional channel values can occur after rebinning.
!
    implicit none

!.................
! Arguments coming in:

    character(len=80), intent(in) ::&
      specSetName
    character(len=8), intent(in) ::&
      rebinSpecMode
    real, intent(in) ::&
      specBinDiv

!.................
! Arguments modified:

    logical, intent(in out) ::&
      useSpecSet,&
      rebinSpec
!    integer, intent(in out) ::&
!      numRBspecRows
 
!.................
! Arguments going out:

    type(DataSetT), intent(out) ::&
      tempSpecSet

!.................
! Others:

    type(DataSetT) ::&
      specSet,&
      RBspecSet
    type(TableT) ::&
      specTab,&
      tempSpecTab
    type(ColumnT) ::&
      specChannelCol,&
      specCountsCol,&
      specRateCol,&
      tempSpecRateCol,&
      tempSpecBCTReVcol,&
      tempSpecBWIDeVcol,&
      tempSpecBCTRchanCol
    integer ::&
      numSpecRows,&
      numTempSpecRows,&
      row,&
      EBindex,&
      numEBchannels,&
      EBstartChannel,&
      EBstartIndex,&
      specStartRow,&
      EBstopIndex,&
      startRow,&
      stopRow
    integer(kind=int16) ::&
      specFirstChan,&
      specLastChan,&
      tempSpecFirstChan,&
      tempSpecLastChan,&
      i
    character(len=20) ::&
      specChanType,&
      tempSpecSetName
    real(kind=single), dimension(:), pointer ::&
      specRate,&
      tempSpecRate,&
      tempSpecBCTReV,&
      tempSpecBCTRchan,&
      tempSpecBWIDeV
!    real(kind=single) ::&
!      tempSpecBWIDchan
    integer(kind=int16), dimension(:), pointer ::&
      specChannel,&
      EBchannels
    integer(kind=int32), dimension(:), pointer ::&
      specCounts
    real(kind=double), dimension(:), pointer ::&
      Emin,&
      Emax
    logical ::&
      specSetIsOpen,&
      hasMin,&
      hasMax,&
      hasDet,&
      hasCounts,&
      hasRate

!........................................................................
! First do some checks on the spectrum file:

    call message(AppMsg, Verbose, '  Checking, rebinning spectrum.')

    tempSpecSetName = './tempspecset.fits'

    specSetIsOpen = .false. ! default.
    if (useSpecSet) then
      specSet = dataSet(specSetName, READ)
      specTab = table(specSet, 'SPECTRUM')
      specSetIsOpen = .true.
      numSpecRows = numberOfRows(specTab)
    end if

    if (useSpecSet) then
      if (hasAttribute(specTab, 'CHANTYPE')) then
        specChanType = stringAttribute(specTab, 'CHANTYPE')
        if (trim(specChanType) .ne. 'PI') then
          useSpecSet = .false.
        end if
      else
        useSpecSet = .false.
      end if
    end if

    if (useSpecSet) then
      hasMin = hasAttribute(specTab, 'TLMIN001')
      hasMax = hasAttribute(specTab, 'TLMAX001')
      hasDet = hasAttribute(specTab, 'DETCHANS')
      if (.not. ((hasMin .and. hasMax .and. .not. hasDet)&
      .or. (.not. hasMin .and. .not. hasMax .and. hasDet))) then
        useSpecSet = .false.
      end if
    end if

    if (useSpecSet) then
      if (hasAttribute(specTab, 'TLMIN001')) then
        specFirstChan = int16Attribute(specTab, 'TLMIN001')
        specLastChan = int16Attribute(specTab, 'TLMAX001')
      else ! can now assume there is a DETCHANS keyword.
        specFirstChan = 0
        specLastChan = int16Attribute(specTab, 'DETCHANS') - 1
      end if

      if (1 + specLastChan - specFirstChan .ne. numSpecRows) then
        useSpecSet = .false.
      end if
    end if

    if (useSpecSet) then
      hasCounts = hasColumn(specTab, 'COUNTS')
      hasRate = hasColumn(specTab, 'RATE')
      if (.not. (hasColumn(specTab, 'CHANNEL')&
      .and. ((.not. hasCounts .and. hasRate)&
      .or. (hasCounts .and. .not. hasRate)))) then
        useSpecSet = .false.
      end if
    end if

    if (useSpecSet) then
      if (numSpecRows .lt. 1) then
        useSpecSet = .false.
      end if
    end if

!    if (useSpecSet) then
!      specChannelCol = column(specTab, 'CHANNEL')
!      specChannel => int16Data(specChannelCol)
!      do row = 1, numSpecRows
!        if (specChannel(row - 1) .ne. specFirstChan - 1 + row) then
!          useSpecSet = .false.
!      exit
!        end if
!      end do ! next row
!      call release(specChannelCol)
!    end if

    if (specSetIsOpen .and. .not. useSpecSet) then ! one of the previous
    ! little loops must have discovered something not right with specSet.

      write(errstr, *) "The structure and content of the &
      &spectrum dataset "//trim(specSetName)//" was not correct - wrong or &
      &missing keywords or columns, no data values, or inconsecutive &
      &channel numbers."

      call release(specSet)
call error('specBadStruct')
    end if

!........................................................................
! Now do some checks of CAL_getEbounds:

    if (useSpecSet) then
      call CAL_getEbounds(EBchannels, Emin, Emax)
      numEBchannels = size(EBchannels)
      EBstartChannel = EBchannels(0)
      do i = 2, numEBchannels
        if (EBchannels(i - 1) .ne. EBstartChannel - 1 + i) then
          useSpecSet = .false.
          write(errstr, *) "Non-consecutive array of channels &
          &returned by CAL_getEbounds."

        call CAL_releaseMemory(EBchannels)
        call CAL_releaseMemory(Emin, Emax)
        call release(specSet)
call error('EBbadChanOrder')
      exit
        end if
      end do ! next i
    end if

    if (useSpecSet) then
      if (Emin(0) .gt. srcTsEmin&
      .or. Emax(numEBchannels - 1) .lt. srcTsEmax) then
!print *, numEBchannels, Emin(0), srcTsEmin, Emax(numEBchannels - 1), srcTsEmax
        useSpecSet = .false.
        write(errstr, *) "The energy range returned by CAL_getEbounds &
        &does not include that of the source timeseries."

        call CAL_releaseMemory(EBchannels)
        call CAL_releaseMemory(Emin, Emax)
        call release(specSet)
call error('EBbadErange')
      end if
    end if

!........................................................................
! Check the energy bounds of the spectrum:

    if (useSpecSet) then
      if (EBchannels(0) .gt. specFirstChan) then
        specStartRow = EBchannels(0) - specFirstChan
      else
        specStartRow = 0
      end if
      tempSpecFirstChan = max(specFirstChan, EBchannels(0))
      tempSpecLastChan  = min(specLastChan,  EBchannels(numEBchannels - 1))

print *, tempSpecFirstChan, tempSpecLastChan, EBchannels(0), size(EBchannels), Emax(size(EBchannels)-1)
print *, Emin(EBstartIndex), srcTsEmin
print *, Emax(EBstopIndex), srcTsEmax
      EBstartIndex = tempSpecFirstChan - EBchannels(0)
      EBstopIndex  = tempSpecLastChan  - EBchannels(0)
      if (Emin(EBstartIndex) .gt. srcTsEmin&
      .or. Emax(EBstopIndex) .lt. srcTsEmax) then
        useSpecSet = .false.
        write(errstr, *) "The energy range in "//trim(specSetName)//&
        " does not entirely overlap that of the source timeseries."

        call CAL_releaseMemory(EBchannels)
        call CAL_releaseMemory(Emin, Emax)
        call release(specSet)
call error('specBadErange')
      end if
    end if

!........................................................................
! Make the temp measured spectrum file:

    tempSpecSet = dataSet(tempSpecSetName, TEMP)

    if (useSpecSet) then
      numTempSpecRows = 1 + tempSpecLastChan - tempSpecFirstChan
      tempSpecTab = addTable(tempSpecSet, 'SPEC', numTempSpecRows)

      ! Write the BINCTREV column of the temp spec file:
      tempSpecBCTReVcol = addColumn(tempSpecTab, 'BINCTREV', Real32, 'eV')
      tempSpecBCTReV => real32Data(tempSpecBCTReVcol)
      do row = 1, numTempSpecRows
        EBindex = (row - 1) + EBstartIndex
        tempSpecBCTReV(row - 1) = (Emin(EBindex) + Emax(EBindex)) / 2.0
      end do ! next row
      call release(tempSpecBCTReVcol)

      ! Write the BINWIDEV column of the temp spec file:
      tempSpecBWIDeVcol = addColumn(tempSpecTab, 'BINWIDEV', Real32, 'eV')
      tempSpecBWIDeV => real32Data(tempSpecBWIDeVcol)
      do row = 1, numTempSpecRows
        EBindex = (row - 1) + EBstartIndex
        tempSpecBWIDeV(row - 1) = Emax(EBindex) - Emin(EBindex)
      end do ! next row
      call release(tempSpecBWIDeVcol)
      call CAL_releaseMemory(EBchannels)
      call CAL_releaseMemory(Emin, Emax)

      ! Write the BCTRCHAN column of the temp spec file:
      tempSpecBCTRchanCol = addColumn(tempSpecTab, 'BCTRCHAN',&
                                      Real32, 'CHANNELS')
      tempSpecBCTRchan => real32Data(tempSpecBCTRchanCol)
      specChannelCol = column(specTab, 'CHANNEL')
      specChannel => int16Data(specChannelCol)
      do row = 1, numTempSpecRows
        tempSpecBCTRchan(row - 1)&
        = real(specChannel(specStartRow + row - 1), single)
      end do ! next row
      call release(specChannelCol)
      call release(tempSpecBCTRchanCol)

      ! Write the RATE column of the temp spec file:
      tempSpecRateCol = addColumn(tempSpecTab, 'RATE', Real32, 'COUNTS/CHANNEL')
      tempSpecRate => real32Data(tempSpecRateCol)
      if (hasColumn(specTab, 'COUNTS')) then
        specCountsCol = column(specTab, 'COUNTS')
        specCounts => int32Data(specCountsCol)
        do row = 1, numTempSpecRows
          tempSpecRate(row - 1)&
          = real(specCounts(specStartRow + row - 1), single)
        end do ! next row
        call release(specCountsCol)

      else ! RATE
        specRateCol = column(specTab, 'RATE')
        specRate => real32Data(specRateCol)
        do row = 1, numTempSpecRows
          tempSpecRate(row - 1) = specRate(specStartRow + row - 1)
        end do ! next row
        call release(specRateCol)
      end if

      call setAttribute(tempSpecTab, 'BWIDCHAN', 1.0,&
                        'Width of the bins in PI channels.')

      call release(tempSpecTab)

!........................................................................
! Prepare for rebinning the measured spectrum:

      if (rebinSpec .and. trim(rebinSpecMode) .eq. 'tonumber'&
     .and. numRBspecRows .ge. numTempSpecRows) then
        rebinSpec = .false.
        write(errstr, *) "Number of spectrum bins requested &
        &is greater than the number of channels in "//trim(specSetName)&
        //". Rebinning not done."

        call warning('tooManySpecBins')
      end if

      ! Calculate the number of bins in the rebinned spectrum (ie, the
      ! number of rows in the RBspec temp file):

      if (rebinSpec .and. trim(rebinSpecMode) .eq. 'divideby') then
!********** nint????? make it int(, int16)
        numRBspecRows = nint(numSpecRows / specBinDiv) ! The parameter
        ! limits should prevent specBinDiv being less than 1, and its
        ! existence should have been tested in the calling routine.

        if (numRBspecRows .ge. numSpecRows) then ! equality can happen
        ! if numRows is large and specBinDiv is just a little greater than
        ! 1. 'Greater than' shouldn't be able to happen, but one never knows.

          rebinSpec = .false. ! No need.
        end if
      end if

      if (rebinSpec .and. numRBspecRows .lt. leastNSpecBins) then
        numRBspecRows = leastNSpecBins
        write(errstr, *) "Requested number of spectrum bins &
        &is smaller than the minimum value of", leastNSpecBins, ". This &
        &minimum number has been used instead."

        call warning('tooFewSpecBins')
      end if

!........................................................................
! Now rebin the measured spectrum (if required):

      if (rebinSpec) then
        call old_rebinSpectrum(tempSpecSet, RBspecSet, numRBspecRows)

        tempSpecTab = table(tempSpecSet, 'SPEC')
        tempSpecBCTRchanCol = column(tempSpecTab, 'BCTRCHAN')
        tempSpecBCTRchan => real32Data(tempSpecBCTRchanCol)
      end if ! rebin spectrum requested.

      numTempSpecRows = numberOfRows(tempSpecTab)
      tempSpecBCTReVcol = column(tempSpecTab, 'BINCTREV')
      tempSpecBCTReV => real32Data(tempSpecBCTReVcol)

      startRow = numTempSpecRows + 1 ! sic
      do row = 1, numTempSpecRows
        if (tempSpecBCTReV(row - 1) .gt. srcTsEmin) then
          startRow = row
      exit
        end if
      end do

      stopRow = 0         ! sic
      do row = numTempSpecRows, 1, -1
        if (tempSpecBCTReV(row - 1) .lt. srcTsEmax) then
          stopRow = row
      exit
        end if
      end do

      if (stopRow .lt. startRow) then
        write(errstr, *) "There are no spectral bins within the energy &
        &range of the timeseries."
call error('noSpecBins')
      end if

      call release(tempSpecTab)

    else ! use model spectrum:
      call makeModelSpec(tempSpecSet)
      ! This adds a table 'SPEC' to tempSpecSet, + all the necessary columns;
      ! it then releases everything but the set (which is needed by the
      ! spatial routines below).
    end if ! use measured spectrum.

  end subroutine checkAndRebinSpectrum

  !---------------------------------------------------------------------
  subroutine old_rebinSpectrum(inSpecSet, outSpecSet, numOutSpecRows)
  !---------------------------------------------------------------------

    implicit none

!.................
! Arguments coming in:

    integer(kind=int16), intent(in) ::&
      numOutSpecRows

!.................
! Arguments modified:

    type(DataSetT), intent(in out) ::&
      inSpecSet

!.................
! Arguments going out:

    type(DataSetT), intent(out) ::&
      outSpecSet

!.................
! Others:

    integer(kind=int16) ::&
      row,&
      numInSpecRows,&
      inSpecStartRow,&
      inSpecStopRow,&
      inSpecRow
    real(kind=single) ::&
      inSpecBinWidth,&
      outSpecBinWidth,&
      startP,&
      stopP,&
      deltaP,&
      inStartP,&
      inStopP,&
      inStartE,&
      inStopE,&
      outStartE,&
      outStopE
    type(BlockT) ::&
      outSpecBlock
    type(TableT) ::&
      inSpecTab,&
      outSpecTab
    type(ColumnT) ::&
      inSpecBCTRchanCol,&
      inSpecBINCTREVcol,&
      inSpecBINWIDEVcol,&
      inSpecRateCol,&
      outSpecBCTRchanCol,&
      outSpecBINCTREVcol,&
      outSpecBINWIDEVcol,&
      outSpecRateCol
    real(kind=single), dimension(:), pointer ::&
      inSpecBCTRchan,&
      inSpecBINCTREV,&
      inSpecBINWIDEV,&
      inSpecRate,&
      outSpecBCTRchan,&
      outSpecBINCTREV,&
      outSpecBINWIDEV,&
      outSpecRate

!........................................................................
! Start:

    ! InSet should have the following columns:
    !	BCTRCHAN	real32
    !	RATE		real32
    !	BINCTREV	real32
    !	BINWIDEV	real32
    ! plus a BWIDCHAN keyword.
    !
    ! The following routine regards the channels of the input spectrum
    ! as bins over a real-valued coordinate p, such that channel C
    ! extends from p = C - 0.5 to C + 0.5. In the output (ie, rebinned)
    ! spectrum the 'channels' may no longer have these pseudo-integer
    ! values.

    inSpecTab = table(inSpecSet, 'SPEC')
    numInSpecRows = numberOfRows(inSpecTab)
    if (numInSpecRows .lt. numOutSpecRows) then
  return
    end if
    inSpecBinWidth = real32attribute(inSpecTab, 'BWIDCHAN')

    outSpecSet = dataSet('rebinnedspec.fits', TEMP)
    outSpecTab = addTable(outSpecSet, 'SPEC', int(numOutSpecRows))

    inSpecBCTRchanCol = column(inSpecTab, 'BCTRCHAN')
    inSpecBCTRchan => real32Data(inSpecBCTRchanCol)
    startP = inSpecBCTRchan(0) - 0.5
    stopP = inSpecBCTRchan(numInSpecRows - 1) + 0.5
    deltaP = stopP - startP
    outSpecBinWidth = deltaP / numOutSpecRows
    call setAttribute(outSpecTab, 'BWIDCHAN', outSpecBinWidth,&
                      'Width of the bins in PI channels.')

    ! Calculate the new bin 'channel' centres:
    outSpecBCTRchanCol = addColumn(outSpecTab, 'BCTRCHAN', Real32, 'CHANNELS')
    outSpecBCTRchan => real32Data(outSpecBCTRchanCol)
    do row = 1, numOutSpecRows
      outSpecBCTRchan(row - 1) = startP + deltaP&
      * ((row - 0.5) / numOutSpecRows)
    end do ! next row
    call release(outSpecBCTRchanCol)
    call release(inSpecBCTRchanCol)

    ! Rebin the rate values:
    inSpecRateCol = column(inSpecTab, 'RATE')
    inSpecRate => real32Data(inSpecRateCol)
    outSpecRateCol = addColumn(outSpecTab, 'RATE', Real32, 'COUNTS/CHANNEL')
    outSpecRate => real32Data(outSpecRateCol)

    do row = 1, numOutSpecRows
      inStartP = (row - 1.0)      * outSpecBinWidth / inSpecBinWidth
      inStopP = real(row, single) * outSpecBinWidth / inSpecBinWidth
      inSpecStartRow = 1 + floor(inStartP)
      inSpecStopRow  = 1 + floor(inStopP)
      if (inSpecStopRow .gt. numInSpecRows) then
        inSpecStopRow = numInSpecRows
      end if
      outSpecRate(row - 1) = (inSpecStartRow - inStartP / inSpecBinWidth)&
      * inSpecRate(inSpecStartRow - 1)

      do inSpecRow = inSpecStartRow + 1, inSpecStopRow - 1
        outSpecRate(row - 1) = outSpecRate(row - 1) + inSpecRate(inSpecRow)
      end do

      outSpecRate(row - 1) = outSpecRate(row - 1) + (inStopP / inSpecBinWidth&
      - inSpecStopRow - 1) * inSpecRate(inSpecStopRow - 1)
    end do ! next row
    call release(outSpecRateCol)
    call release(inSpecRateCol)

    ! Rebin the energy width and centre:
    inSpecBINCTREVcol = column(inSpecTab, 'BINCTREV')
    inSpecBINWIDEVcol = column(inSpecTab, 'BINWIDEV')
    inSpecBINCTREV => real32Data(inSpecBINCTREVcol)
    inSpecBINWIDEV => real32Data(inSpecBINWIDEVcol)
    outSpecBINCTREVcol = addColumn(outSpecTab, 'BINCTREV', Real32, 'eV')
    outSpecBINWIDEVcol = addColumn(outSpecTab, 'BINWIDEV', Real32, 'eV')
    outSpecBINCTREV => real32Data(outSpecBINCTREVcol)
    outSpecBINWIDEV => real32Data(outSpecBINWIDEVcol)
    do row = 1, numOutSpecRows
      inStartP = (row - 1.0) * outSpecBinWidth / inSpecBinWidth
      inStopP =   row        * outSpecBinWidth / inSpecBinWidth
      inSpecStartRow = 1 + floor(inStartP)
      inSpecStopRow  = 1 + floor(inStopP)
      if (inSpecStopRow .gt. numInSpecRows) then
        inSpecStopRow = numInSpecRows
      end if

      inStartE = inSpecBINCTREV(inSpecStartRow - 1)&
      - inSpecBINWIDEV(inSpecStartRow - 1) / 2.0
      inStopE = inSpecBINCTREV(inSpecStartRow - 1)&
      + inSpecBINWIDEV(inSpecStartRow - 1) / 2.0
      outStartE = inStartE + (inSpecStartRow - inStartP / inSpecBinWidth)&
      * inSpecBINWIDEV(inSpecStartRow - 1)

      inStartE = inSpecBINCTREV(inSpecStopRow - 1)&
      - inSpecBINWIDEV(inSpecStopRow - 1) / 2.0
      inStopE = inSpecBINCTREV(inSpecStopRow - 1)&
      + inSpecBINWIDEV(inSpecStopRow - 1) / 2.0
      outStopE = inStartE + (inStopP / inSpecBinWidth - inSpecStopRow - 1)&
      * inSpecBINWIDEV(inSpecStopRow - 1)

      outSpecBINCTREV(row - 1) = (outStopE + outStartE) / 2.0
      outSpecBINWIDEV(row - 1) =  outStopE - outStartE
    end do ! next row

    call release(inSpecTab)
    call deleteBlock(inSpecSet, 'SPEC')
    outSpecBlock = block(outSpecSet, 'SPEC')
    call copyBlock(inSpecSet, outSpecBlock)
    call release(outSpecSet)

  end subroutine old_rebinSpectrum

  !---------------------------------------------------------------------
  subroutine makeModelSpec(specSet)
  !---------------------------------------------------------------------

    implicit none

!.................
! Arguments coming in:

    type(DataSetT), intent(in) ::&
      specSet

!.................
! Others:

    real(kind=single) ::&
      eVrange
    type(TableT)   ::&
      specTab                      
    type(ColumnT)  ::&
      specRateCol,&
      specBCTReVcol,&
      specBWIDeVcol
    real(kind=single), dimension(:), pointer ::&
      specRate,&
      specBCTReV,&
      specBWIDeV
    integer ::&
      row
!,&
!      numSpecRows = 10

!........................................................................
! Start:

    numSpecSamples = 10
    eVrange = srcTsEmax - srcTsEmin

    specTab = addTable(specSet, 'SPEC', numSpecSamples)

    specRateCol   = addColumn(specTab, 'RATE', Real32, 'COUNTS/BIN')
    specBCTReVcol = addColumn(specTab, 'BINCTREV', Real32, 'eV')
    specBWIDeVcol = addColumn(specTab, 'BINWIDEV', Real32, 'eV')
    specRate   => real32Data(specRateCol)
    specBCTReV => real32Data(specBCTReVcol)
    specBWIDeV => real32Data(specBWIDeVcol)

    specRate = 1.0 / eVrange ! vector op: normalised, flat spectrum.

    specBWIDeV = eVrange / real(numSpecSamples, single) ! vector op.

    do row = 1, numSpecSamples
      specBCTReV(row - 1) = srcTsEmin + eVrange * (row - 0.5)&
      / real(numSpecSamples, single)
    end do ! next row

    call release(specBWIDeVcol)
    call release(specBCTReVcol)
    call release(specRateCol)
    call release(specTab)

  end subroutine makeModelSpec
end module lccorr_mod
